﻿// 
// Copyright (c) 2008-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.ComponentModel;

namespace Smdn.Net.Imap4.Client {
  partial class ImapClient {
    public ISynchronizeInvoke SynchronizingObject {
      get; set;
    }

    public event EventHandler<ImapMailboxSizeChangedEventArgs> ExistMessageCountChanged;
    public event EventHandler<ImapMailboxSizeChangedEventArgs> RecentMessageCountChanged;
    public event EventHandler<ImapMessageStatusChangedEventArgs> MessageDeleted;
    public event EventHandler<ImapMessageStatusChangedEventArgs> MessageStatusChanged;
    public event EventHandler<ImapAlertReceivedEventArgs> AlertReceived;

    internal enum EventType {
      ExistMessageCountChanged,
      RecentMessageCountChanged,
      MessageDeleted,
      MessageStatusChanged,
      AlertReceived,
    }

    private class EventQueueEntry {
      public EventType EventType { get; private set; }
      public EventArgs EventArgs { get; private set; }

      public EventQueueEntry(EventType eventType, EventArgs eventArgs)
      {
        this.EventType = eventType;
        this.EventArgs = eventArgs;
      }
    }

    private Queue<EventQueueEntry> eventQueue = new Queue<EventQueueEntry>();

    internal void RaiseEvent(EventType eventType, EventArgs eventArgs, bool raiseImmediately)
    {
#if DEBUG
      ThrowIfNotInExclusiveOperation();
#endif

      var entry = new EventQueueEntry(eventType, eventArgs);

      if (raiseImmediately)
        InvokeEvent(entry);
      else
        eventQueue.Enqueue(entry);
    }

    private void RaiseQueuedEvents()
    {
      EventQueueEntry[] entries;

      lock (eventQueue) {
        if (eventQueue.Count == 0)
          return;

        entries = eventQueue.ToArray();

        eventQueue.Clear();
      }

      foreach (var entry in entries) {
        InvokeEvent(entry);
      }
    }

    private void InvokeEvent(EventQueueEntry entry)
    {
      switch (entry.EventType) {
        case EventType.ExistMessageCountChanged:
          InvokeEvent(ExistMessageCountChanged, entry.EventArgs as ImapMailboxSizeChangedEventArgs);
          break;

        case EventType.RecentMessageCountChanged:
          InvokeEvent(RecentMessageCountChanged, entry.EventArgs as ImapMailboxSizeChangedEventArgs);
          break;

        case EventType.MessageDeleted:
          InvokeEvent(MessageDeleted, entry.EventArgs as ImapMessageStatusChangedEventArgs);
          break;

        case EventType.MessageStatusChanged:
          InvokeEvent(MessageStatusChanged, entry.EventArgs as ImapMessageStatusChangedEventArgs);
          break;

        case EventType.AlertReceived:
          InvokeEvent(AlertReceived, entry.EventArgs as ImapAlertReceivedEventArgs);
          break;

        default:
#if DEBUG
          ExceptionUtils.CreateArgumentMustBeValidEnumValue("eventType", entry.EventType);
#endif
          break;

      }
    }

    private void InvokeEvent<TEventArgs>(EventHandler<TEventArgs> ev, TEventArgs args)
#if !NET_4_5
    where TEventArgs : EventArgs
#endif
    {
      if (ev == null)
        return;

      if (SynchronizingObject == null || !SynchronizingObject.InvokeRequired) {
        try {
          ev(this, args);
        }
        catch {
          // ignore exceptions
        }
      }
      else {
        SynchronizingObject.BeginInvoke(ev, new object[] {this, args});
      }
    }
  }
}