//
// Author:
//       smdn <smdn@smdn.jp>
//
// Copyright (c) 2008-2014 smdn
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
using System;

namespace Smdn.Net.Imap4.Client {
  partial class ImapClient {
    private const bool DefaultRecursive = false;

    public static ImapMailboxInfo MoveMailbox(ImapMailboxInfo mailbox,
                                              ImapClient destinationSession,
                                              string destinationMailboxName)
    {
      return MoveMailbox(mailbox,
                         destinationSession,
                         destinationMailboxName,
                         DefaultRecursive,
                         DefaultSubscription);
    }

    public static ImapMailboxInfo MoveMailbox(ImapMailboxInfo mailbox,
                                              ImapClient destinationSession,
                                              string destinationMailboxName,
                                              bool recursive)
    {
      return MoveMailbox(mailbox,
                         destinationSession,
                         destinationMailboxName,
                         recursive,
                         DefaultSubscription);
    }

    public static ImapMailboxInfo MoveMailbox(ImapMailboxInfo mailbox,
                                              ImapClient destinationSession,
                                              string destinationMailboxName,
                                              bool recursive,
                                              bool subscribe)
    {
      if (mailbox == null)
        throw new ArgumentNullException("mailbox");
      if (destinationSession == null)
        throw new ArgumentNullException("destinationSession");

      return CopyMailboxAcrossSession(mailbox,
                                      destinationSession,
                                      GetValidatedMailboxName(destinationMailboxName, "destinationMailboxName"),
                                      recursive,
                                      subscribe,
                                      true);
    }

    public static ImapMailboxInfo CopyMailbox(ImapMailboxInfo mailbox,
                                              ImapClient destinationSession,
                                              string destinationMailboxName)
    {
      return CopyMailbox(mailbox,
                         destinationSession,
                         destinationMailboxName,
                         DefaultRecursive,
                         DefaultSubscription);
    }

    public static ImapMailboxInfo CopyMailbox(ImapMailboxInfo mailbox,
                                              ImapClient destinationSession,
                                              string destinationMailboxName,
                                              bool recursive)
    {
      return CopyMailbox(mailbox,
                         destinationSession,
                         destinationMailboxName,
                         recursive,
                         DefaultSubscription);
    }

    public static ImapMailboxInfo CopyMailbox(ImapMailboxInfo mailbox,
                                              ImapClient destinationSession,
                                              string destinationMailboxName,
                                              bool recursive,
                                              bool subscribe)
    {
      if (mailbox == null)
        throw new ArgumentNullException("mailbox");
      if (destinationSession == null)
        throw new ArgumentNullException("destinationSession");

      return CopyMailboxAcrossSession(mailbox,
                                      destinationSession,
                                      GetValidatedMailboxName(destinationMailboxName, "destinationMailboxName"),
                                      recursive,
                                      subscribe,
                                      false);
    }

    public static void MoveMailbox(ImapMailboxInfo mailbox,
                                   ImapMailboxInfo destinationMailbox)
    {
      MoveMailbox(mailbox,
                  destinationMailbox,
                  DefaultRecursive,
                  DefaultSubscription);
    }

    public static void MoveMailbox(ImapMailboxInfo mailbox,
                                   ImapMailboxInfo destinationMailbox,
                                   bool recursive)
    {
      MoveMailbox(mailbox,
                  destinationMailbox,
                  recursive,
                  DefaultSubscription);
    }

    public static void MoveMailbox(ImapMailboxInfo mailbox,
                                   ImapMailboxInfo destinationMailbox,
                                   bool recursive,
                                   bool subscribe)
    {
      if (mailbox == null)
        throw new ArgumentNullException("mailbox");
      if (destinationMailbox == null)
        throw new ArgumentNullException("destinationMailbox");

      destinationMailbox.ThrowIfCanNotHaveChild();

      CopyMailboxAcrossSession(mailbox,
                               destinationMailbox.Client,
                               destinationMailbox.GetFullNameOf(mailbox.Name),
                               recursive,
                               subscribe,
                               true);
    }

    public static void CopyMailbox(ImapMailboxInfo mailbox,
                                   ImapMailboxInfo destinationMailbox)
    {
      CopyMailbox(mailbox,
                  destinationMailbox,
                  DefaultRecursive,
                  DefaultSubscription);
    }

    public static void CopyMailbox(ImapMailboxInfo mailbox,
                                   ImapMailboxInfo destinationMailbox,
                                   bool recursive)
    {
      CopyMailbox(mailbox,
                  destinationMailbox,
                  recursive,
                  DefaultSubscription);
    }

    public static void CopyMailbox(ImapMailboxInfo mailbox,
                                   ImapMailboxInfo destinationMailbox,
                                   bool recursive,
                                   bool subscribe)
    {
      if (mailbox == null)
        throw new ArgumentNullException("mailbox");
      if (destinationMailbox == null)
        throw new ArgumentNullException("destinationMailbox");

      destinationMailbox.ThrowIfCanNotHaveChild();

      CopyMailboxAcrossSession(mailbox,
                               destinationMailbox.Client,
                               destinationMailbox.GetFullNameOf(mailbox.Name),
                               recursive,
                               subscribe,
                               false);
    }

    private static ImapMailboxInfo CopyMailboxAcrossSession(ImapMailboxInfo sourceMailbox,
                                                            ImapClient destinationSession,
                                                            string destinationMailboxName,
                                                            bool recursive,
                                                            bool subscribe,
                                                            bool deleteOrigin)
    {
      var destinationMailbox = destinationSession.GetOrCreateMailbox(destinationMailboxName, subscribe);

      // copy messages
      using (var sourceSelectedMailbox = sourceMailbox.Open(true)) {
        if (sourceSelectedMailbox.Client == destinationMailbox.Client) {
          sourceSelectedMailbox.CopyMessagesTo(destinationMailbox);
        }
        else {
          foreach (var message in sourceSelectedMailbox.GetMessages(ImapMessageFetchAttributeOptions.None)) {
            destinationMailbox.AppendMessage(message);
          }
        }
      }

      // copy inferior mailboxes
      if (recursive) {
        foreach (var sourceChildMailbox in sourceMailbox.GetChildren(ImapMailboxListOptions.TopLevelOnly)) {
          var destinationChildMailboxName = destinationMailbox.GetFullNameOf(sourceChildMailbox.Name);

          CopyMailboxAcrossSession(sourceChildMailbox,
                                   destinationSession,
                                   destinationChildMailboxName,
                                   recursive,
                                   subscribe,
                                   deleteOrigin);
        }
      }

      if (deleteOrigin && !sourceMailbox.Mailbox.IsInbox)
        sourceMailbox.Delete(subscribe);

      return destinationMailbox;
    }
  }
}

