using System;
using System.IO;
using System.Text;
using System.Threading;
using NUnit.Framework;

namespace Smdn.Net.Imap4.Client.Session {
  [TestFixture]
  public class ImapSessionCommandsAppendTests : ImapSessionTestsBase {
    [Test]
    public void TestAppend()
    {
      SelectMailbox2(delegate(ImapSession session, ImapPseudoServer2 server) {
        server.EnqueueTaggedResponse("APPEND INBOX ",
                                     "$tag OK APPEND completed\r\n");

        var message = new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                            new ImapMessageFlagSet(ImapMessageFlag.Seen));

        Assert.IsTrue((bool)session.Append(message, "INBOX"));

        Assert.AreEqual("0004 APPEND INBOX (\\Seen) {33}\r\n" +
                        "MIME-Version: 1.0\r\n" +
                        "\r\n" +
                        "test message\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestAppend_NonSynchronizedLiterals()
    {
      SelectMailbox2(new[] {"LITERAL+"}, delegate(ImapSession session, ImapPseudoServer2 server) {
        server.EnqueueTaggedResponse("APPEND INBOX ",
                                     "$tag OK APPEND completed\r\n");

        var message = new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                            new ImapMessageFlagSet(ImapMessageFlag.Seen));

        Assert.IsTrue((bool)session.Append(message, "INBOX"));

        Assert.AreEqual("0004 APPEND INBOX (\\Seen) {33+}\r\n" +
                        "MIME-Version: 1.0\r\n" +
                        "\r\n" +
                        "test message\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestAppend_TryCreate()
    {
      SelectMailbox2(delegate(ImapSession session, ImapPseudoServer2 server, Uri expectedAuthority) {
        session.TransactionTimeout = 500;
        server.EnqueueLiteralContinuationResponse("APPEND INBOX.appendto ",
                                                  "$tag NO [TRYCREATE] Mailbox doesn't exist: INBOX.appendto\r\n",
                                                  false);
        server.EnqueueTaggedResponse("CREATE INBOX.appendto\r\n",
                                     "$tag OK CREATE completed\r\n");
        server.EnqueueTaggedResponse("APPEND INBOX.appendto ",
                                     "$tag OK APPEND completed\r\n");

        var message = new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                            new ImapMessageFlagSet(new[] {"$label2"}, ImapMessageFlag.Draft));

        ImapMailbox created;

        Assert.IsTrue((bool)session.Append(message, "INBOX.appendto", out created));

        Assert.AreEqual("INBOX.appendto", created.Name);
        Assert.AreEqual(expectedAuthority + "INBOX.appendto", created.Url.AbsoluteUri);
        Assert.IsNotNull(created.Flags);
        Assert.IsNotNull(created.ApplicableFlags);
        Assert.IsNotNull(created.PermanentFlags);

        Assert.AreEqual("0004 APPEND INBOX.appendto (\\Draft $label2) {33}\r\n",
                        server.DequeueRequest());
        Assert.AreEqual("0005 CREATE INBOX.appendto\r\n",
                        server.DequeueRequest());

        Assert.AreEqual("0006 APPEND INBOX.appendto (\\Draft $label2) {33}\r\n" +
                        "MIME-Version: 1.0\r\n" +
                        "\r\n" +
                        "test message\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestAppend_TryCreate_NoReferralResponseCode()
    {
      SelectMailbox2(delegate(ImapSession session, ImapPseudoServer2 server) {
        server.EnqueueLiteralContinuationResponse("APPEND SHARED/FOO ",
                                                  "$tag NO [REFERRAL IMAP://user;AUTH=*@SERVER2/SHARED/FOO] Remote mailbox. Try SERVER2.\r\n",
                                                  false);

        var message = new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                            new ImapMessageFlagSet(new[] {"$label2"}, ImapMessageFlag.Draft));

        ImapMailbox created;

        Assert.IsFalse((bool)session.Append(message, "SHARED/FOO", out created));

        Assert.IsNull(created);

        Assert.AreEqual("0004 APPEND SHARED/FOO (\\Draft $label2) {33}\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestAppend_TryCreate_DontTryCreate()
    {
      SelectMailbox2(delegate(ImapSession session, ImapPseudoServer2 server) {
        server.EnqueueLiteralContinuationResponse("APPEND INBOX.appendto ",
                                                  "$tag NO [TRYCREATE] Mailbox doesn't exist: INBOX.appendto\r\n",
                                                  false);

        var message = new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                            new DateTimeOffset(2008, 2, 25, 15, 1, 12, new TimeSpan(+9, 0, 0)));

        var internalDate = "\"25-Feb-2008 15:01:12 +0900\"";

        Assert.IsFalse((bool)session.Append(message, "INBOX.appendto"));

        Assert.AreEqual("0004 APPEND INBOX.appendto " + internalDate + " {33}\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestAppend_RemoveNonApplicableFlags()
    {
      SelectMailbox2(delegate(ImapSession session, ImapPseudoServer2 server) {
        server.EnqueueTaggedResponse("APPEND INBOX ",
                                     "$tag OK APPEND completed\r\n");

        var message = new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                            new ImapMessageFlagSet(ImapMessageFlag.Seen,
                                                                    ImapMessageFlag.Recent,
                                                                    ImapMessageFlag.AllowedCreateKeywords));

        Assert.IsTrue((bool)session.Append(message, "INBOX"));

        Assert.AreEqual("0004 APPEND INBOX (\\Seen) {33}\r\n" +
                        "MIME-Version: 1.0\r\n" +
                        "\r\n" +
                        "test message\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestAppend_WithAppendUidResponseCode()
    {
      SelectMailbox2(delegate(ImapSession session, ImapPseudoServer2 server) {
        server.EnqueueTaggedResponse("APPEND INBOX ",
                                     "$tag OK [APPENDUID 38505 3955] APPEND completed\r\n");

        var message = new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                            new ImapMessageFlagSet(ImapMessageFlag.Seen));

        ImapAppendedUidSet appended;

        Assert.IsTrue((bool)session.Append(message, "INBOX", out appended));

        Assert.AreEqual("0004 APPEND INBOX (\\Seen) {33}\r\n" +
                        "MIME-Version: 1.0\r\n" +
                        "\r\n" +
                        "test message\r\n",
                        server.DequeueRequest());

        Assert.IsNotNull(appended);
        Assert.AreEqual(38505L, appended.UidValidity);
        Assert.AreEqual(3955L, appended.ToNumber());
        Assert.AreEqual("3955", appended.ToString());

        Smdn.Net.TestUtils.SerializeBinary(appended, delegate(ImapAppendedUidSet deserialized) {
          Assert.AreEqual(38505L, deserialized.UidValidity);
          Assert.AreEqual("3955", deserialized.ToString());
        });
      });
    }

    [Test]
    public void TestPrepareAppend()
    {
      PrepareAppend(false, false);
    }

    [Test]
    public void TestPrepareAppend_LengthSpecified()
    {
      PrepareAppend(false, true);
    }

    [Test]
    public void TestPrepareAppend_NonSynchronizedLiterals()
    {
      PrepareAppend(true, false);
    }

    [Test]
    public void TestPrepareAppend_LengthSpecified_NonSynchronizedLiterals()
    {
      PrepareAppend(true, true);
    }

    private void WaitForTransactionFinished(ImapSession session)
    {
      var sw = System.Diagnostics.Stopwatch.StartNew();
      var timeout = TimeSpan.FromMilliseconds(1000);

      for (;;) {
        if (!session.IsTransactionProceeding)
          break;

        if (timeout < sw.Elapsed)
          throw new TimeoutException("transaction not finished");

        Thread.Sleep(10);
      }
    }

    private void PrepareAppend(bool nonSynchronizedLiterals, bool specifyLength)
    {
      var messageBody = @"Date: Mon, 7 Feb 1994 21:52:25 -0800 (PST)
From: Fred Foobar <foobar@Blurdybloop.example.COM>
Subject: afternoon meeting
To: mooch@owatagu.example.net
Message-Id: <B27397-0100000@Blurdybloop.example.COM>
MIME-Version: 1.0
Content-Type: TEXT/PLAIN; CHARSET=US-ASCII

Hello Joe, do you think we can meet at 3:30 tomorrow?
".Replace("\r\n", "\n").Replace("\n", "\r\n");

      var caps = nonSynchronizedLiterals
        ? new[] {"IMAP4REV1", "LITERAL+"}
        : new[] {"IMAP4REV1"};

      Authenticate2(caps, delegate(ImapSession session, ImapPseudoServer2 server) {
        var messageBodyBytes = NetworkTransferEncoding.Transfer7Bit.GetBytes(messageBody);
        var length = specifyLength ? (long?)messageBodyBytes.Length : null;
        var appendContext = session.PrepareAppend(length, null,  null, "INBOX");

        Assert.IsNotNull(appendContext);
        Assert.IsNotNull(appendContext.WriteStream);
        Assert.IsTrue(session.IsTransactionProceeding);

        server.EnqueueTaggedResponse("APPEND INBOX ",
                                     "$tag OK [APPENDUID 38505 3955] APPEND completed\r\n");

        appendContext.WriteStream.Write(messageBodyBytes,
                                        0,
                                        messageBodyBytes.Length);

        if (specifyLength) {
          Assert.DoesNotThrow(() => WaitForTransactionFinished(session));
        }
        else {
          Assert.IsTrue(session.IsTransactionProceeding, "IsTransactionProceeding after Write");
          Assert.Throws<TimeoutException>(() => WaitForTransactionFinished(session)); // transaction will not finish (wait for stream closed)
        }

        ImapAppendedUidSet appendedUid;

        Assert.IsTrue((bool)appendContext.GetResult(out appendedUid));
        Assert.IsFalse(session.IsTransactionProceeding, "IsTransactionProceeding after GetResult");

        var expected =
          string.Format("0002 APPEND INBOX {{{0}}}\r\n", messageBodyBytes.Length) +
          messageBody +
          "\r\n";

        Assert.AreEqual(expected, server.DequeueRequest());

        Assert.IsNotNull(appendedUid);
        Assert.AreEqual(38505L, appendedUid.UidValidity);
        Assert.AreEqual(3955L, appendedUid.ToNumber());

        Smdn.Net.TestUtils.SerializeBinary(appendedUid, delegate(ImapAppendedUidSet deserialized) {
          Assert.AreEqual(38505L, deserialized.UidValidity);
          Assert.AreEqual("3955", deserialized.ToString());
        });
      });
    }

    [Test]
    public void TestPrepareAppend_AppendContext_WriteStream()
    {
      Authenticate2(delegate(ImapSession session, ImapPseudoServer2 server) {
        var appendContext = session.PrepareAppend(null, null,  null, "INBOX");

        Assert.IsTrue(session.IsTransactionProceeding, "IsTransactionProceeding after PrepareAppend");

        Assert.IsNotNull(appendContext);
        Assert.IsNotNull(appendContext.WriteStream, "appendContext.WriteStream initial");

        var writeStream = appendContext.WriteStream;

        Assert.Throws<NotSupportedException>(() => writeStream.Position = 0L, "set_Position");
        Assert.Throws<NotSupportedException>(() => Assert.AreEqual(0L, writeStream.Position), "get_Position");
        Assert.Throws<NotSupportedException>(() => Assert.AreEqual(0L, writeStream.Length), "get_Length");
        Assert.Throws<NotSupportedException>(() => writeStream.Seek(0L, SeekOrigin.Begin), "Seek");

        var buffer = new byte[8];
        Assert.Throws<NotSupportedException>(() => writeStream.Read(buffer, 0, buffer.Length), "Read");

        writeStream.Close();

        Assert.IsNull(appendContext.WriteStream, "appendContext.WriteStream after close");
        Assert.IsTrue(session.IsTransactionProceeding, "IsTransactionProceeding after Close");
        Assert.Throws<TimeoutException>(() => WaitForTransactionFinished(session)); // transaction will not finish (wait for response)

        Assert.Throws<ObjectDisposedException>(() => writeStream.WriteByte(0), "Write after Close");
        Assert.DoesNotThrow(() => writeStream.Close(), "Close agin");

        ImapAppendedUidSet appendedUid;

        server.EnqueueTaggedResponse("APPEND INBOX ",
                                     "$tag OK APPEND completed\r\n");

        Assert.IsTrue((bool)appendContext.GetResult(out appendedUid));
        Assert.IsFalse(session.IsTransactionProceeding, "IsTransactionProceeding after GetResult");

        Assert.AreEqual("0002 APPEND INBOX {0}\r\n\r\n", server.DequeueRequest());

        Assert.IsNull(appendedUid);
      });
    }

    [Test]
    public void TestPrepareAppend_AppendContext_WriteStream_BufferUnderrun()
    {
      Authenticate2(delegate(ImapSession session, ImapPseudoServer2 server) {
        session.SendTimeout = 500;
        session.ReceiveTimeout = 500;

        server.EnqueueLiteralContinuationResponse("APPEND INBOX ",
                                                  "+ OK continue\r\n",
                                                  true);

        var appendContext = session.PrepareAppend(null, null, null, "INBOX");

        Assert.IsNotNull(appendContext);
        Assert.IsNotNull(appendContext.WriteStream);
        Assert.IsTrue(session.IsTransactionProceeding, "IsTransactionProceeding after PrepareAppend");

        ImapAppendedUidSet appendedUid;

        Assert.Throws<TimeoutException>(() => appendContext.GetResult(out appendedUid));

        Assert.IsFalse(session.IsTransactionProceeding, "IsTransactionProceeding after GetResult");
      });
    }

    [Test]
    public void TestPrepareAppend_AppendContext_WriteStream_WriteOverSpecifiedLength()
    {
      Authenticate2(delegate(ImapSession session, ImapPseudoServer2 server) {
        var messageBody = "message";
        var messageBodyBytes = NetworkTransferEncoding.Transfer7Bit.GetBytes(messageBody);

        server.EnqueueLiteralContinuationResponse("APPEND INBOX ",
                                                  "+ OK continue\r\n",
                                                  true);

        var appendContext = session.PrepareAppend(messageBodyBytes.Length, null, null, "INBOX");

        Assert.IsNotNull(appendContext);
        Assert.IsNotNull(appendContext.WriteStream);
        Assert.IsTrue(session.IsTransactionProceeding, "IsTransactionProceeding after PrepareAppend");

        appendContext.WriteStream.Write(messageBodyBytes,
                                        0,
                                        messageBodyBytes.Length);

        // extra data
        appendContext.WriteStream.WriteByte(0x40);
        appendContext.WriteStream.WriteByte(0x40);
        appendContext.WriteStream.WriteByte(0x40);
        appendContext.WriteStream.WriteByte(0x40);

        Assert.IsTrue(session.IsTransactionProceeding, "IsTransactionProceeding after Write");

        server.EnqueueTaggedResponse("$tag OK APPEND completed\r\n");

        ImapAppendedUidSet appendedUid;

        Assert.IsTrue((bool)appendContext.GetResult(out appendedUid));
        Assert.IsFalse(session.IsTransactionProceeding, "IsTransactionProceeding after GetResult");

        var expected =
          string.Format("0002 APPEND INBOX {{{0}}}\r\n", messageBodyBytes.Length) +
          messageBody +
          "\r\n";

        Assert.AreEqual(expected, server.DequeueRequest());
      });
    }

    [Test]
    public void TestPrepareAppend_AppendContext_WriteStream_WriteShortageSpecifiedLength()
    {
      Authenticate2(delegate(ImapSession session, ImapPseudoServer2 server) {
        const int shortage = 8;

        var messageBody = "message";
        var messageBodyBytes = NetworkTransferEncoding.Transfer7Bit.GetBytes(messageBody);

        session.SendTimeout = 500;
        session.ReceiveTimeout = 500;

        var appendContext = session.PrepareAppend(messageBodyBytes.Length + shortage, null, null, "INBOX");

        Assert.IsNotNull(appendContext);
        Assert.IsNotNull(appendContext.WriteStream);
        Assert.IsTrue(session.IsTransactionProceeding, "IsTransactionProceeding after PrepareAppend");

        appendContext.WriteStream.Write(messageBodyBytes,
                                        0,
                                        messageBodyBytes.Length);

        ImapAppendedUidSet appendedUid;

        Assert.Throws<TimeoutException>(() => appendContext.GetResult(out appendedUid));
        Assert.IsFalse(session.IsTransactionProceeding, "IsTransactionProceeding after GetResult");
      });
    }

    [Test]
    public void TestPrepareAppend_NestedCall()
    {
      Authenticate2(delegate(ImapSession session, ImapPseudoServer2 server) {
        var appendContext = session.PrepareAppend(null, null, null, "INBOX");

        Assert.IsNotNull(appendContext);
        Assert.IsNotNull(appendContext.WriteStream);
        Assert.IsTrue(session.IsTransactionProceeding, "IsTransactionProceeding after PrepareAppend");

        Assert.Throws<InvalidOperationException>(() => session.PrepareAppend(null, null, null, "INBOX"));

        var messageBodyBytes = NetworkTransferEncoding.Transfer7Bit.GetBytes("message");

        appendContext.WriteStream.Write(messageBodyBytes,
                                        0,
                                        messageBodyBytes.Length);

        Assert.IsTrue(session.IsTransactionProceeding, "IsTransactionProceeding after Write");

        server.EnqueueTaggedResponse("APPEND INBOX ",
                                     "$tag OK APPEND completed\r\n");

        ImapAppendedUidSet appendedUid;

        Assert.IsTrue((bool)appendContext.GetResult(out appendedUid));
        Assert.IsFalse(session.IsTransactionProceeding, "IsTransactionProceeding after GetResult");

        Assert.AreEqual("0002 APPEND INBOX {7}\r\nmessage\r\n",
                        server.DequeueRequest());

        Assert.IsNull(appendedUid);
      });
    }

    [Test]
    public void TestPrepareAppend_AppendContext_GetResult_AlreadyFinished()
    {
      Authenticate2(delegate(ImapSession session, ImapPseudoServer2 server) {
        var appendContext = session.PrepareAppend(null, null, null, "INBOX");

        Assert.IsNotNull(appendContext);
        Assert.IsNotNull(appendContext.WriteStream);
        Assert.IsTrue(session.IsTransactionProceeding, "IsTransactionProceeding after PrepareAppend");

        appendContext.WriteStream.WriteByte(0x40);
        appendContext.WriteStream.Close();

        Assert.IsTrue(session.IsTransactionProceeding, "IsTransactionProceeding after Close");
        Assert.Throws<TimeoutException>(() => WaitForTransactionFinished(session)); // transaction will not finish (wait for response)

        server.EnqueueTaggedResponse("APPEND INBOX ",
                                     "$tag OK APPEND completed\r\n");

        ImapAppendedUidSet appendedUid;

        Assert.IsTrue((bool)appendContext.GetResult(out appendedUid));
        Assert.IsFalse(session.IsTransactionProceeding, "IsTransactionProceeding after GetResult");

        Assert.AreEqual("0002 APPEND INBOX {1}\r\n@\r\n",
                        server.DequeueRequest());

        Assert.IsNull(appendedUid);

        Assert.Throws<InvalidOperationException>(() => appendContext.GetResult(out appendedUid));
      });
    }

    [Test]
    public void TestPrepareAppend_Continuation_No_1()
    {
      Authenticate2(delegate(ImapSession session, ImapPseudoServer2 server) {
        var appendContext = session.PrepareAppend(null, null, null, "INBOX");

        Assert.IsNotNull(appendContext);
        Assert.IsNotNull(appendContext.WriteStream);
        Assert.IsTrue(session.IsTransactionProceeding, "IsTransactionProceeding after PrepareAppend");

        server.EnqueueLiteralContinuationResponse("APPEND INBOX ",
                                                  "$tag NO APPEND failed\r\n",
                                                  false);

        var messageBodyBytes = NetworkTransferEncoding.Transfer7Bit.GetBytes("message");

        appendContext.WriteStream.Write(messageBodyBytes,
                                        0,
                                        messageBodyBytes.Length);

        Assert.IsTrue(session.IsTransactionProceeding, "IsTransactionProceeding after Write");
        Assert.Throws<TimeoutException>(() => WaitForTransactionFinished(session)); // transaction will not finish (wait for stream closed)

        ImapAppendedUidSet appendedUid;

        Assert.IsFalse((bool)appendContext.GetResult(out appendedUid));
        Assert.IsFalse(session.IsTransactionProceeding, "IsTransactionProceeding after GetResult");

        Assert.AreEqual("0002 APPEND INBOX {7}\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestPrepareAppend_Continuation_No_2()
    {
      Authenticate2(delegate(ImapSession session, ImapPseudoServer2 server) {
        var appendContext = session.PrepareAppend(null, null, null, "INBOX");

        Assert.IsNotNull(appendContext);
        Assert.IsNotNull(appendContext.WriteStream);
        Assert.IsTrue(session.IsTransactionProceeding, "IsTransactionProceeding after PrepareAppend");

        server.EnqueueLiteralContinuationResponse("APPEND INBOX ",
                                                  "$tag NO APPEND failed\r\n",
                                                  false);

        appendContext.WriteStream.Close();

        Assert.IsTrue(session.IsTransactionProceeding, "IsTransactionProceeding after Close");
        Assert.DoesNotThrow(() => WaitForTransactionFinished(session));

        ImapAppendedUidSet appendedUid;

        Assert.IsFalse((bool)appendContext.GetResult(out appendedUid));
        Assert.IsFalse(session.IsTransactionProceeding, "IsTransactionProceeding after GetResult");

        Assert.AreEqual("0002 APPEND INBOX {0}\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestAppendMultiple()
    {
      AppendMultiple(false);
    }

    [Test]
    public void TestAppendMultiple_WithAppendUidResponseCode()
    {
      AppendMultiple(true);
    }

    private void AppendMultiple(bool respondAppendUid)
    {
      SelectMailbox2(new[] {"MULTIAPPEND"}, delegate(ImapSession session, ImapPseudoServer2 server) {
        server.EnqueueLiteralContinuationResponse("+ Ready for literal data\r\n", true);
        server.EnqueueLiteralContinuationResponse("+ Ready for literal data\r\n", true);

        if (respondAppendUid)
          server.EnqueueTaggedResponse("APPEND saved-messages ",
                                       "$tag OK [APPENDUID 38505 3955:3956] APPEND completed\r\n");
        else
          server.EnqueueTaggedResponse("APPEND saved-messages ",
                                       "$tag OK APPEND completed\r\n");

        var messageBody1 = @"Date: Mon, 7 Feb 1994 21:52:25 -0800 (PST)
From: Fred Foobar <foobar@Blurdybloop.example.COM>
Subject: afternoon meeting
To: mooch@owatagu.example.net
Message-Id: <B27397-0100000@Blurdybloop.example.COM>
MIME-Version: 1.0
Content-Type: TEXT/PLAIN; CHARSET=US-ASCII

Hello Joe, do you think we can meet at 3:30 tomorrow?
".Replace("\r\n", "\n").Replace("\n", "\r\n");
        var messageBody2 = @"Date: Mon, 7 Feb 1994 22:43:04 -0800 (PST)
From: Joe Mooch <mooch@OWaTaGu.example.net>
Subject: Re: afternoon meeting
To: foobar@blurdybloop.example.com
Message-Id: <a0434793874930@OWaTaGu.example.net>
MIME-Version: 1.0
Content-Type: TEXT/PLAIN; CHARSET=US-ASCII

3:30 is fine with me.
".Replace("\r\n", "\n").Replace("\n", "\r\n");

        var message1 = new ImapAppendMessage(Encoding.ASCII.GetBytes(messageBody1),
                                             new ImapMessageFlagSet(ImapMessageFlag.Seen));
        var message2 = new ImapAppendMessage(Encoding.ASCII.GetBytes(messageBody2),
                                             new DateTimeOffset(new DateTime(1994, 2, 7, 22, 43, 04), new TimeSpan(-8, 0, 0)),
                                             new ImapMessageFlagSet(ImapMessageFlag.Seen));

        ImapAppendedUidSet appended;

        Assert.IsTrue((bool)session.AppendMultiple(new[] {message1, message2}, "saved-messages", out appended));

        var expected =
          string.Format("0004 APPEND saved-messages (\\Seen) {{{0}}}\r\n", Encoding.ASCII.GetByteCount(messageBody1)) +
          messageBody1 +
          string.Format(" (\\Seen) \"07-Feb-1994 22:43:04 -0800\" {{{0}}}\r\n", Encoding.ASCII.GetByteCount(messageBody2)) +
          messageBody2 +
          "\r\n";

        Assert.AreEqual(expected, server.DequeueRequest());

        if (respondAppendUid) {
          Assert.IsNotNull(appended);
          Assert.AreEqual(38505L, appended.UidValidity);

          var appendedUids = appended.ToArray();

          Assert.AreEqual(2, appendedUids.Length);
          Assert.AreEqual(3955L, appendedUids[0]);
          Assert.AreEqual(3956L, appendedUids[1]);
        }
      });
    }

    [Test]
    public void TestAppendMultiple_NonSynchronizedLiterals()
    {
      SelectMailbox2(new[] {"MULTIAPPEND", "LITERAL+"},
                     delegate(ImapSession session, ImapPseudoServer2 server) {
        server.EnqueueTaggedResponse("$tag OK APPEND completed\r\n");

        var message1 = new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message1"));
        var message2 = new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message2"));
        Assert.IsTrue((bool)session.AppendMultiple(new[] {message1, message2}, "saved-messages"));

        Assert.AreEqual("0004 APPEND saved-messages {34+}\r\n" +
                        "MIME-Version: 1.0\r\n" +
                        "\r\n" +
                        "test message1 {34+}\r\n" +
                        "MIME-Version: 1.0\r\n" +
                        "\r\n" +
                        "test message2\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestAppendMultiple_Incapable()
    {
      SelectMailbox2(delegate(ImapSession session, ImapPseudoServer2 server) {
        session.HandlesIncapableAsException = true;

        Assert.Throws<ImapIncapableException>(delegate {
          session.AppendMultiple(new[] {
            new ImapAppendMessage(Encoding.ASCII.GetBytes("message")),
            new ImapAppendMessage(Encoding.ASCII.GetBytes("message2")),
          }, "saved-messages");
        });
      });
    }

    [Test]
    public void TestAppendBinary()
    {
      SelectMailbox2(new[] {"BINARY"}, delegate(ImapSession session, ImapPseudoServer2 server) {
        server.EnqueueTaggedResponse("APPEND INBOX",
                                     "$tag OK APPEND completed\r\n");

        Assert.IsTrue((bool)session.AppendBinary(new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\n\x00\x01\x02\x03\x04\x05\x06\x07")),
                                                 "INBOX"));

        Assert.AreEqual("0004 APPEND INBOX ~{29}\r\n" +
                        "MIME-Version: 1.0\r\n" +
                        "\r\n" +
                        "\x00\x01\x02\x03\x04\x05\x06\x07\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestAppendBinary_NonSynchronizedLiterals()
    {
      SelectMailbox2(new[] {"BINARY", "LITERAL+"}, delegate(ImapSession session, ImapPseudoServer2 server) {
        server.EnqueueTaggedResponse("APPEND INBOX ",
                                     "$tag OK APPEND completed\r\n");

        Assert.IsTrue((bool)session.AppendBinary(new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\n\x00\x01\x02\x03\x04\x05\x06\x07")),
                                                 "INBOX"));

        Assert.AreEqual(("0004 APPEND INBOX ~{29+}\r\n" +
                        "MIME-Version: 1.0\r\n" +
                        "\r\n" +
                        "\x00\x01\x02\x03\x04\x05\x06\x07\r\n").Replace("\x00", @"\0"),
                        server.DequeueRequest().Replace("\x00", @"\0"));
      });
    }

    [Test]
    public void TestAppendBinary_Multiple()
    {
      SelectMailbox2(new[] {"BINARY", "MULTIAPPEND"},
                     delegate(ImapSession session, ImapPseudoServer2 server) {
        server.EnqueueTaggedResponse("APPEND INBOX ",
                                     "$tag OK APPEND completed\r\n");

        var messages = new[] {
          new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\nmessage1\x00\x01\x02\x03\x04\x05\x06\x07")),
          new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\nmessage2\x00\x01\x02\x03\x04\x05\x06\x07")),
        };

        Assert.IsTrue((bool)session.AppendBinaryMultiple(messages,
                                                         "INBOX"));

        Assert.AreEqual(("0004 APPEND INBOX ~{37}\r\n" +
                        "MIME-Version: 1.0\r\n" +
                        "\r\n" +
                        "message1\x00\x01\x02\x03\x04\x05\x06\x07" +
                        " ~{37}\r\n" +
                        "MIME-Version: 1.0\r\n" +
                        "\r\n" +
                        "message2\x00\x01\x02\x03\x04\x05\x06\x07\r\n").Replace("\x00", @"\0"),
                        server.DequeueRequest().Replace("\x00", @"\0"));
      });
    }

    [Test]
    public void TestAppendBinary_Incapable()
    {
      SelectMailbox2(delegate(ImapSession session, ImapPseudoServer2 server) {
        session.HandlesIncapableAsException = true;

        Assert.Throws<ImapIncapableException>(delegate {
          session.AppendBinary(new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\n\x00\x01\x02\x03\x04\x05\x06\x07")),
                               "saved-messages");
        });
      });
    }
  }
}