﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Threading;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapClientEventsTests {
    [Test]
    public void TestExistMessageCountChanged()
    {
      var selectResp =
        "* EXISTS 10\r\n" +
        "* RECENT 3\r\n" +
        "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.AreEqual(10L, mailbox.ExistMessageCount);

        var eventRaisedCount = 0;
        Exception caughtException = null;

        mailbox.Client.ExistMessageCountChanged += delegate(object sender, ImapMailboxSizeChangedEventArgs e) {
          try {
            Assert.IsFalse(mailbox.Client.IsBusy);

            Assert.AreSame(mailbox.Client, sender);

            Assert.IsNotNull(e);
            Assert.AreEqual(mailbox, e.Mailbox);
            Assert.AreEqual(10L, e.PrevCount);
            Assert.AreEqual(12L, e.CurrentCount);

            server.EnqueueTaggedResponse("NOOP\r\n",
                                         "$tag OK done\r\n");

            Assert.DoesNotThrow(() => mailbox.Client.Refresh(), "new operation from event handler");

            eventRaisedCount++;
          }
          catch (Exception ex) {
            caughtException = ex;
          }
        };

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "* EXISTS 12\r\n" +
                                     "$tag OK done\r\n");

        mailbox.Client.Refresh();

        Assert.IsNull(caughtException);

        Assert.AreEqual(1, eventRaisedCount);
        Assert.AreEqual(12L, mailbox.ExistMessageCount);
      });
    }

    [Test]
    public void TestRecentMessageCountChanged()
    {
      var selectResp =
        "* EXISTS 10\r\n" +
        "* RECENT 3\r\n" +
        "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.AreEqual(3L, mailbox.RecentMessageCount);

        var eventRaisedCount = 0;
        Exception caughtException = null;

        mailbox.Client.RecentMessageCountChanged += delegate(object sender, ImapMailboxSizeChangedEventArgs e) {
          try {
            Assert.IsFalse(mailbox.Client.IsBusy);

            Assert.AreSame(mailbox.Client, sender);

            Assert.IsNotNull(e);
            Assert.AreEqual(mailbox, e.Mailbox);
            Assert.AreEqual(3L, e.PrevCount);
            Assert.AreEqual(4L, e.CurrentCount);

            server.EnqueueTaggedResponse("NOOP\r\n",
                                         "$tag OK done\r\n");

            Assert.DoesNotThrow(() => mailbox.Client.Refresh(), "new operation from event handler");

            eventRaisedCount++;
          }
          catch (Exception ex) {
            caughtException = ex;
          }
        };

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "* RECENT 4\r\n" +
                                     "$tag OK done\r\n");

        mailbox.Client.Refresh();

        Assert.IsNull(caughtException);

        Assert.AreEqual(1, eventRaisedCount);
        Assert.AreEqual(4L, mailbox.RecentMessageCount);
      });
    }

    [Test]
    public void TestMessageStatusChanged()
    {
      var selectResp =
        "* EXISTS 3\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.AreEqual(3L, mailbox.ExistMessageCount);

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("FETCH 1:3 (UID FLAGS)\r\n",
                                     "* FETCH 1 (UID 1 FLAGS (\\Seen))\r\n" +
                                     "* FETCH 2 (UID 2 FLAGS (\\Seen))\r\n" +
                                     "* FETCH 3 (UID 3 FLAGS (\\Seen))\r\n" +
                                     "$tag OK done\r\n");

        var messages = mailbox.GetMessages(ImapMessageFetchAttributeOptions.DynamicAttributes).ToArray();

        Assert.IsTrue (messages[0].IsSeen);
        Assert.IsFalse(messages[0].IsMarkedAsDeleted);
        Assert.IsTrue (messages[1].IsSeen);
        Assert.IsFalse(messages[1].IsMarkedAsDeleted);
        Assert.IsTrue (messages[2].IsSeen);
        Assert.IsFalse(messages[2].IsMarkedAsDeleted);

        var eventRaisedCount = 0;
        Exception caughtException = null;

        mailbox.Client.MessageStatusChanged += delegate(object sender, ImapMessageStatusChangedEventArgs e) {
          try {
            Assert.IsFalse(mailbox.Client.IsBusy);

            Assert.AreSame(mailbox.Client, sender);

            Assert.IsNotNull(e);
            Assert.IsNotNull(e.Messages);
            Assert.AreEqual(2, e.Messages.Length);

            Assert.AreEqual(1L, e.Messages[0].Sequence);
            Assert.IsTrue(e.Messages[0].IsSeen);
            Assert.IsTrue(e.Messages[0].IsMarkedAsDeleted);

            Assert.AreEqual(3L, e.Messages[1].Sequence);
            Assert.IsFalse(e.Messages[1].IsSeen);
            Assert.IsFalse(e.Messages[1].IsMarkedAsDeleted);

            server.EnqueueTaggedResponse("NOOP\r\n",
                                         "$tag OK done\r\n");

            Assert.DoesNotThrow(() => mailbox.Client.Refresh(), "new operation from event handler");

            eventRaisedCount++;
          }
          catch (Exception ex) {
            caughtException = ex;
          }
        };

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "* FETCH 1 (FLAGS (\\Seen \\Deleted))\r\n" +
                                     "* FETCH 3 (FLAGS ())\r\n" +
                                     "$tag OK done\r\n");

        mailbox.Client.Refresh();

        Assert.IsNull(caughtException);

        Assert.AreEqual(1, eventRaisedCount);

        Assert.IsTrue (messages[0].IsSeen);
        Assert.IsTrue (messages[0].IsMarkedAsDeleted);
        Assert.IsTrue (messages[1].IsSeen);
        Assert.IsFalse(messages[1].IsMarkedAsDeleted);
        Assert.IsFalse(messages[2].IsSeen);
        Assert.IsFalse(messages[2].IsMarkedAsDeleted);
      });
    }

    [Test]
    public void TestMessageStatusChanged_NotRaisedByFetchBodyResponse()
    {
      MessageStatusChanged_RaisedByFetchBodyResponse(false);
    }

    [Test]
    public void TestMessageStatusChanged_RaisedByFetchBodyResponseWithFlags()
    {
      MessageStatusChanged_RaisedByFetchBodyResponse(true);
    }

    private void MessageStatusChanged_RaisedByFetchBodyResponse(bool setSeen)
    {
      var selectResp =
        "* EXISTS 1\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      var messageBody = @"Return-Path: <test@example.net>
X-Original-To: test@example.net
Delivered-To: test@example.net
Message-ID: <4BAA561E.5000004@test@example.net>
Date: Thu, 25 Mar 2010 03:12:46 +0900
From: test <test@example.net>
Reply-To: test@example.net
User-Agent: Thunderbird 2.0.0.24 (X11/20100317)
MIME-Version: 1.0
To: hoge@example.net
Subject: =?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=
Content-Type: text/plain; charset=ISO-2022-JP
Content-Transfer-Encoding: 7bit

test message

3 uid fetch 6 (body[])
* 6 FETCH (UID 6 BODY[] {1476}
hogehogehoge
)
3 OK Fetch completed.
".Replace("\r\n", "\n").Replace("\n", "\r\n");

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.AreEqual(1L, mailbox.ExistMessageCount);

        server.EnqueueTaggedResponse("FETCH 1 (UID FLAGS)\r\n",
                                     "* FETCH 1 (UID 1 FLAGS ())\r\n" +
                                     "$tag OK done\r\n");

        var message = mailbox.GetMessageBySequence(1L, ImapMessageFetchAttributeOptions.DynamicAttributes);

        Assert.IsFalse(message.IsSeen);

        var eventRaisedCount = 0;
        ImapMessageInfo[] statusChangedMessages = null;

        mailbox.Client.MessageStatusChanged += delegate(object sender, ImapMessageStatusChangedEventArgs e) {
          statusChangedMessages = e.Messages;

          eventRaisedCount++;
        };

        var octets = NetworkTransferEncoding.Transfer8Bit.GetByteCount(messageBody);

        server.EnqueueTaggedResponse(string.Format("UID FETCH 1 (RFC822.SIZE BODY{0}[]<0.10240>)\r\n",
                                                   setSeen ? string.Empty : ".PEEK"),
                                     string.Format("* {0} FETCH (RFC822.SIZE {1} {3}BODY[] {{{1}}}\r\n{2})\r\n",
                                                   message.Uid,
                                                   octets,
                                                   messageBody,
                                                   setSeen ? "FLAGS (\\Seen) " : string.Empty) +
                                     "$tag OK done\r\n");

        if (setSeen) {
          message.ReadAllText(ImapMessageFetchBodyOptions.SetSeen);

          Assert.AreEqual(1, eventRaisedCount);
          Assert.AreEqual(1, statusChangedMessages.Length);
          Assert.AreSame(message, statusChangedMessages[0]);
          Assert.IsTrue(message.IsSeen);
        }
        else {
          message.ReadAllText(ImapMessageFetchBodyOptions.Default);

          Assert.AreEqual(0, eventRaisedCount);
          Assert.IsNull(statusChangedMessages);
        }
      });
    }

    [Test]
    public void TestMessageStatusChanged_NotRaisedByUnmanagedMessage()
    {
      var selectResp =
        "* EXISTS 3\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.AreEqual(3L, mailbox.ExistMessageCount);

        var eventRaisedCount = 0;

        mailbox.Client.MessageStatusChanged += delegate(object sender, ImapMessageStatusChangedEventArgs e) {
          eventRaisedCount++;
        };

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "* FETCH 1 (FLAGS (\\Seen \\Deleted))\r\n" +
                                     "* FETCH 3 (FLAGS ())\r\n" +
                                     "$tag OK done\r\n");

        mailbox.Client.Refresh();

        Assert.AreEqual(0, eventRaisedCount);
      });
    }

    [Test]
    public void TestMessageDeleted()
    {
      var selectResp =
        "* EXISTS 3\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.AreEqual(3L, mailbox.ExistMessageCount);

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("FETCH 1:3 (UID FLAGS)\r\n",
                                     "* FETCH 1 (UID 1 FLAGS (\\Seen))\r\n" +
                                     "* FETCH 2 (UID 2 FLAGS (\\Seen))\r\n" +
                                     "* FETCH 3 (UID 3 FLAGS (\\Seen))\r\n" +
                                     "$tag OK done\r\n");

        var messages = mailbox.GetMessages(ImapMessageFetchAttributeOptions.DynamicAttributes).ToArray();

        Assert.AreEqual(1L, messages[0].Sequence);
        Assert.AreEqual(1L, messages[0].Uid);
        Assert.IsFalse(messages[0].IsDeleted);
        Assert.AreEqual(2L, messages[1].Sequence);
        Assert.AreEqual(2L, messages[1].Uid);
        Assert.IsFalse(messages[0].IsDeleted);
        Assert.AreEqual(3L, messages[2].Sequence);
        Assert.AreEqual(3L, messages[2].Uid);
        Assert.IsFalse(messages[0].IsDeleted);

        var eventRaisedCount = 0;
        Exception caughtException = null;

        mailbox.Client.MessageDeleted += delegate(object sender, ImapMessageStatusChangedEventArgs e) {
          try {
            Assert.IsFalse(mailbox.Client.IsBusy);

            Assert.AreSame(mailbox.Client, sender);

            Assert.IsNotNull(e);
            Assert.IsNotNull(e.Messages);
            Assert.AreEqual(2, e.Messages.Length);

            Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, e.Messages[0].Sequence);
            Assert.AreEqual(3L, e.Messages[0].Uid);
            Assert.IsTrue(e.Messages[0].IsDeleted);

            Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, e.Messages[1].Sequence);
            Assert.AreEqual(1L, e.Messages[1].Uid);
            Assert.IsTrue(e.Messages[1].IsDeleted);

            server.EnqueueTaggedResponse("NOOP\r\n",
                                         "$tag OK done\r\n");

            Assert.DoesNotThrow(() => mailbox.Client.Refresh(), "new operation from event handler");

            eventRaisedCount++;
          }
          catch (Exception ex) {
            caughtException = ex;
          }
        };

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "* EXPUNGE 3\r\n" +
                                     "* EXPUNGE 1\r\n" +
                                     "$tag OK done\r\n");

        mailbox.Client.Refresh();

        Assert.IsNull(caughtException);

        Assert.AreEqual(1, eventRaisedCount);

        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[0].Sequence);
        Assert.AreEqual(1L, messages[0].Uid);
        Assert.IsTrue(messages[0].IsDeleted);
        Assert.AreEqual(1L, messages[1].Sequence);
        Assert.AreEqual(2L, messages[1].Uid);
        Assert.IsFalse(messages[1].IsDeleted);
        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[2].Sequence);
        Assert.AreEqual(3L, messages[2].Uid);
        Assert.IsTrue(messages[2].IsDeleted);
      });
    }

    [Test]
    public void TestAlertReceived_WithUntaggedOkResponse()
    {
      EventAlertReceived("* OK [ALERT] operation successed\r\n" +
                         "$tag OK done\r\n",
                         "operation successed",
                         ImapResponseCondition.Ok);
    }

    [Test]
    public void TestAlertReceived_WithUntaggedNoResponse()
    {
      EventAlertReceived("* NO [ALERT] operation failed\r\n" +
                         "$tag OK done\r\n",
                         "operation failed",
                         ImapResponseCondition.No);
    }

    [Test]
    public void TestAlertReceived_WithTaggedOkResponse()
    {
      EventAlertReceived("$tag OK [ALERT] done\r\n",
                         "done",
                         ImapResponseCondition.Ok);
    }

    private void EventAlertReceived(string noopResponse,
                                    string expectedAlert,
                                    ImapResponseCondition expectedCondition)
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var eventRaisedCount = 0;
        Exception caughtException = null;

        client.AlertReceived += delegate(object sender, ImapAlertReceivedEventArgs e) {
          try {
            Assert.IsFalse(client.IsBusy);

            Assert.AreSame(client, sender);

            Assert.IsNotNull(e);
            Assert.AreEqual(expectedAlert, e.Alert);
            Assert.AreEqual(expectedCondition, e.Condition);

            server.EnqueueTaggedResponse("NOOP\r\n",
                                         "$tag OK done\r\n");

            Assert.DoesNotThrow(() => client.Refresh(), "new operation from event handler");

            eventRaisedCount++;
          }
          catch (Exception ex) {
            caughtException = ex;
          }
        };

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     noopResponse);

        client.Refresh();

        Assert.IsNull(caughtException);

        Assert.AreEqual(1, eventRaisedCount);
      });
    }

    [Test]
    public void TestEvents_WhileIdling()
    {
      var selectResp =
        "* EXISTS 3\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.AreEqual(3L, mailbox.ExistMessageCount);

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("FETCH 1:3 (UID FLAGS)\r\n",
                                     "* FETCH 1 (UID 1 FLAGS (\\Seen))\r\n" +
                                     "* FETCH 2 (UID 2 FLAGS (\\Seen))\r\n" +
                                     "* FETCH 3 (UID 3 FLAGS (\\Seen))\r\n" +
                                     "$tag OK done\r\n");

        var messages = mailbox.GetMessages(ImapMessageFetchAttributeOptions.DynamicAttributes).ToArray();

        var eventMessageDeletedRaisedCount = 0;
        var eventRecentMessageCountChangedRaisedCount = 0;
        Exception caughtExceptionInMessageDeleted = null;
        Exception caughtExceptionInRecentMessageCountChanged = null;

        mailbox.Client.MessageDeleted += delegate(object sender, ImapMessageStatusChangedEventArgs e) {
          try {
            Assert.IsTrue(mailbox.Client.IsBusy, "client.IsBusy");
            Assert.IsTrue(mailbox.Client.IsIdling, "client.IsIdling");
            Assert.IsTrue(mailbox.IsIdling, "mailbox.IsIdling");

            Assert.AreSame(mailbox.Client, sender);

            Assert.IsNotNull(e);

            Assert.IsNotNull(e.Messages);
            Assert.AreEqual(1, e.Messages.Length);
            Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, e.Messages[0].Sequence);
            Assert.AreEqual(2L, e.Messages[0].Uid);
            Assert.IsTrue(e.Messages[0].IsDeleted);

            Assert.AreEqual(1L, messages[0].Sequence);
            Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[1].Sequence);
            Assert.AreEqual(2L, messages[2].Sequence);

            Assert.Throws<InvalidOperationException>(() => mailbox.Refresh());

            eventMessageDeletedRaisedCount++;
          }
          catch (Exception ex) {
            caughtExceptionInMessageDeleted = ex;
          }
        };

        mailbox.Client.RecentMessageCountChanged += delegate(object sender, ImapMailboxSizeChangedEventArgs e) {
          try {
            Assert.IsTrue(mailbox.Client.IsBusy, "client.IsBusy");
            Assert.IsTrue(mailbox.Client.IsIdling, "client.IsIdling");
            Assert.IsTrue(mailbox.IsIdling, "mailbox.IsIdling");

            Assert.AreSame(mailbox.Client, sender);

            Assert.IsNotNull(e);
            Assert.AreEqual(mailbox, e.Mailbox);
            Assert.AreEqual(0L, e.PrevCount);
            Assert.AreEqual(1L, e.CurrentCount);

            Assert.Throws<InvalidOperationException>(() => mailbox.Refresh());

            eventRecentMessageCountChangedRaisedCount++;
          }
          catch (Exception ex) {
            caughtExceptionInRecentMessageCountChanged = ex;
          }
        };

        var delay = TimeSpan.FromMilliseconds(100);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n");

          Thread.Sleep(delay);

          server.EnqueuePushResponse("* EXISTS 4\r\n");
          server.EnqueuePushResponse("* RECENT 1\r\n");
          server.EnqueuePushResponse("* EXPUNGE 2\r\n");

          server.WaitForClientRequest(); // DONE

          server.EnqueueTaggedResponse("$tag OK done\r\n");
        });

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

        var timeout = TimeSpan.FromSeconds(1.0);

        Smdn.Net.TestUtils.Elapses(timeout, delegate {
          mailbox.Idle(timeout);
        });

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");

        Assert.IsNull(caughtExceptionInMessageDeleted);
        Assert.IsNull(caughtExceptionInRecentMessageCountChanged);

        Assert.AreEqual(1, eventMessageDeletedRaisedCount, "MessageDeleted raised count");
        Assert.AreEqual(1, eventRecentMessageCountChangedRaisedCount, "RecentMessageCountChanged raised count");
      });
    }

    [Test]
    public void TestEvents_IgnoreExceptionsFromEventHandler()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var eventRaisedCount = 0;

        client.AlertReceived += delegate(object sender, ImapAlertReceivedEventArgs e) {
          eventRaisedCount++;

          throw new ApplicationException("this exception will be ignored");
        };

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK [ALERT] done\r\n");

        Assert.DoesNotThrow(() => client.Refresh());

        Assert.AreEqual(1, eventRaisedCount);
      });
    }

    [Test]
    public void TestEvents_RaisedInEachOperation()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var receivedAlerts = new List<string>();

        client.AlertReceived += delegate(object sender, ImapAlertReceivedEventArgs e) {
          receivedAlerts.Add(e.Alert);
        };

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK [ALERT] 1st\r\n");

        Assert.DoesNotThrow(() => client.Refresh());

        Assert.AreEqual(1, receivedAlerts.Count);
        Assert.AreEqual("1st", receivedAlerts[0]);

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK [ALERT] 2nd\r\n");

        Assert.DoesNotThrow(() => client.Refresh());

        Assert.AreEqual(2, receivedAlerts.Count);
        Assert.AreEqual("2nd", receivedAlerts[1]);
      });
    }

    [Test]
    public void TestEvents_NoEventsRaised()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var eventRaisedCount = 0;

        client.AlertReceived += delegate(object sender, ImapAlertReceivedEventArgs e) {
          eventRaisedCount++;
        };

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");

        Assert.DoesNotThrow(() => client.Refresh());

        Assert.AreEqual(0, eventRaisedCount);
      });
    }

    private class PseudoSynchronizingObject : ISynchronizeInvoke {
      public bool InvokeRequired {
        get; private set;
      }

      private bool expectSynchronousCall;

      public PseudoSynchronizingObject(bool invokeRequired, bool expectSynchronousCall)
      {
        this.InvokeRequired = invokeRequired;
        this.expectSynchronousCall = expectSynchronousCall;
      }

      public IAsyncResult BeginInvoke(Delegate method, object[] args)
      {
        if (expectSynchronousCall)
          throw new InvalidOperationException("synchronous call expected");

        ThreadPool.QueueUserWorkItem(delegate {
          method.DynamicInvoke(args);
        });

        return null;
      }

      public object EndInvoke(IAsyncResult result)
      {
        throw new NotImplementedException();
      }

      public object Invoke(Delegate method, object[] args)
      {
        if (!expectSynchronousCall)
          throw new InvalidOperationException("asynchronous call expected");

        return method.DynamicInvoke(args);
      }
    }

    [Test]
    public void TestEvents_RaiseEventWithSynchronizingObject_Null()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        const bool expectSynchronousCall = true;

        client.SynchronizingObject = null;

        HandleEventWithSynchronizingObject(server, client, expectSynchronousCall);
      });
    }

    [Test]
    public void TestEvents_RaiseEventWithSynchronizingObject_InvokeRequired()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        const bool expectSynchronousCall = false;

        client.SynchronizingObject = new PseudoSynchronizingObject(true, expectSynchronousCall);

        HandleEventWithSynchronizingObject(server, client, expectSynchronousCall);
      });
    }

    [Test]
    public void TestEvents_RaiseEventWithSynchronizingObject_InvokeNotRequired()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        const bool expectSynchronousCall = true;

        client.SynchronizingObject = new PseudoSynchronizingObject(false, expectSynchronousCall);

        HandleEventWithSynchronizingObject(server, client, expectSynchronousCall);
      });
    }

    private void HandleEventWithSynchronizingObject(ImapPseudoServer2 server, ImapClient client, bool expectSynchronousCall)
    {
      using (var waitHandle = new ManualResetEvent(false)) {
        var eventRaisedCount = 0;
        var eventHandlerThreadID = -1;
        Exception caughtException = null;

        client.AlertReceived += delegate(object sender, ImapAlertReceivedEventArgs e) {
          eventHandlerThreadID = Thread.CurrentThread.ManagedThreadId;

          try {
            try {
              Assert.AreSame(client, sender);

              Assert.IsNotNull(e);
              Assert.AreEqual("done",e.Alert);

              eventRaisedCount++;
            }
            catch (Exception ex) {
              caughtException = ex;
            }
          }
          finally {
            waitHandle.Set();
          }
        };

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK [ALERT] done\r\n");

        Assert.DoesNotThrow(() => client.Refresh());

        if (!waitHandle.WaitOne(1000)) // wait for event handler finished
          Assert.Fail("event handler not called or not finished");

        Assert.IsNull(caughtException);

        Assert.AreEqual(1, eventRaisedCount);

        if (expectSynchronousCall)
          Assert.AreEqual(Thread.CurrentThread.ManagedThreadId,
                          eventHandlerThreadID);
        else
          Assert.AreNotEqual(Thread.CurrentThread.ManagedThreadId,
                             eventHandlerThreadID);
      }
    }
  }
}

