using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Net;
using System.Threading;
using NUnit.Framework;

using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Protocol;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapClientOperationsTests {
    [Test]
    public void TestGetInbox()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                       "* LIST () \"\" INBOX\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailbox = client.GetInbox();

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.AreEqual("INBOX", mailbox.Name);
        Assert.AreEqual("INBOX", mailbox.FullName);
        Assert.IsTrue(mailbox.IsInbox);
      });
    }

    [Test]
    public void TestGetMailbox()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" Draft\r\n",
                                       "* LIST () \"\" Draft\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailbox = client.GetMailbox("Draft");

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.AreEqual("Draft", mailbox.Name);
        Assert.AreEqual("Draft", mailbox.FullName);
        Assert.IsFalse(mailbox.IsInbox);
      });
    }

    [Test]
    public void TestGetMailbox_NotFound()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST ",
                                       "$tag OK done\r\n");
        });

        try {
          client.GetMailbox("NonExistent");
          Assert.Fail("ImapMailboxNotFoundException not thrown");
        }
        catch (ImapMailboxNotFoundException ex) {
          Assert.IsNotNull(ex.Mailbox);
          Assert.AreEqual("NonExistent", ex.Mailbox);

          Smdn.Net.TestUtils.SerializeBinary(ex, delegate(ImapMailboxNotFoundException deserialized) {
            Assert.IsNotNull(deserialized.Mailbox);
            Assert.AreEqual(ex.Mailbox, deserialized.Mailbox);
          });
        }

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestGetMailbox_NameContainsWildcard()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" Box*\r\n",
                                       "* LIST () \"/\" Box*/Child1\r\n" +
                                       "* LIST () \"/\" Box*/Child2\r\n" +
                                       "* LIST () \"/\" Box*\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailbox = client.GetMailbox("Box*");

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.AreEqual("Box*", mailbox.Name);
        Assert.AreEqual("Box*", mailbox.FullName);
        Assert.IsFalse(mailbox.IsInbox);
      });
    }

    [Test]
    public void TestGetMailbox_NameContainsWildcard_RequestStatus()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" Box*\r\n",
                                       "* LIST () \"/\" Box*/Child1\r\n" +
                                       "* LIST () \"/\" Box*/Child2\r\n" +
                                       "* LIST () \"/\" Box*\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("STATUS \"Box*\" (MESSAGES RECENT UIDNEXT UIDVALIDITY UNSEEN)\r\n",
                                       "* STATUS \"Box*\" (MESSAGES 3 RECENT 1 UIDNEXT 4 UIDVALIDITY 1 UNSEEN 1)\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailbox = client.GetMailbox("Box*", ImapMailboxListOptions.RequestStatus);

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.AreEqual("Box*", mailbox.Name);
        Assert.AreEqual("Box*", mailbox.FullName);
        Assert.AreEqual(3L, mailbox.ExistMessageCount, "ExistMessageCount");
        Assert.AreEqual(4L, mailbox.NextUid, "NextUid");
        Assert.IsFalse(mailbox.IsInbox);
      });
    }

    [Test]
    public void TestGetMailbox_NameNull()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        Assert.Throws<ArgumentNullException>(() => client.GetMailbox((string)null));
      });
    }

    [Test]
    public void TestGetMailbox_NameEmpty()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        Assert.Throws<ArgumentException>(() => client.GetMailbox(string.Empty));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestGetMailbox_SpecialMailbox_SpecialUse()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.SpecialUse},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        GetMailboxSpecialMailboxSpecialUse(server, client, false);

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestGetMailbox_SpecialMailbox_SpecialUse_RequestStatus()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.SpecialUse},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        GetMailboxSpecialMailboxSpecialUse(server, client, true);
      });
    }

    [Test]
    public void TestGetMailbox_SpecialMailbox_SpecialUse_ListExtendedDisabled()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.SpecialUse, ImapCapability.ListExtended},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        client.DisabledCapabilities.Add(ImapCapability.ListExtended);

        GetMailboxSpecialMailboxSpecialUse(server, client, false);
      });
    }

    private void GetMailboxSpecialMailboxSpecialUse(ImapPseudoServer2 server, ImapClient client, bool requestStatus)
    {
      var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
        server.EnqueueTaggedResponse("LIST \"\" *\r\n",
                                     "* LIST (\\Marked \\HasNoChildren) \"/\" Inbox\r\n" +
                                     "* LIST (\\HasNoChildren) \"/\" ToDo\r\n" +
                                     "* LIST (\\HasChildren) \"/\" Projects\r\n" +
                                     "* LIST (\\Sent \\HasNoChildren) \"/\" SentMail\r\n" +
                                     "* LIST (\\Marked \\Drafts \\HasNoChildren) \"/\" MyDrafts\r\n" +
                                     "* LIST (\\Trash \\HasNoChildren) \"/\" Trash\r\n" +
                                     "$tag OK done\r\n");

        if (requestStatus)
          server.EnqueueTaggedResponse("STATUS SentMail (MESSAGES RECENT UIDNEXT UIDVALIDITY UNSEEN)\r\n",
                                       "* STATUS SentMail (MESSAGES 3 RECENT 1 UIDNEXT 4 UIDVALIDITY 1 UNSEEN 1)\r\n" +
                                       "$tag OK done\r\n");
      });

      var mailbox = requestStatus
        ? client.GetMailbox(ImapSpecialMailbox.Sent, ImapMailboxListOptions.RequestStatus)
        : client.GetMailbox(ImapSpecialMailbox.Sent);

      responseTask.Wait();

      Assert.IsFalse(client.IsBusy);

      Assert.IsNotNull(mailbox);
      Assert.AreSame(mailbox.Client, client);
      Assert.AreEqual("SentMail", mailbox.Name);
      Assert.AreEqual("SentMail", mailbox.FullName);
      CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.Sent);

      if (requestStatus) {
        Assert.AreEqual(3L, mailbox.ExistMessageCount, "ExistMessageCount");
        Assert.AreEqual(4L, mailbox.NextUid, "NextUid");
      }
    }

    [Test]
    public void TestGetMailbox_SpecialMailbox_SpecialUse_SubscribedOnly()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.SpecialUse},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          // LSUB
          server.EnqueueTaggedResponse("LSUB \"\" *\r\n",
                                       "* LSUB (\\Marked \\HasNoChildren) \"/\" Inbox\r\n" +
                                       "* LSUB (\\Sent \\HasNoChildren) \"/\" SentMail\r\n" +
                                       "* LSUB (\\Trash \\HasNoChildren) \"/\" Trash\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailbox = client.GetMailbox(ImapSpecialMailbox.Trash,
                                        ImapMailboxListOptions.SubscribedOnly);

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.AreEqual("Trash", mailbox.Name);
        Assert.AreEqual("Trash", mailbox.FullName);
        CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.Trash);
      });
    }

    [Test]
    public void TestGetMailbox_SpecialMailbox_SpecialUse_Remote()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.MailboxReferrals, ImapCapability.SpecialUse},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("RLIST \"\" *\r\n",
                                       "* LIST (\\Marked \\HasNoChildren) \"/\" Inbox\r\n" +
                                       "* LIST (\\HasNoChildren) \"/\" ToDo\r\n" +
                                       "* LIST (\\HasChildren) \"/\" Projects\r\n" +
                                       "* LIST (\\Sent \\HasNoChildren) \"/\" SentMail\r\n" +
                                       "* LIST (\\Marked \\Drafts \\HasNoChildren) \"/\" MyDrafts\r\n" +
                                       "* LIST (\\Trash \\HasNoChildren) \"/\" Trash\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailbox = client.GetMailbox(ImapSpecialMailbox.Drafts,
                                        ImapMailboxListOptions.Remote);

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.AreEqual("MyDrafts", mailbox.Name);
        Assert.AreEqual("MyDrafts", mailbox.FullName);
        CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.Drafts);
      });
    }

    [Test]
    public void TestGetMailbox_SpecialMailbox_SpecialUse_NotFound()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.SpecialUse},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST ",
                                       "* LIST (\\Marked \\HasNoChildren) \"/\" Inbox\r\n" +
                                       "* LIST (\\HasNoChildren) \"/\" ToDo\r\n" +
                                       "* LIST (\\HasChildren) \"/\" Projects\r\n" +
                                       "* LIST (\\Sent \\HasNoChildren) \"/\" SentMail\r\n" +
                                       "* LIST (\\Marked \\Drafts \\HasNoChildren) \"/\" MyDrafts\r\n" +
                                       "* LIST (\\Trash \\HasNoChildren) \"/\" Trash\r\n" +
                                       "$tag OK done\r\n");
        });

        var ex = Assert.Throws<ImapMailboxNotFoundException>(() => client.GetMailbox(ImapSpecialMailbox.All));

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.AreEqual(ex.Mailbox, @"\All");
      });
    }

    [Test]
    public void TestGetMailbox_SpecialMailbox_ListExtended()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.ListExtended, ImapCapability.SpecialUse},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        GetMailboxSpecialMailboxListExtended(server, client, false);
      });
    }

    [Test]
    public void TestGetMailbox_SpecialMailbox_ListExtended_RequestStatus()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.ListExtended, ImapCapability.SpecialUse},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        GetMailboxSpecialMailboxListExtended(server, client, true);
      });
    }

    [Test]
    public void TestGetMailbox_SpecialMailbox_ListExtended_RequestStatus_ListStatusDisabled()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.ListExtended, ImapCapability.SpecialUse, ImapCapability.ListStatus},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        client.DisabledCapabilities.Add(ImapCapability.ListStatus);

        GetMailboxSpecialMailboxListExtended(server, client, true);
      });
    }

    private void GetMailboxSpecialMailboxListExtended(ImapPseudoServer2 server, ImapClient client, bool requestStatus)
    {
      var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
        server.EnqueueTaggedResponse("LIST (SPECIAL-USE) \"\" * RETURN (CHILDREN)\r\n",
                                     "* LIST (\\Sent) \"/\" SentMail\r\n" +
                                     "* LIST (\\Marked \\Drafts) \"/\" MyDrafts\r\n" +
                                     "* LIST (\\Trash) \"/\" Trash\r\n" +
                                     "$tag OK done\r\n");
        if (requestStatus)
          server.EnqueueTaggedResponse("STATUS Trash (MESSAGES RECENT UIDNEXT UIDVALIDITY UNSEEN)\r\n",
                                       "* STATUS Trash (MESSAGES 3 RECENT 1 UIDNEXT 4 UIDVALIDITY 1 UNSEEN 1)\r\n" +
                                       "$tag OK done\r\n");
      });

      var mailbox = requestStatus
        ? client.GetMailbox(ImapSpecialMailbox.Trash, ImapMailboxListOptions.RequestStatus)
        : client.GetMailbox(ImapSpecialMailbox.Trash);

      responseTask.Wait();

      Assert.IsFalse(client.IsBusy);

      Assert.IsNotNull(mailbox);
      Assert.AreSame(mailbox.Client, client);
      Assert.AreEqual("Trash", mailbox.Name);
      Assert.AreEqual("Trash", mailbox.FullName);
      CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.Trash);

      if (requestStatus) {
        Assert.AreEqual(3L, mailbox.ExistMessageCount, "ExistMessageCount");
        Assert.AreEqual(4L, mailbox.NextUid, "NextUid");
      }
    }

    [Test]
    public void TestGetMailbox_SpecialMailbox_ListExtended_RequestStatus_ListStatusCapable()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.ListExtended, ImapCapability.SpecialUse, ImapCapability.ListStatus},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST (SPECIAL-USE) \"\" * RETURN (CHILDREN STATUS (MESSAGES RECENT UIDNEXT UIDVALIDITY UNSEEN))\r\n",
                                       "* LIST (\\Sent) \"/\" SentMail\r\n" +
                                       "* STATUS SentMail ()\r\n" +
                                       "* LIST (\\Marked \\Drafts) \"/\" MyDrafts\r\n" +
                                       "* STATUS MyDrafts (MESSAGES 3 RECENT 1 UIDNEXT 4 UIDVALIDITY 1 UNSEEN 1 HIGHESTMODSEQ 1)\r\n" +
                                       "* LIST (\\Trash) \"/\" Trash\r\n" +
                                       "* STATUS Trash ()\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailbox = client.GetMailbox(ImapSpecialMailbox.Drafts, ImapMailboxListOptions.RequestStatus);

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.AreEqual("MyDrafts", mailbox.Name);
        Assert.AreEqual("MyDrafts", mailbox.FullName);
        CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.Drafts);

        Assert.AreEqual(3L, mailbox.ExistMessageCount, "ExistMessageCount");
        Assert.AreEqual(4L, mailbox.NextUid, "NextUid");
      });
    }

    [Test]
    public void TestGetMailbox_SpecialMailbox_ListExtended_SubscribedOnly()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.ListExtended, ImapCapability.SpecialUse},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST (SPECIAL-USE SUBSCRIBED) \"\" * RETURN (CHILDREN)\r\n",
                                       "* LIST (\\Subscribed \\Sent) \"/\" SentMail\r\n" +
                                       "* LIST (\\Subscribed \\Marked \\Drafts) \"/\" MyDrafts\r\n" +
                                       "* LIST (\\Subscribed \\Trash) \"/\" Trash\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailbox = client.GetMailbox(ImapSpecialMailbox.Sent,
                                        ImapMailboxListOptions.SubscribedOnly);

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.AreEqual("SentMail", mailbox.Name);
        Assert.AreEqual("SentMail", mailbox.FullName);
        CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.Sent);
      });
    }

    [Test]
    public void TestGetMailbox_SpecialMailbox_ListExtended_Remote()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.ListExtended, ImapCapability.SpecialUse},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST (SPECIAL-USE REMOTE) \"\" * RETURN (CHILDREN)\r\n",
                                       "* LIST (\\Sent) \"/\" SentMail\r\n" +
                                       "* LIST (\\Marked \\Drafts) \"/\" MyDrafts\r\n" +
                                       "* LIST (\\Trash) \"/\" Trash\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailbox = client.GetMailbox(ImapSpecialMailbox.Sent,
                                        ImapMailboxListOptions.Remote);

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.AreEqual("SentMail", mailbox.Name);
        Assert.AreEqual("SentMail", mailbox.FullName);
        CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.Sent);
      });
    }

    [Test]
    public void TestGetMailbox_SpecialMailbox_ListExtended_NotFound()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.ListExtended, ImapCapability.SpecialUse},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST ",
                                       "* LIST (\\Sent) \"/\" SentMail\r\n" +
                                       "* LIST (\\Marked \\Drafts) \"/\" MyDrafts\r\n" +
                                       "* LIST (\\Trash) \"/\" Trash\r\n" +
                                       "$tag OK done\r\n");
        });

        var ex = Assert.Throws<ImapMailboxNotFoundException>(() => client.GetMailbox(ImapSpecialMailbox.Junk));

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.AreEqual(ex.Mailbox, @"\Junk");
      });
    }

    [Test]
    public void TestGetMailbox_SpecialMailbox_SpecialUseIncapable()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        Assert.Throws<ImapIncapableException>(() => client.GetMailbox(ImapSpecialMailbox.All));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestGetMailbox_SpecialMailbox_SpecialUseDisabled()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.SpecialUse},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        client.DisabledCapabilities.Add(ImapCapability.SpecialUse);

        Assert.Throws<ImapIncapableException>(() => client.GetMailbox(ImapSpecialMailbox.All));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestGetMailbox_SpecialMailbox_InvalidSpecailMailbox()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.ListExtended, ImapCapability.SpecialUse},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        var values = Enum.GetValues(typeof(ImapSpecialMailbox)).Cast<int>().ToArray();

        Array.Sort(values);

        var invalid = (ImapSpecialMailbox)(values[values.Length - 1] + 1);

        Assert.Throws<ArgumentException>(() => client.GetMailbox(invalid));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestGetOrCreateMailbox_Existent()
    {
      GetOrCreateMailboxExistent(false);
    }

    [Test]
    public void TestGetOrCreateMailbox_Existent_Subscribe()
    {
      GetOrCreateMailboxExistent(true);
    }

    private void GetOrCreateMailboxExistent(bool subscribe)
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" Trash\r\n",
                                       "* LIST () \"\" Trash\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailbox = subscribe
          ? client.GetOrCreateMailbox("Trash", true)
          : client.GetOrCreateMailbox("Trash");

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.AreEqual("Trash", mailbox.Name);
        Assert.AreEqual("Trash", mailbox.FullName);
        Assert.IsFalse(mailbox.IsInbox);
      });
    }

    [Test]
    public void TestGetOrCreateMailbox_NonExistent()
    {
      GetOrCreateMailboxNonExistent(false);
    }

    [Test]
    public void TestGetOrCreateMailbox_NonExistent_Subscribe()
    {
      GetOrCreateMailboxNonExistent(true);
    }

    private void GetOrCreateMailboxNonExistent(bool subscribe)
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" Trash\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("CREATE Trash\r\n",
                                       "$tag OK done\r\n");
          if (subscribe)
            server.EnqueueTaggedResponse("SUBSCRIBE Trash\r\n",
                                         "$tag OK done\r\n");
          server.EnqueueTaggedResponse("LIST \"\" Trash\r\n",
                                       "* LIST () \"\" Trash\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailbox = subscribe
          ? client.GetOrCreateMailbox("Trash", true)
          : client.GetOrCreateMailbox("Trash");

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.AreEqual("Trash", mailbox.Name);
        Assert.AreEqual("Trash", mailbox.FullName);
        Assert.IsFalse(mailbox.IsInbox);
      });
    }

    [Test]
    public void TestGetOrCreateMailbox_NameContainsWildcard()
    {
      GetOrCreateMailboxNameContainsWildcard(false);
    }

    [Test]
    public void TestGetOrCreateMailbox_NameContainsWildcard_Subscribe()
    {
      GetOrCreateMailboxNameContainsWildcard(true);
    }

    private void GetOrCreateMailboxNameContainsWildcard(bool subscribe)
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" Box*/Child\r\n",
                                       "* LIST () \"/\" Box*Box/Child\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("CREATE \"Box*/Child\"\r\n",
                                       "$tag OK done\r\n");
          if (subscribe)
            server.EnqueueTaggedResponse("SUBSCRIBE \"Box*/Child\"\r\n",
                                         "$tag OK done\r\n");
          server.EnqueueTaggedResponse("LIST \"\" Box*/Child\r\n",
                                       "* LIST () \"/\" Box*/Child\r\n" +
                                       "* LIST () \"/\" Box*Box/Child\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailbox = subscribe
          ? client.GetOrCreateMailbox("Box*/Child", true)
          : client.GetOrCreateMailbox("Box*/Child");

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.AreEqual("Child", mailbox.Name);
        Assert.AreEqual("Box*/Child", mailbox.FullName);
        Assert.IsFalse(mailbox.IsInbox);
      });
    }

    [Test]
    public void TestGetOrCreateMailbox_NameNull()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        Assert.Throws<ArgumentNullException>(() => client.GetOrCreateMailbox((string)null));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestGetOrCreateMailbox_NameEmpty()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        Assert.Throws<ArgumentException>(() => client.GetOrCreateMailbox(string.Empty));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestGetMailboxes()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" *\r\n",
                                       "* LIST () \".\" Trash\r\n" +
                                       "* LIST () \".\" INBOX\r\n" +
                                       "* LIST () \".\" INBOX.Child1\r\n" +
                                       "* LIST () \".\" INBOX.Child2\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailboxes = client.GetMailboxes();

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(4, mailboxes.Count, "Count");
#if !NET_4_5
        Assert.Throws<NotSupportedException>(() => mailboxes.Clear(), "readonly");
#endif

        using (var enumerator = mailboxes.GetEnumerator()) {
          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreSame(enumerator.Current.Client, client);
          Assert.AreEqual("Trash", enumerator.Current.Name);
          Assert.AreEqual("Trash", enumerator.Current.FullName);
          Assert.IsFalse(enumerator.Current.IsInbox);

          StringAssert.EndsWith("LIST \"\" *\r\n", server.DequeueRequest());

          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreSame(enumerator.Current.Client, client);
          Assert.AreEqual("INBOX", enumerator.Current.Name);
          Assert.AreEqual("INBOX", enumerator.Current.FullName);
          Assert.IsTrue(enumerator.Current.IsInbox);

          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreSame(enumerator.Current.Client, client);
          Assert.AreEqual("Child1", enumerator.Current.Name);
          Assert.AreEqual("INBOX.Child1", enumerator.Current.FullName);
          Assert.IsFalse(enumerator.Current.IsInbox);

          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreSame(enumerator.Current.Client, client);
          Assert.AreEqual("Child2", enumerator.Current.Name);
          Assert.AreEqual("INBOX.Child2", enumerator.Current.FullName);
          Assert.IsFalse(enumerator.Current.IsInbox);

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMailboxes_NothingMatched()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" *\r\n",
                                       "$tag OK done\r\n");
        });

        var mailboxes = client.GetMailboxes();

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(0, mailboxes.Count, "Count");
#if !NET_4_5
        Assert.Throws<NotSupportedException>(() => mailboxes.Clear(), "readonly");
#endif

        using (var enumerator = mailboxes.GetEnumerator()) {
          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMailboxes_TopLevelOnly()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" %\r\n",
                                       "* LIST () \".\" Trash\r\n" +
                                       "* LIST () \".\" INBOX\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailboxes = client.GetMailboxes(ImapMailboxListOptions.TopLevelOnly);

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(2, mailboxes.Count, "Count");
#if !NET_4_5
        Assert.Throws<NotSupportedException>(() => mailboxes.Clear(), "readonly");
#endif

        using (var enumerator = mailboxes.GetEnumerator()) {
          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreSame(enumerator.Current.Client, client);
          Assert.AreEqual("Trash", enumerator.Current.Name);
          Assert.AreEqual("Trash", enumerator.Current.FullName);

          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreSame(enumerator.Current.Client, client);
          Assert.AreEqual("INBOX", enumerator.Current.Name);
          Assert.AreEqual("INBOX", enumerator.Current.FullName);

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMailboxes_SubscribedOnly()
    {
      GetMailboxesSubscribedOnly(false);
    }

    [Test]
    public void TestGetMailboxes_SubscribedOnly_MailboxReferralsCapable()
    {
      GetMailboxesSubscribedOnly(true);
    }

    private void GetMailboxesSubscribedOnly(bool mailboxReferralsCapable)
    {
      var capas = mailboxReferralsCapable
        ? new ImapCapability[] {ImapCapability.MailboxReferrals}
        : new ImapCapability[] {};

      TestUtils2.TestAuthenticated(capas, delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          var expectedCommand = mailboxReferralsCapable ? "RLSUB \"\" *\r\n" : "LSUB \"\" *\r\n";

          server.EnqueueTaggedResponse(expectedCommand,
                                       "* LSUB () \".\" Trash\r\n" +
                                       "* LSUB () \".\" INBOX\r\n" +
                                       "* LSUB () \".\" INBOX.Child2\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailboxes = client.GetMailboxes(ImapMailboxListOptions.SubscribedOnly | ImapMailboxListOptions.Remote);

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(3, mailboxes.Count, "Count");
#if !NET_4_5
        Assert.Throws<NotSupportedException>(() => mailboxes.Clear(), "readonly");
#endif

        using (var enumerator = mailboxes.GetEnumerator()) {
          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreSame(enumerator.Current.Client, client);
          Assert.AreEqual("Trash", enumerator.Current.Name);
          Assert.AreEqual("Trash", enumerator.Current.FullName);

          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreSame(enumerator.Current.Client, client);
          Assert.AreEqual("INBOX", enumerator.Current.Name);
          Assert.AreEqual("INBOX", enumerator.Current.FullName);

          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreSame(enumerator.Current.Client, client);
          Assert.AreEqual("Child2", enumerator.Current.Name);
          Assert.AreEqual("INBOX.Child2", enumerator.Current.FullName);

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMailboxes_SubscribedOnly_ListExtendedCapable()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.ListExtended},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST (SUBSCRIBED) \"\" *",
                                       "* LIST (\\Marked \\NoInferiors \\Subscribed) \"/\" \"inbox\"\r\n" +
                                       "* LIST (\\Subscribed) \"/\" \"Fruit/Banana\"\r\n" +
                                       "* LIST (\\Subscribed \\NonExistent) \"/\" \"Fruit/Peach\"\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailboxes = client.GetMailboxes(ImapMailboxListOptions.SubscribedOnly);

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(3, mailboxes.Count, "Count");
#if !NET_4_5
        Assert.Throws<NotSupportedException>(() => mailboxes.Clear(), "readonly");
#endif

        using (var enumerator = mailboxes.GetEnumerator()) {
          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreSame(enumerator.Current.Client, client);
          Assert.AreEqual("INBOX", enumerator.Current.Name);
          Assert.AreEqual("INBOX", enumerator.Current.FullName);
          Assert.IsFalse(enumerator.Current.CanHaveChild);

          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreSame(enumerator.Current.Client, client);
          Assert.AreEqual("Banana", enumerator.Current.Name);
          Assert.AreEqual("Fruit/Banana", enumerator.Current.FullName);
          Assert.IsTrue(enumerator.Current.Exists);
          Assert.IsTrue(enumerator.Current.CanHaveChild);

          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreSame(enumerator.Current.Client, client);
          Assert.AreEqual("Peach", enumerator.Current.Name);
          Assert.AreEqual("Fruit/Peach", enumerator.Current.FullName);
          Assert.IsFalse(enumerator.Current.Exists);
          Assert.IsTrue(enumerator.Current.CanHaveChild);

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMailboxes_Remote_ListExtendedCapable()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.ListExtended},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST (REMOTE) \"\" *",
                                       "* LIST () \"/\" \"Fruit/Banana\"\r\n" +
                                       "* LIST (\\Remote) \"/\" \"Bread\"\r\n" +
                                       "$tag OK done\r\n");

        });

        var mailboxes = client.GetMailboxes(ImapMailboxListOptions.Remote);

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(2, mailboxes.Count, "Count");
#if !NET_4_5
        Assert.Throws<NotSupportedException>(() => mailboxes.Clear(), "readonly");
#endif

        using (var enumerator = mailboxes.GetEnumerator()) {
          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreSame(enumerator.Current.Client, client);
          Assert.AreEqual("Banana", enumerator.Current.Name);
          Assert.AreEqual("Fruit/Banana", enumerator.Current.FullName);

          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreSame(enumerator.Current.Client, client);
          Assert.AreEqual("Bread", enumerator.Current.Name);
          Assert.AreEqual("Bread", enumerator.Current.FullName);

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMailboxes_Remote_MailboxReferralsCapable()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.MailboxReferrals},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("RLIST \"\" *\r\n",
                                       "* LIST () \".\" Trash\r\n" +
                                       "* LIST () \".\" INBOX\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailboxes = client.GetMailboxes(ImapMailboxListOptions.Remote);

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(2, mailboxes.Count, "Count");
#if !NET_4_5
        Assert.Throws<NotSupportedException>(() => mailboxes.Clear(), "readonly");
#endif

        using (var enumerator = mailboxes.GetEnumerator()) {
          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreSame(enumerator.Current.Client, client);
          Assert.AreEqual("Trash", enumerator.Current.Name);
          Assert.AreEqual("Trash", enumerator.Current.FullName);

          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreSame(enumerator.Current.Client, client);
          Assert.AreEqual("INBOX", enumerator.Current.Name);
          Assert.AreEqual("INBOX", enumerator.Current.FullName);

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMailboxes_RequestStatus()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        GetMailboxesRequestStatus(server, client, false);
      });
    }

    [Test]
    public void TestGetMailboxes_RequestStatus_CondStoreDisabled()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.CondStore},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        client.DisabledCapabilities.Add(ImapCapability.CondStore);

        GetMailboxesRequestStatus(server, client, false);
      });
    }

    [Test]
    public void TestGetMailboxes_RequestStatus_CondStoreCapable()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.CondStore},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        GetMailboxesRequestStatus(server, client, true);
      });
    }


    private void GetMailboxesRequestStatus(ImapPseudoServer2 server, ImapClient client, bool condStoreAvailable)
    {
      var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
        var expectedStatusDataItems = condStoreAvailable
                                      ? "(MESSAGES RECENT UIDNEXT UIDVALIDITY UNSEEN HIGHESTMODSEQ)"
                                      : "(MESSAGES RECENT UIDNEXT UIDVALIDITY UNSEEN)";
        server.EnqueueTaggedResponse("LIST \"\" *\r\n",
                                     "* LIST () \".\" Trash\r\n" +
                                     "* LIST () \".\" INBOX\r\n" +
                                     "* LIST () \".\" INBOX.Child\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse(string.Format("STATUS Trash {0}\r\n", expectedStatusDataItems),
                                     "* STATUS Trash (MESSAGES 3 RECENT 1 UIDNEXT 4 UIDVALIDITY 1 UNSEEN 1 HIGHESTMODSEQ 1)\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse(string.Format("STATUS INBOX {0}\r\n", expectedStatusDataItems),
                                     "* STATUS INBOX ()\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse(string.Format("STATUS INBOX.Child {0}\r\n", expectedStatusDataItems),
                                     "* STATUS INBOX.Child ()\r\n" +
                                     "$tag OK done\r\n");
      });

      var mailboxes = client.GetMailboxes(ImapMailboxListOptions.RequestStatus);

      responseTask.Wait();

      Assert.IsFalse(client.IsBusy);

      Assert.IsNotNull(mailboxes);
      Assert.AreEqual(3, mailboxes.Count, "Count");
#if !NET_4_5
      Assert.Throws<NotSupportedException>(() => mailboxes.Clear(), "readonly");
#endif

      using (var enumerator = mailboxes.GetEnumerator()) {
        Assert.IsNotNull(mailboxes);

        Assert.IsTrue(enumerator.MoveNext());
        Assert.AreSame(enumerator.Current.Client, client);
        Assert.AreEqual("Trash", enumerator.Current.Name);
        Assert.AreEqual("Trash", enumerator.Current.FullName);
        Assert.AreEqual(3L, enumerator.Current.ExistMessageCount);
        Assert.AreEqual(1L, enumerator.Current.RecentMessageCount);
        Assert.AreEqual(4L, enumerator.Current.NextUid);
        Assert.AreEqual(1L, enumerator.Current.UidValidity);
        if (condStoreAvailable)
          Assert.AreEqual(1L, enumerator.Current.HighestModSeq);

        Assert.IsTrue(enumerator.MoveNext());
        Assert.AreSame(enumerator.Current.Client, client);
        Assert.AreEqual("INBOX", enumerator.Current.Name);
        Assert.AreEqual("INBOX", enumerator.Current.FullName);

        Assert.IsTrue(enumerator.MoveNext());
        Assert.AreSame(enumerator.Current.Client, client);
        Assert.AreEqual("Child", enumerator.Current.Name);
        Assert.AreEqual("INBOX.Child", enumerator.Current.FullName);

        Assert.IsFalse(enumerator.MoveNext());
      }
    }

    [Test]
    public void TestGetMailboxes_RequestStatus_NothingMatched()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" *\r\n",
                                       "$tag OK done\r\n");
        });

        var mailboxes = client.GetMailboxes(ImapMailboxListOptions.RequestStatus);

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(0, mailboxes.Count, "Count");
#if !NET_4_5
        Assert.Throws<NotSupportedException>(() => mailboxes.Clear(), "readonly");
#endif

        using (var enumerator = mailboxes.GetEnumerator()) {
          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMailboxes_RequestStatus_UnselectableMailbox()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" *\r\n",
                                       "* LIST () \"/\" blurdybloop\r\n" +
                                       "* LIST (\\Noselect) \"/\" foo\r\n" +
                                       "* LIST () \"/\" foo/bar\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("STATUS blurdybloop ",
                                       "* STATUS blurdybloop ()\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("STATUS foo/bar ",
                                       "* STATUS foo/bar ()\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailboxes = client.GetMailboxes(ImapMailboxListOptions.RequestStatus);

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(3, mailboxes.Count, "Count");
#if !NET_4_5
        Assert.Throws<NotSupportedException>(() => mailboxes.Clear(), "readonly");
#endif

        using (var enumerator = mailboxes.GetEnumerator()) {
          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreSame(enumerator.Current.Client, client);
          Assert.AreEqual("blurdybloop", enumerator.Current.Name);
          Assert.AreEqual("blurdybloop", enumerator.Current.FullName);

          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreSame(enumerator.Current.Client, client);
          Assert.AreEqual("foo", enumerator.Current.Name);
          Assert.AreEqual("foo", enumerator.Current.FullName);
          Assert.IsTrue(enumerator.Current.IsUnselectable);

          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreSame(enumerator.Current.Client, client);
          Assert.AreEqual("bar", enumerator.Current.Name);
          Assert.AreEqual("foo/bar", enumerator.Current.FullName);
        }
      });
    }

    [Test]
    public void TestGetMailboxes_RequestStatus_SelectedMailbox()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        /*
         * OpenInbox
         */
        server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                     "* LIST () \"\" INBOX\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                     "$tag OK done\r\n");

        Assert.IsNotNull(client.OpenInbox());
        Assert.IsNotNull(client.OpenedMailbox);

        /*
         * GetMailboxes
         */
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" *\r\n",
                                       "* LIST () \".\" Trash\r\n" +
                                       "* LIST () \".\" INBOX\r\n" +
                                       "* LIST () \".\" INBOX.Child\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("STATUS Trash ",
                                       "* STATUS Trash ()\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("STATUS INBOX.Child ",
                                       "* STATUS INBOX.Child ()\r\n" +
                                       "$tag OK done\r\n");

        });

        var mailboxes = client.GetMailboxes(ImapMailboxListOptions.RequestStatus);

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(3, mailboxes.Count, "Count");
#if !NET_4_5
        Assert.Throws<NotSupportedException>(() => mailboxes.Clear(), "readonly");
#endif

        using (var enumerator = mailboxes.GetEnumerator()) {
          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreSame(enumerator.Current.Client, client);
          Assert.AreEqual("Trash", enumerator.Current.Name);
          Assert.AreEqual("Trash", enumerator.Current.FullName);

          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreSame(enumerator.Current.Client, client);
          Assert.AreEqual("INBOX", enumerator.Current.Name);
          Assert.AreEqual("INBOX", enumerator.Current.FullName);

          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreSame(enumerator.Current.Client, client);
          Assert.AreEqual("Child", enumerator.Current.Name);
          Assert.AreEqual("INBOX.Child", enumerator.Current.FullName);

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMailboxes_RequestStatus_ListStatusCapable()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.ListExtended, ImapCapability.ListStatus},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        GetMailboxesRequestStatusListStatusCapable(server, client, false);
      });
    }

    [Test]
    public void TestGetMailboxes_RequestStatus_ListStatusCapable_CondStoreDisabled()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.ListExtended, ImapCapability.ListStatus, ImapCapability.CondStore},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        client.DisabledCapabilities.Add(ImapCapability.CondStore);

        GetMailboxesRequestStatusListStatusCapable(server, client, false);
      });
    }

    [Test]
    public void TestGetMailboxes_RequestStatus_ListStatusCapable_CondStoreCapable()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.ListExtended, ImapCapability.ListStatus, ImapCapability.CondStore},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        GetMailboxesRequestStatusListStatusCapable(server, client, true);
      });
    }

    private void GetMailboxesRequestStatusListStatusCapable(ImapPseudoServer2 server, ImapClient client, bool condStoreAvailable)
    {
      var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
        var expectedStatusDataItems = condStoreAvailable
                                    ? "(MESSAGES RECENT UIDNEXT UIDVALIDITY UNSEEN HIGHESTMODSEQ)"
                                    : "(MESSAGES RECENT UIDNEXT UIDVALIDITY UNSEEN)";

        server.EnqueueTaggedResponse(string.Format("LIST () \"\" * RETURN (CHILDREN STATUS {0})\r\n", expectedStatusDataItems),
                                     "* LIST () \".\" Trash\r\n" +
                                     "* STATUS Trash (MESSAGES 3 RECENT 1 UIDNEXT 4 UIDVALIDITY 1 UNSEEN 1 HIGHESTMODSEQ 1)\r\n" +
                                     "* LIST () \".\" INBOX\r\n" +
                                     "* STATUS INBOX ()\r\n" +
                                     "* LIST () \".\" INBOX.Child\r\n" +
                                     "* STATUS INBOX.Child ()\r\n" +
                                     "$tag OK done\r\n");
      });

      var mailboxes = client.GetMailboxes(ImapMailboxListOptions.RequestStatus);

      responseTask.Wait();

      Assert.IsFalse(client.IsBusy);

      Assert.IsNotNull(mailboxes);
      Assert.AreEqual(3, mailboxes.Count, "Count");
#if !NET_4_5
      Assert.Throws<NotSupportedException>(() => mailboxes.Clear(), "readonly");
#endif

      using (var enumerator = mailboxes.GetEnumerator()) {
        Assert.IsTrue(enumerator.MoveNext());
        Assert.AreSame(enumerator.Current.Client, client);
        Assert.AreEqual("Trash", enumerator.Current.Name);
        Assert.AreEqual("Trash", enumerator.Current.FullName);
        Assert.AreEqual(3L, enumerator.Current.ExistMessageCount);
        Assert.AreEqual(1L, enumerator.Current.RecentMessageCount);
        Assert.AreEqual(4L, enumerator.Current.NextUid);
        Assert.AreEqual(1L, enumerator.Current.UidValidity);
        Assert.AreEqual(1L, enumerator.Current.UnseenMessageCount);
        if (condStoreAvailable)
          Assert.AreEqual(1L, enumerator.Current.HighestModSeq);

        Assert.IsTrue(enumerator.MoveNext());
        Assert.AreSame(enumerator.Current.Client, client);
        Assert.AreEqual("INBOX", enumerator.Current.Name);
        Assert.AreEqual("INBOX", enumerator.Current.FullName);

        Assert.IsTrue(enumerator.MoveNext());
        Assert.AreSame(enumerator.Current.Client, client);
        Assert.AreEqual("Child", enumerator.Current.Name);
        Assert.AreEqual("INBOX.Child", enumerator.Current.FullName);

        Assert.IsFalse(enumerator.MoveNext());
      }
    }

    [Test]
    public void TestCreateMailbox()
    {
      CreateMailbox(false);
    }

    [Test]
    public void TestCreateMailbox_Subscribe()
    {
      CreateMailbox(true);
    }

    private void CreateMailbox(bool subscribe)
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("CREATE INBOX.Recent\r\n",
                                       "$tag OK done\r\n");
          if (subscribe)
            server.EnqueueTaggedResponse("SUBSCRIBE INBOX.Recent\r\n",
                                         "$tag OK done\r\n");
          server.EnqueueTaggedResponse("LIST \"\" INBOX.Recent\r\n",
                                       "* LIST () \".\" INBOX.Recent\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailbox = subscribe
          ? client.CreateMailbox("INBOX.Recent", true)
          : client.CreateMailbox("INBOX.Recent");

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.AreEqual("Recent", mailbox.Name);
        Assert.AreEqual("INBOX.Recent", mailbox.FullName);
      });
    }

    [Test]
    public void TestCreateMailbox_NameContainsWildcard_Asterisk()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("CREATE \"Box*\"\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("LIST \"\" Box*\r\n",
                                       "* LIST (\\Marked) \".\" Box*\r\n" +
                                       "* LIST () \".\" Box\r\n" +
                                       "* LIST () \".\" Box.Child\r\n" +
                                       "* LIST () \".\" Boxes\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailbox = client.CreateMailbox("Box*");

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.AreEqual("Box*", mailbox.Name);
        Assert.AreEqual("Box*", mailbox.FullName);
        Assert.AreEqual(1, mailbox.Flags.Count);
        CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.Marked);
        Assert.IsFalse(mailbox.IsInbox);
      });
    }

    [Test]
    public void TestCreateMailbox_NameContainsWildcard_Percent()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("CREATE \"Box%\"\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("LIST \"\" Box%\r\n",
                                       "* LIST (\\Marked) \".\" Box%\r\n" +
                                       "* LIST () \".\" Box\r\n" +
                                       "* LIST () \".\" Boxes\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailbox = client.CreateMailbox("Box%");

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.AreEqual("Box%", mailbox.Name);
        Assert.AreEqual("Box%", mailbox.FullName);
        Assert.AreEqual(1, mailbox.Flags.Count);
        CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.Marked);
        Assert.IsFalse(mailbox.IsInbox);
      });
    }

    [Test]
    public void TestCreateMailbox_NameNull()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        Assert.Throws<ArgumentNullException>(() => client.CreateMailbox(null));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestCreateMailbox_NameEmpty()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        Assert.Throws<ArgumentException>(() => client.CreateMailbox(string.Empty));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestCreateMailbox_AlreadyExists()
    {
      CreateMailboxAlreadyExists(false);
    }

    [Test]
    public void TestCreateMailbox_AlreadyExists_Subscribe()
    {
      CreateMailboxAlreadyExists(true);
    }

    private void CreateMailboxAlreadyExists(bool subscribe)
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("CREATE INBOX.Recent\r\n",
                                       "$tag NO [ALREADYEXISTS] mailbox already exists\r\n");
        });

        try {
          if (subscribe)
            client.CreateMailbox("INBOX.Recent", true);
          else
            client.CreateMailbox("INBOX.Recent");
        }
        catch (ImapErrorResponseException ex) {
          Assert.AreEqual(ImapResponseCode.AlreadyExists,
                          ex.Result.TaggedStatusResponse.ResponseText.Code);
        }

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestOpenInbox()
    {
      OpenInbox(false);
    }

    [Test]
    public void TestOpenInbox_AsReadOnly()
    {
      OpenInbox(true);
    }

    private void OpenInbox(bool asReadOnly)
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                       "* LIST () \"\" INBOX\r\n" +
                                       "$tag OK done\r\n");
          if (asReadOnly)
            server.EnqueueTaggedResponse("EXAMINE INBOX\r\n",
                                         "* OK [READ-ONLY]\r\n" +
                                         "$tag OK done\r\n");
          else
            server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                         "* OK [READ-WRITE]\r\n" +
                                         "$tag OK done\r\n");
        });

        var mailbox = asReadOnly
          ? client.OpenInbox(true)
          : client.OpenInbox();

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.IsNotNull(client.OpenedMailbox);
        Assert.AreSame(mailbox, client.OpenedMailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.IsTrue(mailbox.IsOpen);
        Assert.AreEqual("INBOX", mailbox.Name);
        Assert.AreEqual("INBOX", mailbox.FullName);

        if (asReadOnly)
          Assert.IsTrue(mailbox.IsReadOnly);
        else
          Assert.IsFalse(mailbox.IsReadOnly);
      });
    }

    [Test]
    public void TestOpenMailbox()
    {
      OpenMailbox(false);
    }

    [Test]
    public void TestOpenMailbox_AsReadOnly()
    {
      OpenMailbox(true);
    }

    private void OpenMailbox(bool asReadOnly)
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" Draft\r\n",
                                       "* LIST () \"\" Draft\r\n" +
                                       "$tag OK done\r\n");
          if (asReadOnly)
            server.EnqueueTaggedResponse("EXAMINE Draft\r\n",
                                         "* OK [READ-ONLY]\r\n" +
                                         "$tag OK done\r\n");
          else
            server.EnqueueTaggedResponse("SELECT Draft\r\n",
                                         "* OK [READ-WRITE]\r\n" +
                                         "$tag OK done\r\n");
        });

        var mailbox = asReadOnly
          ? client.OpenMailbox("Draft", true)
          : client.OpenMailbox("Draft");

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.IsNotNull(client.OpenedMailbox);
        Assert.AreSame(mailbox, client.OpenedMailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.IsTrue(mailbox.IsOpen);
        Assert.AreEqual("Draft", mailbox.Name);
        Assert.AreEqual("Draft", mailbox.FullName);
        Assert.IsFalse(mailbox.IsModSequencesAvailable);

        if (asReadOnly)
          Assert.IsTrue(mailbox.IsReadOnly);
        else
          Assert.IsFalse(mailbox.IsReadOnly);
      });
    }

    [Test]
    public void TestOpenMailbox_NotFound()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST ",
                                       "$tag OK done\r\n");
        });

        var ex = Assert.Throws<ImapMailboxNotFoundException>(() => client.OpenMailbox("NonExistent"));

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(ex.Mailbox);
        Assert.AreEqual("NonExistent", ex.Mailbox);

        Smdn.Net.TestUtils.SerializeBinary(ex, delegate(ImapMailboxNotFoundException deserialized) {
          Assert.IsNotNull(deserialized.Mailbox);
          Assert.AreEqual(ex.Mailbox, deserialized.Mailbox);
        });
      });
    }

    [Test]
    public void TestOpenMailbox_CondStoreCapable()
    {
      OpenMailboxCondStore(false);
    }

    [Test]
    public void TestOpenMailbox_CondStoreCapable_AsReadOnly()
    {
      OpenMailboxCondStore(true);
    }

    private void OpenMailboxCondStore(bool asReadOnly)
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.CondStore},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                       "* LIST () \"\" INBOX\r\n" +
                                       "$tag OK done\r\n");
          if (asReadOnly)
            server.EnqueueTaggedResponse("EXAMINE INBOX (CONDSTORE)\r\n",
                                         "* OK [HIGHESTMODSEQ 1]\r\n" +
                                         "* OK [READ-ONLY]\r\n" +
                                         "$tag OK done\r\n");
          else
            server.EnqueueTaggedResponse("SELECT INBOX (CONDSTORE)\r\n",
                                         "* OK [HIGHESTMODSEQ 1]\r\n" +
                                         "* OK [READ-WRITE]\r\n" +
                                         "$tag OK done\r\n");
        });

        var mailbox = asReadOnly
          ? client.OpenMailbox("INBOX", true)
          : client.OpenMailbox("INBOX");

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.IsNotNull(client.OpenedMailbox);
        Assert.AreSame(mailbox, client.OpenedMailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.IsTrue(mailbox.IsOpen);
        Assert.AreEqual("INBOX", mailbox.Name);
        Assert.AreEqual("INBOX", mailbox.FullName);
        Assert.IsTrue(mailbox.IsModSequencesAvailable);

        if (asReadOnly)
          Assert.IsTrue(mailbox.IsReadOnly);
        else
          Assert.IsFalse(mailbox.IsReadOnly);
      });
    }

    [Test]
    public void TestOpenMailbox_CondStoreCapable_NoModSeq()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.CondStore},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                       "* LIST () \"\" INBOX\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("SELECT INBOX (CONDSTORE)\r\n",
                                       "* OK [NOMODSEQ]\r\n" +
                                       "* OK [READ-WRITE]\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailbox = client.OpenMailbox("INBOX");

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.IsNotNull(client.OpenedMailbox);
        Assert.AreSame(mailbox, client.OpenedMailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.IsTrue(mailbox.IsOpen);
        Assert.AreEqual("INBOX", mailbox.Name);
        Assert.AreEqual("INBOX", mailbox.FullName);
        Assert.IsFalse(mailbox.IsModSequencesAvailable);
        Assert.IsFalse(mailbox.IsReadOnly);
      });
    }

    [Test]
    public void TestOpenMailbox_NonExistent()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.ListExtended},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST (SUBSCRIBED) \"\" INBOX/Child",
                                       "* LIST (\\Subscribed \\NonExistent) \"/\" \"INBOX/Child\"\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailbox = client.GetMailbox("INBOX/Child", ImapMailboxListOptions.SubscribedOnly);

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.AreEqual("Child", mailbox.Name);
        Assert.AreEqual("INBOX/Child", mailbox.FullName);
        Assert.IsFalse(mailbox.Exists);

        Assert.Throws<ImapProtocolViolationException>(() => client.OpenMailbox(mailbox));
      });
    }

    [Test]
    public void TestOpenMailbox_Unselectable()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" [Gmail]\r\n",
                                       "* LIST (\\Noselect \\HasChildren) \"/\" \"[Gmail]\"\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailbox = client.GetMailbox("[Gmail]");

        responseTask.Wait();

        Assert.IsNotNull(mailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.AreEqual("[Gmail]", mailbox.Name);
        Assert.AreEqual("[Gmail]", mailbox.FullName);
        Assert.IsTrue(mailbox.IsUnselectable);

        Assert.Throws<ImapProtocolViolationException>(() => client.OpenMailbox(mailbox));

        Assert.IsNull(client.OpenedMailbox);

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestOpenMailbox_Deleted()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        server.EnqueueTaggedResponse("LIST \"\" INBOX/Child\r\n",
                                     "* LIST () \"/\" \"INBOX/Child\"\r\n" +
                                     "$tag OK done\r\n");

        var mailbox = client.GetMailbox("INBOX/Child");

        Assert.IsNotNull(mailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.AreEqual("Child", mailbox.Name);
        Assert.AreEqual("INBOX/Child", mailbox.FullName);
        Assert.IsTrue(mailbox.Exists);

        server.EnqueueTaggedResponse("DELETE INBOX/Child\r\n",
                                     "$tag OK done\r\n");

        mailbox.Delete();

        Assert.Throws<ImapProtocolViolationException>(() => client.OpenMailbox(mailbox));

        Assert.IsFalse(client.IsBusy);

        Assert.IsNull(client.OpenedMailbox);
      });
    }

    [Test]
    public void TestOpenMailbox_MailboxOfDifferentSession()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server1, ImapClient client1) {
        server1.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                      "* LIST () \"\" \"INBOX\"\r\n" +
                                     "$tag OK done\r\n");

        var mailbox = client1.GetMailbox("INBOX");

        Assert.IsNotNull(mailbox);
        Assert.AreSame(mailbox.Client, client1);
        Assert.AreEqual("INBOX", mailbox.Name);
        Assert.AreEqual("INBOX", mailbox.FullName);
        Assert.IsTrue(mailbox.Exists);

        TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server2, ImapClient client2) {
          Assert.Throws<ArgumentException>(() => client2.OpenMailbox(mailbox));

          Assert.IsFalse(client2.IsBusy);

          Assert.IsNull(client1.OpenedMailbox);
          Assert.IsNull(client2.OpenedMailbox);
        });
      });
    }

    [Test]
    public void TestOpenMailbox_NameContainsWildcard()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" Box*\r\n",
                                       "* LIST () \".\" Box*\r\n" +
                                       "* LIST () \".\" Box\r\n" +
                                       "* LIST () \".\" Box.Child\r\n" +
                                       "* LIST () \".\" Boxes\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("SELECT \"Box*\"\r\n",
                                       "* OK [READ-WRITE]\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailbox = client.OpenMailbox("Box*");

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.IsNotNull(client.OpenedMailbox);
        Assert.AreSame(mailbox, client.OpenedMailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.IsTrue(mailbox.IsOpen);
        Assert.AreEqual("Box*", mailbox.Name);
        Assert.AreEqual("Box*", mailbox.FullName);
        Assert.IsFalse(mailbox.IsModSequencesAvailable);
        Assert.IsFalse(mailbox.IsReadOnly);
      });
    }

    [Test]
    public void TestOpenMailbox_NameNull()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        Assert.Throws<ArgumentNullException>(() => client.OpenMailbox((string)null));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestOpenMailbox_NameEmpty()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        Assert.Throws<ArgumentException>(() => client.OpenMailbox(string.Empty));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test, Ignore("not implemented; reopen")]
    public void TestOpenMailbox_ReopenCurrentlySelected()
    {
    }

    [Test, Ignore("not implemented; reopen")]
    public void TestOpenMailbox_OpenNewly()
    {
    }

    [Test]
    public void TestOpenMailbox_Error()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                       "* LIST () \"\" INBOX\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                       "$tag NO unselectable\r\n");
        });

        Assert.IsNull(client.OpenedMailbox);

        Assert.Throws<ImapErrorResponseException>(() => client.OpenMailbox("INBOX"));

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsFalse(client.IsBusy);
        Assert.IsNull(client.OpenedMailbox);
      });
    }

    [Test]
    public void TestOpenOrCreateMailbox_Existent()
    {
      OpenOrCreateMailbox(true, false, false);
    }

    [Test]
    public void TestOpenOrCreateMailbox_Existent_AsReadOnly()
    {
      OpenOrCreateMailbox(true, true, false);
    }

    [Test]
    public void TestOpenOrCreateMailbox_NonExistent()
    {
      OpenOrCreateMailbox(false, false, false);
    }

    [Test]
    public void TestOpenOrCreateMailbox_NonExistent_AsReadOnly()
    {
      OpenOrCreateMailbox(false, true, false);
    }

    [Test]
    public void TestOpenOrCreateMailbox_RequestStatus_Existent()
    {
      OpenOrCreateMailbox(true, false, true);
    }

    [Test]
    public void TestOpenOrCreateMailbox_RequestStatus_NonExistent()
    {
      OpenOrCreateMailbox(false, false, true);
    }

    private void OpenOrCreateMailbox(bool exist, bool asReadOnly, bool requestStatus)
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          if (!exist) {
            server.EnqueueTaggedResponse("LIST \"\" INBOX.child\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("CREATE INBOX.child\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE INBOX.child\r\n",
                                         "$tag OK done\r\n");
          }

          server.EnqueueTaggedResponse("LIST \"\" INBOX.child\r\n",
                                       "* LIST () \".\" INBOX.child\r\n" +
                                       "$tag OK done\r\n");

          if (asReadOnly)
            server.EnqueueTaggedResponse("EXAMINE INBOX.child\r\n",
                                         "* OK [READ-ONLY]\r\n" +
                                         "$tag OK done\r\n");
          else
            server.EnqueueTaggedResponse("SELECT INBOX.child\r\n",
                                         "* OK [READ-WRITE]\r\n" +
                                         "$tag OK done\r\n");
        });

        var options = requestStatus ? ImapMailboxListOptions.RequestStatus : ImapMailboxListOptions.Default;

        var mailbox = client.OpenOrCreateMailbox("INBOX.child",
                                                 true, // subscribe
                                                 asReadOnly,
                                                 options);

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.IsNotNull(client.OpenedMailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.IsTrue(mailbox.IsOpen);
        Assert.AreEqual("child", mailbox.Name);
        Assert.AreEqual("INBOX.child", mailbox.FullName);
        Assert.IsFalse(mailbox.IsModSequencesAvailable);
        if (asReadOnly)
          Assert.IsTrue(mailbox.IsReadOnly);
        else
          Assert.IsFalse(mailbox.IsReadOnly);
      });
    }

    [Test]
    public void TestOpenOrCreateMailbox_Unselectable_Existent()
    {
      OpenOrCreateMailbox_Unselectable(true, false);
    }

    [Test]
    public void TestOpenOrCreateMailbox_Unselectable_Existent_AsReadOnly()
    {
      OpenOrCreateMailbox_Unselectable(true, true);
    }

    [Test]
    public void TestOpenOrCreateMailbox_Unselectable_NonExistent()
    {
      OpenOrCreateMailbox_Unselectable(false, false);
    }

    [Test]
    public void TestOpenOrCreateMailbox_Unselectable_NonExistent_AsReadOnly()
    {
      OpenOrCreateMailbox_Unselectable(false, true);
    }

    private void OpenOrCreateMailbox_Unselectable(bool exist, bool asReadOnly)
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          if (!exist) {
            server.EnqueueTaggedResponse("LIST \"\" INBOX.child\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("CREATE INBOX.child\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE INBOX.child\r\n",
                                         "$tag OK done\r\n");
          }

          server.EnqueueTaggedResponse("LIST \"\" INBOX.child\r\n",
                                       "* LIST (\\Noselect) \".\" INBOX.child\r\n" +
                                       "$tag OK done\r\n");
        });

        Assert.Throws<ImapProtocolViolationException>(() => {
          client.OpenOrCreateMailbox("INBOX.child",
                                     true, // subscribe
                                     asReadOnly,
                                     ImapMailboxListOptions.Default);
        });

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNull(client.OpenedMailbox);
      });
    }

    [Test]
    public void TestOpenOrCreateMailbox_NameContainsWildcard()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" Box*\r\n",
                                       "* LIST () \".\" Box\r\n" +
                                       "* LIST () \".\" Box.Child\r\n" +
                                       "* LIST () \".\" Boxes\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("CREATE \"Box*\"\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("LIST \"\" Box*\r\n",
                                       "* LIST () \".\" Box*\r\n" +
                                       "* LIST () \".\" Box\r\n" +
                                       "* LIST () \".\" Box.Child\r\n" +
                                       "* LIST () \".\" Boxes\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("SELECT \"Box*\"\r\n",
                                       "* OK [READ-WRITE]\r\n" +
                                       "$tag OK done\r\n");
        });

        var mailbox = client.OpenOrCreateMailbox("Box*");

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(mailbox);
        Assert.IsNotNull(client.OpenedMailbox);
        Assert.AreSame(mailbox, client.OpenedMailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.IsTrue(mailbox.IsOpen);
        Assert.AreEqual("Box*", mailbox.Name);
        Assert.AreEqual("Box*", mailbox.FullName);
        Assert.IsFalse(mailbox.IsModSequencesAvailable);
        Assert.IsFalse(mailbox.IsReadOnly);
      });
    }

    [Test]
    public void TestOpenOrCreateMailbox_NameNull()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        Assert.Throws<ArgumentNullException>(() => client.OpenOrCreateMailbox((string)null));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestOpenOrCreateMailbox_NameEmpty()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        Assert.Throws<ArgumentException>(() => client.OpenOrCreateMailbox(string.Empty));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestCloseMailbox()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                     "* LIST () \"\" INBOX\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                     "$tag OK done\r\n");

        var mailbox = client.OpenInbox();

        Assert.IsNotNull(mailbox);
        Assert.IsNotNull(client.OpenedMailbox);
        Assert.AreSame(mailbox, client.OpenedMailbox);
        Assert.IsTrue(mailbox.IsOpen);
        Assert.AreEqual("INBOX", mailbox.Name);
        Assert.AreEqual("INBOX", mailbox.FullName);

        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("CLOSE\r\n",
                                       "$tag OK done\r\n");
        });

        client.CloseMailbox();

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNull(client.OpenedMailbox);
        Assert.IsFalse(mailbox.IsOpen);
      });
    }

    [Test]
    public void TestCloseMailbox_NotOpened()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        Assert.IsNull(client.OpenedMailbox);

        client.CloseMailbox();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNull(client.OpenedMailbox);

        client.CloseMailbox();

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestGetQuota()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.Quota},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("GETQUOTA \"\"\r\n",
                                       "* QUOTA \"\" (STORAGE 10 512)\r\n" +
                                       "$tag OK done\r\n");
        });

        var quota = client.GetQuota(string.Empty);

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.IsNotNull(quota);
        Assert.AreEqual(string.Empty, quota.Root);
        Assert.AreEqual(1, quota.Resources.Count);
        Assert.AreEqual("STORAGE", quota.Resources[0].Name);
        Assert.AreEqual(10, quota.Resources[0].Usage);
        Assert.AreEqual(512, quota.Resources[0].Limit);
      });
    }

    [Test]
    public void TestGetQuota_QuotaIncapable()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var quota = client.GetQuota(string.Empty);

        Assert.IsNull(quota);

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestGetQuota_QuotaDisabled()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.Quota},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        client.DisabledCapabilities.Add(ImapCapability.Quota);

        var quota = client.GetQuota(string.Empty);

        Assert.IsNull(quota);

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestGetQuotaUsage()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.Quota},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("GETQUOTA \"\"\r\n",
                                       "* QUOTA \"\" (STORAGE 10 512)\r\n" +
                                       "$tag OK done\r\n");
        });

        var usage = client.GetQuotaUsage(string.Empty, "STORAGE");

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.AreEqual(10.0 / 512.0, usage);
      });
    }

    [Test]
    public void TestGetQuotaUsage_NoSuchQuotaRoot()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.Quota},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("GETQUOTA ",
                                       "$tag NO no such quota root\r\n");
        });

        Assert.Throws<ImapErrorResponseException>(() => client.GetQuotaUsage("NON-EXISTENT", "STORAGE"));

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestGetQuotaUsage_NoSuchResourceName()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.Quota},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("GETQUOTA ",
                                       "* QUOTA \"\" (STORAGE 10 512)\r\n" +
                                       "$tag OK done\r\n");
        });

        Assert.Throws<ImapProtocolViolationException>(() => client.GetQuotaUsage(string.Empty, "NON-EXITENT"));

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestGetQuotaUsage_QuotaIncapable()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var usage = client.GetQuotaUsage(string.Empty, "STORAGE");

        Assert.AreEqual(0.0, usage);
      });
    }

    [Test]
    public void TestGetQuotaUsage_QuotaDisabled()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.Quota},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        client.DisabledCapabilities.Add(ImapCapability.Quota);

        var usage = client.GetQuotaUsage(string.Empty, "STORAGE");

        Assert.AreEqual(0.0, usage);

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestRefresh()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("NOOP\r\n",
                                       "$tag OK done\r\n");
        });

        client.Refresh();

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestRefresh_MailboxOpened()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                     "* LIST () \"\" INBOX\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                     "* EXISTS 10\r\n" +
                                     "* RECENT 3\r\n" +
                                     "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
                                     "* OK [READ-WRITE]\r\n" +
                                     "$tag OK done\r\n");

        var mailbox = client.OpenInbox();

        Assert.IsNotNull(mailbox);
        Assert.IsNotNull(client.OpenedMailbox);
        Assert.AreSame(mailbox, client.OpenedMailbox);
        Assert.AreSame(mailbox.Client, client);
        Assert.IsTrue(mailbox.IsOpen);
        Assert.AreEqual("INBOX", mailbox.Name);
        Assert.AreEqual("INBOX", mailbox.FullName);
        Assert.IsFalse(mailbox.IsReadOnly);
        Assert.AreEqual(10L, mailbox.ExistMessageCount);
        Assert.AreEqual(3L, mailbox.RecentMessageCount);
        Assert.IsNotNull(mailbox.ApplicableFlags);
        Assert.AreEqual(5, mailbox.ApplicableFlags.Count);
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Answered));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Flagged));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Draft));

        server.EnqueueTaggedResponse("UID FETCH 3 (UID FLAGS)\r\n",
                                     "* FETCH 3 (UID 3 FLAGS ())\r\n" +
                                     "$tag OK done\r\n");

        var message = mailbox.GetMessageByUid(3L, ImapMessageFetchAttributeOptions.DynamicAttributes);

        Assert.AreEqual(3L, message.Sequence);
        Assert.AreEqual(3L, message.Uid);
        Assert.IsNotNull(message.Flags);
        Assert.AreEqual(0, message.Flags.Count);
        Assert.IsFalse(message.IsSeen);

        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("NOOP\r\n",
                                       "* EXPUNGE 1\r\n" +
                                       "* EXPUNGE 5\r\n" +
                                       "* EXISTS 8\r\n" +
                                       "* RECENT 4\r\n" +
                                       "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft $label1)\r\n" +
                                       "* FETCH 2 (FLAGS (\\Seen \\Deleted $label1))\r\n" +
                                       "$tag OK done\r\n");
        });

        client.Refresh();

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        Assert.AreEqual(8L, mailbox.ExistMessageCount);
        Assert.AreEqual(4L, mailbox.RecentMessageCount);
        Assert.IsNotNull(mailbox.ApplicableFlags);
        Assert.AreEqual(6, mailbox.ApplicableFlags.Count);
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Answered));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Flagged));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains("$label1"));

        Assert.AreEqual(2L, message.Sequence);
        Assert.AreEqual(3L, message.Uid);
        Assert.IsNotNull(message.Flags);
        Assert.AreEqual(3, message.Flags.Count);
        Assert.IsTrue(message.Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(message.Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(message.Flags.Contains("$label1"));
        Assert.IsTrue(message.IsSeen);
        Assert.IsTrue(message.IsMarkedAsDeleted);
      });
    }

    [Test]
    public void TestRefresh_Bye()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueUntaggedByeResponse("* BYE\r\n");
        });

        Assert.Throws<Smdn.Net.Imap4.Protocol.ImapConnectionException>(() => client.Refresh());

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestRefresh_Disconnected()
    {
      TestUtils2.TestDisconnected(delegate(ImapPseudoServer2 server, ImapClient client) {
        Assert.Throws<Smdn.Net.Imap4.Protocol.ImapConnectionException>(() => client.Refresh());

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestRefresh_LoggedOut()
    {
      TestUtils2.TestLoggedOut(delegate(ImapPseudoServer2 server, ImapClient client) {
        Assert.Throws<InvalidOperationException>(() => client.Refresh());

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestCombineMailboxName_With2Names()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST INBOX \"\"\r\n",
                                       "* LIST (\\Noselect) \"/\" \"\"\r\n" +
                                       "$tag OK done\r\n");
        });

        Assert.AreEqual("INBOX/child1", client.CombineMailboxName("INBOX", "child1"));

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        // sends no command
        Assert.AreEqual("INBOX/child2", client.CombineMailboxName("INBOX", "child2"));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestCombineMailboxName_With2Names_NullOrEmpty()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        Assert.Throws<ArgumentNullException>(() => client.CombineMailboxName(null, "sub1"));
        Assert.Throws<ArgumentException>(() => client.CombineMailboxName(string.Empty, "sub1"));

        Assert.Throws<ArgumentNullException>(() => client.CombineMailboxName("INBOX", null));
        Assert.Throws<ArgumentException>(() => client.CombineMailboxName("INBOX", string.Empty));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestCombineMailboxName_With3Names()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST INBOX \"\"\r\n",
                                       "* LIST (\\Noselect) \"/\" \"\"\r\n" +
                                       "$tag OK done\r\n");
        });

        Assert.AreEqual("INBOX/dir/child1", client.CombineMailboxName("INBOX", "dir", "child1"));

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        // sends no command
        Assert.AreEqual("INBOX/dir/child2", client.CombineMailboxName("INBOX", "dir", "child2"));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestCombineMailboxName_With3Names_NullOrEmpty()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        Assert.Throws<ArgumentNullException>(() => client.CombineMailboxName(null, "sub1", "sub2"));
        Assert.Throws<ArgumentException>(() => client.CombineMailboxName(string.Empty, "sub1", "sub2"));

        Assert.Throws<ArgumentNullException>(() => client.CombineMailboxName("INBOX", null, "sub2"));
        Assert.Throws<ArgumentException>(() => client.CombineMailboxName("INBOX", string.Empty, "sub2"));

        Assert.Throws<ArgumentNullException>(() => client.CombineMailboxName("INBOX", "sub1", null));
        Assert.Throws<ArgumentException>(() => client.CombineMailboxName("INBOX", "sub1", string.Empty));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestCombineMailboxName_With4Names()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST INBOX \"\"\r\n",
                                       "* LIST (\\Noselect) \"/\" \"\"\r\n" +
                                       "$tag OK done\r\n");
        });

        Assert.AreEqual("INBOX/dir/subdir/child1", client.CombineMailboxName("INBOX", "dir", "subdir", "child1"));

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        // sends no command
        Assert.AreEqual("INBOX/dir/subdir/child2", client.CombineMailboxName("INBOX", "dir", "subdir", "child2"));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestCombineMailboxName_With4Names_NullOrEmpty()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        Assert.Throws<ArgumentNullException>(() => client.CombineMailboxName(null, "sub1", "sub2", "sub3"));
        Assert.Throws<ArgumentException>(() => client.CombineMailboxName(string.Empty, "sub1", "sub2", "sub3"));

        Assert.Throws<ArgumentNullException>(() => client.CombineMailboxName("INBOX", null, "sub2", "sub3"));
        Assert.Throws<ArgumentException>(() => client.CombineMailboxName("INBOX", string.Empty, "sub2", "sub3"));

        Assert.Throws<ArgumentNullException>(() => client.CombineMailboxName("INBOX", "sub1", null, "sub3"));
        Assert.Throws<ArgumentException>(() => client.CombineMailboxName("INBOX", "sub1", string.Empty, "sub3"));

        Assert.Throws<ArgumentNullException>(() => client.CombineMailboxName("INBOX", "sub1", "sub2", null));
        Assert.Throws<ArgumentException>(() => client.CombineMailboxName("INBOX", "sub1", "sub2", string.Empty));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestCombineMailboxName_WithArrayOfNames()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST INBOX \"\"\r\n",
                                       "* LIST (\\Noselect) \"/\" \"\"\r\n" +
                                       "$tag OK done\r\n");
        });

        Assert.AreEqual("INBOX/1/2/3/4/5/6", client.CombineMailboxName("INBOX", "1", "2", "3", "4", "5", "6"));

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        // sends no command
        Assert.AreEqual("INBOX/1/2/3/4/5/6", client.CombineMailboxName("INBOX", "1", "2", "3", "4", "5", "6"));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestCombineMailboxName_WithArrayOfNames_ContainsNullOrEmpty()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        Assert.Throws<ArgumentNullException>(() => client.CombineMailboxName(new string[] {null}));
        Assert.Throws<ArgumentException>(() => client.CombineMailboxName(new string[] {string.Empty}));

        Assert.Throws<ArgumentNullException>(() => client.CombineMailboxName(new string[] {null, "sub1", "sub2", "sub3"}));
        Assert.Throws<ArgumentException>(() => client.CombineMailboxName(new string[] {string.Empty, "sub1", "sub2", "sub3"}));

        Assert.Throws<ArgumentNullException>(() => client.CombineMailboxName(new string[] {"INBOX", null, "sub2", "sub3"}));
        Assert.Throws<ArgumentException>(() => client.CombineMailboxName(new string[] {"INBOX", string.Empty, "sub2", "sub3"}));

        Assert.Throws<ArgumentNullException>(() => client.CombineMailboxName(new string[] {"INBOX", "sub1", null, "sub3"}));
        Assert.Throws<ArgumentException>(() => client.CombineMailboxName(new string[] {"INBOX", "sub1", string.Empty, "sub3"}));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestCombineMailboxName_WithListOfNames()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var responseTask = TestUtils2.QueueResponseTask(server, client, delegate {
          server.EnqueueTaggedResponse("LIST INBOX \"\"\r\n",
                                       "* LIST (\\Noselect) \"/\" \"\"\r\n" +
                                       "$tag OK done\r\n");
        });

        var paths = new List<string>() {"INBOX", "1", "2", "3"};

        Assert.AreEqual("INBOX/1/2/3", client.CombineMailboxName(paths));

        responseTask.Wait();

        Assert.IsFalse(client.IsBusy);

        // sends no command
        Assert.AreEqual("INBOX/1/2/3", client.CombineMailboxName(paths));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestCombineMailboxName_WithArrayOfNames_Empty()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        Assert.Throws<ArgumentException>(() => client.CombineMailboxName(new string[0]));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestCombineMailboxName_WithArrayOfNames_Single()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        Assert.AreEqual("INBOX", client.CombineMailboxName(new string[] {"INBOX"}));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestCombineMailboxName_WithListOfNames_Empty()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var paths = new List<string>();

        Assert.Throws<ArgumentException>(() => client.CombineMailboxName(paths));

        Assert.IsFalse(client.IsBusy);
      });
    }

    [Test]
    public void TestCombineMailboxName_WithListOfNames_Single()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        var paths = new List<string>() {"INBOX"};

        Assert.AreEqual("INBOX", client.CombineMailboxName(paths));

        Assert.IsFalse(client.IsBusy);
      });
    }
  }
}
