using System;
using System.Collections.Generic;
using System.Linq;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapMailboxInfoTests {
    [Test]
    public void TestRefresh()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             "INBOX",
                             Refresh);
    }

    [Test]
    public void TestRefresh_CondStoreDisabled()
    {
      TestUtils2.TestMailbox(new[] {ImapCapability.CondStore},
                             new[] {ImapMailboxFlag.NoInferiors},
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        mailbox.Client.DisabledCapabilities.Add(ImapCapability.CondStore);

        Refresh(server, mailbox);
      });
    }

    private void Refresh(ImapPseudoServer2 server, ImapMailboxInfo mailbox)
    {
      var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
        server.EnqueueTaggedResponse("STATUS INBOX (MESSAGES RECENT UIDNEXT UIDVALIDITY UNSEEN)\r\n",
                                     "* STATUS INBOX ()\r\n" +
                                     "$tag OK done\r\n");
      });

      mailbox.Refresh();

      responseTask.Wait();

      Assert.IsFalse(mailbox.Client.IsBusy);
    }

    [Test]
    public void TestRefresh_ReRetrieveMailboxFlags()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" INBOX",
                                       "* LIST (\\Marked \\NoInferiors) \"/\" \"INBOX\"\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("STATUS INBOX (",
                                       "* STATUS INBOX ()\r\n" +
                                       "$tag OK done\r\n");
        });

        CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.NoInferiors);
        CollectionAssert.DoesNotContain(mailbox.Flags, ImapMailboxFlag.Marked);

        mailbox.Refresh(true);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.NoInferiors);
        CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.Marked);
      });
    }

    [Test]
    public void TestRefresh_ReRetrieveMailboxFlags_MailboxDeleted()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             "INBOX.child",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" INBOX.child",
                                       "$tag OK done\r\n");
        });

        Assert.IsTrue(mailbox.Exists);

        CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.NoInferiors);

        mailbox.Refresh(true);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsFalse(mailbox.Exists);

        CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.NoInferiors);
      });
    }

    [Test]
    public void TestRefresh_ReRetrieveMailboxFlags_MailboxReCreated()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             "INBOX.child",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("DELETE INBOX.child",
                                       "$tag OK done\r\n");
        });

        mailbox.Delete();

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsFalse(mailbox.Exists);

        server.EnqueueTaggedResponse("LIST \"\" INBOX.child",
                                     "* LIST (\\NoInferiors) \"/\" \"INBOX.child\"\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("STATUS INBOX.child (",
                                     "* STATUS INBOX.child ()\r\n" +
                                     "$tag OK done\r\n");

        mailbox.Refresh(true);

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsTrue(mailbox.Exists);

        CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.NoInferiors);
      });
    }

    [Test]
    public void TestRefresh_CondStoreCapable()
    {
      TestUtils2.TestMailbox(new[] {ImapCapability.CondStore},
                             new[] {ImapMailboxFlag.NoInferiors},
                             "INBOX.child",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("STATUS INBOX.child (MESSAGES RECENT UIDNEXT UIDVALIDITY UNSEEN HIGHESTMODSEQ)\r\n",
                                       "* STATUS INBOX.child ()\r\n" +
                                       "$tag OK done\r\n");
        });

        mailbox.Refresh();

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestRefresh_NonExistentMailbox()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NonExistent},
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Exists);

        // do nothing
        mailbox.Refresh();

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestRefresh_UnselectableMailbox()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoSelect},
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.IsUnselectable);

        // do nothing
        mailbox.Refresh();

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpen()
    {
      Open(false);
    }

    [Test]
    public void TestOpen_AsReadOnly()
    {
      Open(true);
    }

    private void Open(bool asReadOnly)
    {
      TestUtils2.TestMailbox("INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse(asReadOnly ? "EXAMINE INBOX\r\n" : "SELECT INBOX\r\n",
                                       "* 172 EXISTS\r\n" +
                                       "* 1 RECENT\r\n" +
                                       "* OK [UNSEEN 12] Message 12 is first unseen\r\n" +
                                       "* OK [UIDVALIDITY 3857529045] UIDs valid\r\n" +
                                       "* OK [UIDNEXT 4392] Predicted next UID\r\n" +
                                       "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
                                       "* OK [PERMANENTFLAGS (\\Deleted \\Seen \\*)] Limited\r\n" +
                                       string.Format("$tag OK [{0}] done\r\n", asReadOnly ? "READ-ONLY" : "READ-WRITE"));
        });

        Assert.IsNull(mailbox.Client.OpenedMailbox);
        Assert.IsFalse(mailbox.IsOpen);

        using (var opened = asReadOnly ? mailbox.Open(true) : mailbox.Open()) {
          responseTask.Wait();

          Assert.IsFalse(mailbox.Client.IsBusy);

          try {
            Assert.IsNotNull(opened);
            Assert.IsTrue(opened.IsOpen);

            Assert.AreEqual(172L, opened.ExistMessageCount);
            Assert.AreEqual(1L, opened.RecentMessageCount);
            Assert.AreEqual(12L, opened.FirstUnseenMessageNumber);
            Assert.AreEqual(4392L, opened.NextUid);

            Assert.AreSame(opened, mailbox.Client.OpenedMailbox);
            Assert.IsTrue(mailbox.IsOpen);
            Assert.AreEqual(opened.ExistMessageCount, mailbox.ExistMessageCount);
            Assert.AreEqual(opened.UnseenMessageCount, mailbox.UnseenMessageCount);
            Assert.AreEqual(opened.NextUid, mailbox.NextUid);

            Assert.IsNotNull(opened.ApplicableFlags);
            Assert.AreEqual(5, opened.ApplicableFlags.Count);

            Assert.IsNotNull(opened.PermanentFlags);
            Assert.AreEqual(3, opened.PermanentFlags.Count);

            Assert.IsTrue(opened.IsAllowedToCreateKeywords);
            Assert.IsTrue(opened.IsUidPersistent);

            if (asReadOnly)
              Assert.IsTrue(opened.IsReadOnly);
            else
              Assert.IsFalse(opened.IsReadOnly);
          }
          finally {
            server.EnqueueTaggedResponse("CLOSE\r\n",
                                         "$tag OK done\r\n");
          }
        }

        Assert.IsNull(mailbox.Client.OpenedMailbox);
        Assert.IsFalse(mailbox.IsOpen);
      });
    }

    [Test]
    [TestCase(true)]
    [TestCase(false)]
    public void TestOpen_Unselectable(bool asReadOnly)
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoSelect},
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        if (asReadOnly)
          Assert.Throws<ImapProtocolViolationException>(() => mailbox.Open(true));
        else
          Assert.Throws<ImapProtocolViolationException>(() => mailbox.Open());

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNull(mailbox.Client.OpenedMailbox);
        Assert.IsFalse(mailbox.IsOpen);
      });
    }

    [Test, Ignore("not implemented; reopen")]
    public void TestOpen_ReopenItself()
    {
    }

    [Test, Ignore("not implemented; reopen")]
    public void TestOpen_OpenAnotherNewly()
    {
    }

    [Test]
    public void TestDelete()
    {
      TestUtils2.TestMailbox("Trash", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("DELETE Trash\r\n",
                                       "$tag OK done\r\n");
        });

        Assert.IsTrue(mailbox.Exists);

        mailbox.Delete();

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsFalse(mailbox.Exists);
      });
    }

    [Test]
    public void TestDelete_Unsubscribe()
    {
      TestUtils2.TestMailbox("Trash", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("DELETE Trash\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("UNSUBSCRIBE Trash\r\n",
                                       "$tag OK done\r\n");
        });

        Assert.IsTrue(mailbox.Exists);

        mailbox.Delete(true);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsFalse(mailbox.Exists);
      });
    }

    [Test]
    public void TestDelete_DeleteOpened()
    {
      TestUtils2.TestMailbox("Trash", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("SELECT Trash\r\n",
                                       "$tag OK done\r\n");
        });

        var opened = mailbox.Open();

        Assert.AreSame(opened, mailbox.Client.OpenedMailbox);
        Assert.IsTrue(mailbox.IsOpen);
        Assert.IsTrue(opened.IsOpen);

        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("DELETE Trash\r\n",
                                     "$tag OK done\r\n");

        Assert.IsTrue(opened.Exists);

        opened.Delete();

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNull(mailbox.Client.OpenedMailbox);
        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsFalse(opened.IsOpen);
        Assert.IsFalse(opened.Exists);
      });
    }

    [Test]
    public void TestDelete_MailboxNonExistent()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NonExistent},
                             "Trash",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Exists);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("UNSUBSCRIBE Trash\r\n",
                                       "$tag OK done\r\n");
        });

        mailbox.Delete(true);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsFalse(mailbox.Exists);
      });
    }

    [Test]
    public void TestDelete_NonExistent()
    {
      DeleteNonExistent(false);
    }

    [Test]
    public void TestDelete_NonExistent_Unsubscribe()
    {
      DeleteNonExistent(true);
    }

    private void DeleteNonExistent(bool unsubscribe)
    {
      TestUtils2.TestMailbox("Trash",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.Exists);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          // someone deletes 'Trash'
          // C2: TAG CREATE "Trash"
          // S2: TAG OK deleted "Trash"

          server.EnqueueTaggedResponse("DELETE Trash\r\n",
                                       "$tag NO [NONEXISTENT] done\r\n");

          if (unsubscribe)
            server.EnqueueTaggedResponse("UNSUBSCRIBE Trash\r\n",
                                         "$tag OK done\r\n");
        });

        mailbox.Delete(unsubscribe);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsFalse(mailbox.Exists);
      });
    }

    [Test]
    public void TestCreate_RecreateDeleted()
    {
      CreateRecreateDeleted(false);
    }

    [Test]
    public void TestCreate_RecreateDeleted_Subscribe()
    {
      CreateRecreateDeleted(true);
    }

    private void CreateRecreateDeleted(bool subscribe)
    {
      TestUtils2.TestMailbox("Trash", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.Exists);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("DELETE Trash\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("UNSUBSCRIBE Trash\r\n",
                                       "$tag OK done\r\n");
        });

        mailbox.Delete(true);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Exists);

        server.EnqueueTaggedResponse("CREATE Trash\r\n",
                                     "$tag OK done\r\n");

        if (subscribe)
          server.EnqueueTaggedResponse("SUBSCRIBE Trash\r\n",
                                       "$tag OK done\r\n");

        if (subscribe)
          mailbox.Create(true);
        else
          mailbox.Create();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsTrue(mailbox.Exists);
      });
    }

    [Test]
    public void TestCreate_NonExistent()
    {
      CreateNonExistent(false);
    }

    [Test]
    public void TestCreate_NonExistent_Subscribe()
    {
      CreateNonExistent(true);
    }

    private void CreateNonExistent(bool subscribe)
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NonExistent},
                             "deleted",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Exists);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("CREATE deleted\r\n",
                                       "$tag OK done\r\n");

          if (subscribe)
            server.EnqueueTaggedResponse("SUBSCRIBE deleted\r\n",
                                         "$tag OK done\r\n");
        });

        if (subscribe)
          mailbox.Create(true);
        else
          mailbox.Create();

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsTrue(mailbox.Exists);
      });
    }

    [Test]
    public void TestCreate_AlreadyExists()
    {
      CreateAlreadyExists(false);
    }

    [Test]
    public void TestCreate_AlreadyExists_Subscribe()
    {
      CreateAlreadyExists(true);
    }

    private void CreateAlreadyExists(bool subscribe)
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NonExistent},
                             "Sent",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Exists);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          // someone creates 'Sent'
          // C2: TAG CREATE "Sent"
          // S2: TAG OK created "Sent"

          server.EnqueueTaggedResponse("CREATE Sent\r\n",
                                       "$tag NO [ALREADYEXISTS] Mailbox 'Sent' already exists\r\n");

          if (subscribe)
            server.EnqueueTaggedResponse("SUBSCRIBE Sent\r\n",
                                         "$tag OK done\r\n");
        });

        if (subscribe)
          mailbox.Create(true);
        else
          mailbox.Create();

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsTrue(mailbox.Exists);
      });
    }

    [Test]
    public void TestMoveTo_MoveToInbox()
    {
      MoveToInbox(false);
    }

    [Test]
    public void TestMoveTo_MoveToInbox_Subscribe()
    {
      MoveToInbox(true);
    }

    private void MoveToInbox(bool subscribe)
    {
      TestUtils2.TestMailbox("/", "Draft", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("Draft", mailbox.Name);
        Assert.AreEqual("Draft", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);
        Assert.IsFalse(mailbox.IsInbox);

        if (subscribe)
          Assert.Throws<ImapProtocolViolationException>(() => mailbox.MoveTo("INBOX", true));
        else
          Assert.Throws<ImapProtocolViolationException>(() => mailbox.MoveTo("INBOX"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestMoveTo_MoveInbox()
    {
      MoveToMoveInbox(false);
    }

    [Test]
    public void TestMoveTo_MoveInbox_Subscribe()
    {
      MoveToMoveInbox(true);
    }

    private void MoveToMoveInbox(bool subscribe)
    {
      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX", mailbox.Name);
        Assert.AreEqual("INBOX", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);
        Assert.IsTrue(mailbox.IsInbox);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("RENAME INBOX INBOX/backup\r\n",
                                       "$tag OK done\r\n");

          if (subscribe)
            server.EnqueueTaggedResponse("SUBSCRIBE INBOX/backup\r\n",
                                         "$tag OK done\r\n");
        });

        if (subscribe)
          mailbox.MoveTo("INBOX/backup", true);
        else
          mailbox.MoveTo("INBOX/backup");

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("backup", mailbox.Name);
        Assert.AreEqual("INBOX/backup", mailbox.FullName);
        Assert.AreEqual("INBOX", mailbox.ParentMailboxName);
      });
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox()
    {
      MoveToDestinationMailbox(false);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_Subscribe()
    {
      MoveToDestinationMailbox(true);
    }

    private void MoveToDestinationMailbox(bool subscribe)
    {
      TestUtils2.TestMailbox("/", "src", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("src", mailbox.Name);
        Assert.AreEqual("src", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);

        server.EnqueueTaggedResponse("LIST \"\" dest\r\n",
                                     "* LIST () \"/\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = mailbox.Client.GetMailbox("dest");

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          if (subscribe) {
            server.EnqueueTaggedResponse("LSUB src/ *",
                                         "* LSUB () \"/\" src/sub1\r\n" +
                                         "* LSUB () \"/\" src/sub2/sub3\r\n" +
                                         "* LSUB () \"/\" src/sub4/sub5/sub6\r\n" +
                                         "$tag OK done\r\n");
          }

          server.EnqueueTaggedResponse("RENAME src dest/src\r\n",
                                       "$tag OK done\r\n");

          if (subscribe) {
            server.EnqueueTaggedResponse("UNSUBSCRIBE src\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE dest/src\r\n",
                                         "$tag OK done\r\n");

            server.EnqueueTaggedResponse("UNSUBSCRIBE src/sub1\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE dest/src/sub1\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("UNSUBSCRIBE src/sub2/sub3\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE dest/src/sub2/sub3\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("UNSUBSCRIBE src/sub4/sub5/sub6\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE dest/src/sub4/sub5/sub6\r\n",
                                         "$tag OK done\r\n");
          }
        });

        if (subscribe)
          mailbox.MoveTo(destMailbox, true);
        else
          mailbox.MoveTo(destMailbox);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("src", mailbox.Name);
        Assert.AreEqual("dest/src", mailbox.FullName);
        Assert.AreEqual("dest", mailbox.ParentMailboxName);
      });
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_DestinationNonHierarchical()
    {
      TestUtils2.TestMailbox(null,
                             null,
                             string.Empty,
                             "dest",
                             MoveToDestinationMailboxDestinationCanNotHaveChild);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_DestinationNoInferiors()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             ".",
                             "dest",
                             MoveToDestinationMailboxDestinationCanNotHaveChild);
    }

    private void MoveToDestinationMailboxDestinationCanNotHaveChild(ImapPseudoServer2 server, ImapMailboxInfo destMailbox)
    {
      server.EnqueueTaggedResponse("LIST \"\" src\r\n",
                                   "* LIST () \"/\" src\r\n" +
                                   "$tag OK done\r\n");

      var sourceMailbox = destMailbox.Client.GetMailbox("src");

      Assert.Throws<ImapProtocolViolationException>(() => sourceMailbox.MoveTo(destMailbox));

      Assert.IsFalse(destMailbox.Client.IsBusy);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_AlreadyExists()
    {
      MoveToDestinationMailboxAlreadyExists(false);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_AlreadyExists_Subscribe()
    {
      MoveToDestinationMailboxAlreadyExists(true);
    }

    private void MoveToDestinationMailboxAlreadyExists(bool subscribe)
    {
      TestUtils2.TestMailbox("/", "src", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("src", mailbox.Name);
        Assert.AreEqual("src", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);

        server.EnqueueTaggedResponse("LIST \"\" dest\r\n",
                                     "* LIST () \"/\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = mailbox.Client.GetMailbox("dest");

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          if (subscribe) {
            server.EnqueueTaggedResponse("LSUB src/ *",
                                         "* LSUB () \"/\" src/sub1\r\n" +
                                         "* LSUB () \"/\" src/sub2/sub3\r\n" +
                                         "$tag OK done\r\n");
          }

          server.EnqueueTaggedResponse("RENAME src dest/src\r\n",
                                       "$tag NO [ALREADYEXISTS] mailbox exists\r\n");
        });

        var ex = Assert.Throws<ImapErrorResponseException>(() => {
          if (subscribe)
            mailbox.MoveTo(destMailbox, true);
          else
            mailbox.MoveTo(destMailbox);
        });

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(ImapResponseCode.AlreadyExists,
                        ex.Result.TaggedStatusResponse.ResponseText.Code);

        Assert.AreEqual("src", mailbox.Name);
        Assert.AreEqual("src", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);
      });
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_MailboxOfDifferentSession()
    {
      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 serverSrc, ImapMailboxInfo mailboxSrc) {
        Assert.AreEqual("INBOX", mailboxSrc.Name);
        Assert.AreEqual("INBOX", mailboxSrc.FullName);
        Assert.AreEqual(string.Empty, mailboxSrc.ParentMailboxName);
        Assert.IsTrue(mailboxSrc.IsInbox);

        TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 serverDest, ImapMailboxInfo mailboxDest) {
          Assert.AreEqual("INBOX", mailboxDest.Name);
          Assert.AreEqual("INBOX", mailboxDest.FullName);
          Assert.AreEqual(string.Empty, mailboxDest.ParentMailboxName);
          Assert.IsTrue(mailboxDest.IsInbox);

          Assert.AreNotSame(mailboxSrc.Client, mailboxDest.Client);

          Assert.Throws<NotSupportedException>(() => mailboxSrc.MoveTo(mailboxDest));

          Assert.IsFalse(mailboxSrc.Client.IsBusy);
        });
      });
    }

    [Test]
    public void TestMoveTo_MoveToNewName_ContainsWildcard()
    {
      TestUtils2.TestMailbox(".", "Box", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("RENAME Box \"Box*\"\r\n",
                                       "$tag OK done\r\n");
        });

        mailbox.MoveTo("Box*");

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("Box*", mailbox.Name);
        Assert.AreEqual("Box*", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);
      });
    }

    [Test]
    public void TestMoveTo_MoveToNewName_Null()
    {
      TestUtils2.TestMailbox(".", "Box", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.MoveTo((string)null));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestMoveTo_MoveToNewName_Empty()
    {
      TestUtils2.TestMailbox(".", "Box", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentException>(() => mailbox.MoveTo(string.Empty));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestMoveTo_MoveToSameName1()
    {
      TestUtils2.TestMailbox("/", "Draft", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("Draft", mailbox.Name);
        Assert.AreEqual("Draft", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);
        Assert.IsFalse(mailbox.IsInbox);

        Assert.Throws<ImapProtocolViolationException>(() => mailbox.MoveTo("Draft"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestMoveTo_MoveToSameName2()
    {
      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX", mailbox.Name);
        Assert.AreEqual("INBOX", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);
        Assert.IsTrue(mailbox.IsInbox);

        Assert.Throws<ImapProtocolViolationException>(() => mailbox.MoveTo("inbox"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestMoveTo_MoveToNewName_Deepen()
    {
      MoveToNewNameDeepen(false);
    }

    [Test]
    public void TestMoveTo_MoveToNewName_Deepen_Subscribe()
    {
      MoveToNewNameDeepen(true);
    }

    private void MoveToNewNameDeepen(bool subscribe)
    {
      TestUtils2.TestMailbox(new[] {ImapCapability.ListExtended},
                             null,
                             ".",
                             "src",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          if (subscribe) {
            server.EnqueueTaggedResponse("LIST (SUBSCRIBED) src. *",
                                         "* LIST (\\Subscribed) \".\" src.sub1\r\n" +
                                         "* LIST (\\Subscribed \\NonExistent) \".\" src.sub2\r\n" +
                                         "$tag OK done\r\n");
          }

          server.EnqueueTaggedResponse("RENAME src Trash.src\r\n",
                                       "$tag OK done\r\n");

          if (subscribe) {
            server.EnqueueTaggedResponse("UNSUBSCRIBE src\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE Trash.src\r\n",
                                         "$tag OK done\r\n");

            server.EnqueueTaggedResponse("UNSUBSCRIBE src.sub1\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE Trash.src.sub1\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("UNSUBSCRIBE src.sub2\r\n",
                                         "$tag OK done\r\n");
          }
        });

        if (subscribe)
          mailbox.MoveTo("Trash.src", true);
        else
          mailbox.MoveTo("Trash.src");

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("src", mailbox.Name);
        Assert.AreEqual("Trash.src", mailbox.FullName);
        Assert.AreEqual("Trash", mailbox.ParentMailboxName);
      });
    }

    [Test]
    public void TestMoveTo_MoveToNewName_MakeShallow()
    {
      MoveToNewNameShallow(false);
    }

    [Test]
    public void TestMoveTo_MoveToNewName_MakeShallow_Subscribe()
    {
      MoveToNewNameShallow(true);
    }

    private void MoveToNewNameShallow(bool subscribe)
    {
      TestUtils2.TestMailbox("/",
                             "nested/src",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("src", mailbox.Name);
        Assert.AreEqual("nested/src", mailbox.FullName);
        Assert.AreEqual("nested", mailbox.ParentMailboxName);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          if (subscribe) {
            server.EnqueueTaggedResponse("LSUB nested/src/ *",
                                         "* LSUB () \"/\" nested/src/sub1\r\n" +
                                         "* LSUB () \"/\" nested/src/sub2/sub3\r\n" +
                                         "$tag OK done\r\n");
          }

          server.EnqueueTaggedResponse("RENAME nested/src dest\r\n",
                                       "$tag OK done\r\n");

          if (subscribe) {
            server.EnqueueTaggedResponse("UNSUBSCRIBE nested/src\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE dest\r\n",
                                         "$tag OK done\r\n");

            server.EnqueueTaggedResponse("UNSUBSCRIBE nested/src/sub1\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE dest/sub1\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("UNSUBSCRIBE nested/src/sub2/sub3\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE dest/sub2/sub3\r\n",
                                         "$tag OK done\r\n");
          }
        });

        if (subscribe)
          mailbox.MoveTo("dest", true);
        else
          mailbox.MoveTo("dest");

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("dest", mailbox.Name);
        Assert.AreEqual("dest", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);
      });
    }

    [Test]
    public void TestMoveTo_MoveToNewName_Opened()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                     "$tag OK done\r\n");

        var opened = mailbox.Open();

        Assert.AreSame(opened, mailbox.Client.OpenedMailbox);
        Assert.IsTrue(mailbox.IsOpen);
        Assert.IsTrue(opened.IsOpen);
        Assert.AreEqual("INBOX", opened.FullName);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("CLOSE\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("RENAME INBOX Trash.INBOX\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("SELECT Trash.INBOX\r\n",
                                       "$tag OK done\r\n");
        });

        mailbox.MoveTo("Trash.INBOX");

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("INBOX", mailbox.Name);
        Assert.AreEqual("Trash.INBOX", mailbox.FullName);
        Assert.AreEqual("Trash", mailbox.ParentMailboxName);

        Assert.IsNull(mailbox.Client.OpenedMailbox);
        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsFalse(opened.IsOpen);
      });
    }

    [Test]
    public void TestMoveTo_MoveToNewName_Subscribe_HasNoChildren()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.HasNoChildren},
                             ".",
                             "Draft",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LSUB Draft. *\r\n",
                                       "* LSUB () \".\" Draft.Created\r\n" + // other client created children
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("RENAME Draft Sent\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("UNSUBSCRIBE Draft\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("SUBSCRIBE Sent\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("UNSUBSCRIBE Draft.Created\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("SUBSCRIBE Sent.Created\r\n",
                                       "$tag OK done\r\n");
        });

        mailbox.MoveTo("Sent", true);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("Sent", mailbox.Name);
        Assert.AreEqual("Sent", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);
      });
    }

    [Test]
    public void TestMoveTo_MoveToNewName_Subscribe_NonHierarchical()
    {
      TestUtils2.TestMailbox(null,
                             null,
                             string.Empty,
                             "Sent",
                             MoveToNewNameSubscribeCanNotHaveChild);
    }

    [Test]
    public void TestMoveTo_MoveToNewName_Subscribe_NoInferiors()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             ".",
                             "Sent",
                             MoveToNewNameSubscribeCanNotHaveChild);
    }

    private void MoveToNewNameSubscribeCanNotHaveChild(ImapPseudoServer2 server, ImapMailboxInfo mailbox)
    {
      var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
        server.EnqueueTaggedResponse("RENAME Sent OldSent\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("UNSUBSCRIBE Sent\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("SUBSCRIBE OldSent\r\n",
                                     "$tag OK done\r\n");
      });

      mailbox.MoveTo("OldSent", true);

      responseTask.Wait();

      Assert.IsFalse(mailbox.Client.IsBusy);
    }

    [Test]
    public void TestMoveTo_MoveToNewName_AlreadyExists()
    {
      MoveToNewNameAlreadyExists(false);
    }

    [Test]
    public void TestMoveTo_MoveToNewName_AlreadyExists_Subscribe()
    {
      MoveToNewNameAlreadyExists(true);
    }

    private void MoveToNewNameAlreadyExists(bool subscribe)
    {
      TestUtils2.TestMailbox(new[] {ImapCapability.ListExtended},
                             null,
                             ".",
                             "src",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("src", mailbox.Name);
        Assert.AreEqual("src", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          if (subscribe) {
            server.EnqueueTaggedResponse("LIST (SUBSCRIBED) src. * RETURN (CHILDREN)\r\n",
                                         "* LIST (\\Subscribed) \".\" src.sub1\r\n" +
                                         "* LIST (\\Subscribed \\NonExistent) \".\" src.sub2\r\n" +
                                         "$tag OK done\r\n");
          }

          server.EnqueueTaggedResponse("RENAME src Trash.src\r\n",
                                       "$tag NO [ALREADYEXISTS] mailbox exists\r\n");
        });

        var ex = Assert.Throws<ImapErrorResponseException>(() => {
          if (subscribe)
            mailbox.MoveTo("Trash.src", true);
          else
            mailbox.MoveTo("Trash.src");
        });

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(ImapResponseCode.AlreadyExists,
                        ex.Result.TaggedStatusResponse.ResponseText.Code);

        Assert.AreEqual("src", mailbox.Name);
        Assert.AreEqual("src", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);
      });
    }

    [Test]
    public void TestSubscribe()
    {
      Subscribe(false);
    }

    [Test]
    public void TestSubscribe_Recursive()
    {
      Subscribe(true);
    }

    private void Subscribe(bool recursive)
    {
      TestUtils2.TestMailbox("INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("SUBSCRIBE INBOX\r\n",
                                       "$tag OK done\r\n");

          if (recursive) {
            server.EnqueueTaggedResponse("LIST INBOX. *\r\n",
                                         "* LIST () \".\" INBOX.Child1\r\n" +
                                         "* LIST () \".\" INBOX.Child2\r\n" +
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE INBOX.Child1\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE INBOX.Child2\r\n",
                                         "$tag OK done\r\n");
          }
        });

        if (recursive)
          mailbox.Subscribe(true);
        else
          mailbox.Subscribe();

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestUnsubscribe()
    {
      Unsubscribe(false);
    }

    [Test]
    public void TestUnsubscribe_Recursive()
    {
      Unsubscribe(true);
    }

    private void Unsubscribe(bool recursive)
    {
      TestUtils2.TestMailbox("INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("UNSUBSCRIBE INBOX\r\n",
                                       "$tag OK done\r\n");

          if (recursive) {
            server.EnqueueTaggedResponse("LSUB INBOX. *\r\n",
                                         "* LSUB () \".\" INBOX.Child1\r\n" +
                                         "* LSUB () \".\" INBOX.Child2\r\n" +
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("UNSUBSCRIBE INBOX.Child1\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("UNSUBSCRIBE INBOX.Child2\r\n",
                                         "$tag OK done\r\n");
          }
        });

        if (recursive)
          mailbox.Unsubscribe(true);
        else
          mailbox.Unsubscribe();

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetChildren()
    {
      GetChildren(new ImapMailboxFlag[0]);
    }

    [Test]
    public void TestGetChildren_HasNoChildren()
    {
      GetChildren(new[] {ImapMailboxFlag.HasNoChildren});
    }

    private void GetChildren(ImapMailboxFlag[] flags)
    {
      TestUtils2.TestMailbox(null, flags, "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LIST INBOX. *\r\n",
                                       "* LIST () \".\" INBOX.Child1\r\n" +
                                       "* LIST () \".\" INBOX.Child2\r\n" +
                                       "* LIST () \".\" INBOX.Child2.Child3\r\n" +
                                       "$tag OK done\r\n");
        });

        var children = mailbox.GetChildren();

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(children);
        Assert.AreEqual(3, children.Count, "Count");
#if !NET_4_5
        Assert.Throws<NotSupportedException>(() => children.Clear(), "readonly");
#endif

        using (var enumerator = children.GetEnumerator()) {
          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreEqual("Child1", enumerator.Current.Name);
          Assert.AreEqual("INBOX.Child1", enumerator.Current.FullName);

          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreEqual("Child2", enumerator.Current.Name);
          Assert.AreEqual("INBOX.Child2", enumerator.Current.FullName);

          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreEqual("Child3", enumerator.Current.Name);
          Assert.AreEqual("INBOX.Child2.Child3", enumerator.Current.FullName);

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetChildren_NonHierarchical()
    {
      TestUtils2.TestMailbox(null,
                             null,
                             string.Empty,
                             "INBOX",
                             GetChildrenNoInferiorsOrNonHierarchical);
    }

    [Test]
    public void TestGetChildren_NoInferiors()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             ".",
                             "INBOX",
                             GetChildrenNoInferiorsOrNonHierarchical);
    }

    private void GetChildrenNoInferiorsOrNonHierarchical(ImapPseudoServer2 server, ImapMailboxInfo mailbox)
    {
      var children = mailbox.GetChildren();

      Assert.IsFalse(mailbox.Client.IsBusy);

      Assert.IsNotNull(children);
      Assert.AreEqual(0, children.Count, "Count");
#if !NET_4_5
      Assert.Throws<NotSupportedException>(() => children.Clear(), "readonly");
#endif

      using (var enumerator = children.GetEnumerator()) {
        Assert.IsFalse(enumerator.MoveNext());
      }
    }

    [Test]
    public void TestGetParent()
    {
      TestUtils2.TestMailbox(".", "INBOX.Child", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                       "* LIST () \".\" INBOX\r\n" +
                                       "$tag OK done\r\n");
        });

        var parent = mailbox.GetParent();

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(parent);
        Assert.AreEqual("INBOX", parent.Name);
        Assert.AreEqual("INBOX", parent.FullName);
      });
    }

    [Test]
    public void TestGetParent_OfTopLevelMailbox()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetParent());

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetParent_NotFound()
    {
      TestUtils2.TestMailbox(".", "Sent.Child", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" Sent\r\n",
                                       "$tag OK done\r\n");
        });

        var ex = Assert.Throws<ImapMailboxNotFoundException>(() => {
          mailbox.GetParent();
          Assert.Fail("ImapMailboxNotFoundException not thrown");
        });

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(ex.Mailbox);
        Assert.AreEqual("Sent", ex.Mailbox);

        Smdn.Net.TestUtils.SerializeBinary(ex, delegate(ImapMailboxNotFoundException deserialized) {
          Assert.IsNotNull(deserialized.Mailbox);
          Assert.AreEqual(ex.Mailbox, deserialized.Mailbox);
        });
      });
    }

    [Test]
    public void TestGetOrCreateParent()
    {
      GetOrCreateParent(false);
    }

    [Test]
    public void TestGetOrCreateParent_Subscribe()
    {
      GetOrCreateParent(true);
    }

    private void GetOrCreateParent(bool subscribe)
    {
      TestUtils2.TestMailbox(".", "Sent.Child", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" Sent\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("CREATE Sent\r\n",
                                       "$tag OK done\r\n");
          if (subscribe)
            server.EnqueueTaggedResponse("SUBSCRIBE Sent\r\n",
                                         "$tag OK done\r\n");
          server.EnqueueTaggedResponse("LIST \"\" Sent\r\n",
                                       "* LIST () \".\" Sent\r\n" +
                                       "$tag OK done\r\n");
        });

        var parent = subscribe
          ? mailbox.GetOrCreateParent(true)
          : mailbox.GetOrCreateParent();

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(parent);
        Assert.AreEqual("Sent", parent.Name);
        Assert.AreEqual("Sent", parent.FullName);
      });
    }

    [Test]
    public void TestGetOrCreateParent_OfTopLevelMailbox()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetOrCreateParent());

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetFullNameOf()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX.Child", mailbox.GetFullNameOf("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX/Child", mailbox.GetFullNameOf("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetFullNameOf_ArgumentEmpty()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf(string.Empty));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetFullNameOf_ArgumentNull()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf((string)null));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetFullNameOf_CanNotHaveChild()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             ".",
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetFullNameOf("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetFullNameOf_With2Inferiors()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX.1.2", mailbox.GetFullNameOf("1", "2"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX/1/2", mailbox.GetFullNameOf("1", "2"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             ".",
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetFullNameOf("1", "2"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf(null, "2"));
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf("1", null));

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf(string.Empty, "2"));
        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf("1", string.Empty));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetFullNameOf_With3Inferiors()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX.1.2.3", mailbox.GetFullNameOf("1", "2", "3"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX/1/2/3", mailbox.GetFullNameOf("1", "2", "3"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             ".",
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetFullNameOf("1", "2", "3"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf(null, "2", "3"));
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf("1", null, "3"));
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf("1", "2", null));

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf(string.Empty, "2", "3"));
        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf("1", string.Empty, "3"));
        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf("1", "2", string.Empty));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetFullNameOf_WithArrayOfInferiors()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX.1.2.3.4", mailbox.GetFullNameOf(new[] {"1", "2", "3", "4"}));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX/1/2/3/4", mailbox.GetFullNameOf(new[] {"1", "2", "3", "4"}));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox(null,
                  new[] {ImapMailboxFlag.NoInferiors},
                  ".",
                  "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetFullNameOf(new[] {"1", "2", "3", "4"}));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf(new string[] {null, "2", "3", "4"}));
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf(new string[] {"1", null, "3", "4"}));
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf(new string[] {"1", "2", null, "4"}));
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf(new string[] {"1", "2", "3", null}));

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf(new string[] {string.Empty, "2", "3", "4"}));
        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf(new string[] {"1", string.Empty, "3", "4"}));
        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf(new string[] {"1", "2", string.Empty, "4"}));
        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf(new string[] {"1", "2", "3", string.Empty}));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetFullNameOf_WithListOfInferiors()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX.1.2.3.4", mailbox.GetFullNameOf(new List<string>() {"1", "2", "3", "4"}));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX/1/2/3/4", mailbox.GetFullNameOf(new List<string>() {"1", "2", "3", "4"}));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             ".",
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetFullNameOf(new List<string>() {"1", "2", "3", "4"}));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf(new List<string>() {null, "2", "3", "4"}));
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf(new List<string>() {"1", null, "3", "4"}));
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf(new List<string>() {"1", "2", null, "4"}));
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf(new List<string>() {"1", "2", "3", null}));

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf(new List<string>() {string.Empty, "2", "3", "4"}));
        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf(new List<string>() {"1", string.Empty, "3", "4"}));
        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf(new List<string>() {"1", "2", string.Empty, "4"}));
        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf(new List<string>() {"1", "2", "3", string.Empty}));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestSplitMailboxHierarchy()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        CollectionAssert.AreEqual(new[] {"INBOX"},
                                  mailbox.SplitMailboxHierarchy("INBOX"));
        CollectionAssert.AreEqual(new[] {"INBOX", "1", "2"},
                                  mailbox.SplitMailboxHierarchy("INBOX.1.2"));
        CollectionAssert.AreEqual(new[] {"INBOX/1/2"},
                                  mailbox.SplitMailboxHierarchy("INBOX/1/2"));

        Assert.Throws<ArgumentException>(() => mailbox.SplitMailboxHierarchy(string.Empty));

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.Throws<ArgumentNullException>(() => mailbox.SplitMailboxHierarchy(null));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        CollectionAssert.AreEqual(new[] {"INBOX"},
                                  mailbox.SplitMailboxHierarchy("INBOX"));
        CollectionAssert.AreEqual(new[] {"INBOX.1.2"},
                                  mailbox.SplitMailboxHierarchy("INBOX.1.2"));
        CollectionAssert.AreEqual(new[] {"INBOX", "1", "2"},
                                  mailbox.SplitMailboxHierarchy("INBOX/1/2"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             ".",
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.SplitMailboxHierarchy("INBOX.1.2"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestCreateChild()
    {
      CreateChild(false);
    }

    [Test]
    public void TestCreateChild_Subscribe()
    {
      CreateChild(true);
    }

    private void CreateChild(bool subscribe)
    {
      TestUtils2.TestMailbox(".", "Sent", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("CREATE Sent.Child\r\n",
                                       "$tag OK done\r\n");
          if (subscribe)
            server.EnqueueTaggedResponse("SUBSCRIBE Sent.Child\r\n",
                                         "$tag OK done\r\n");
          server.EnqueueTaggedResponse("LIST \"\" Sent.Child\r\n",
                                       "* LIST () \".\" Sent.Child\r\n" +
                                       "$tag OK done\r\n");
        });

        var child = subscribe
          ? mailbox.CreateChild("Child", true)
          : mailbox.CreateChild("Child");

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(child);
        Assert.AreEqual("Child", child.Name);
        Assert.AreEqual("Sent.Child", child.FullName);
      });
    }

    [Test]
    public void TestCreateChild_NameEmpty()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentException>(() => mailbox.CreateChild(string.Empty));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestCreateChild_NameNull()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.CreateChild((string)null));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestCreateChild_AlreadyExists()
    {
      CreateChildAlreadyExists(false);
    }

    [Test]
    public void TestCreateChild_AlreadyExists_Subscribe()
    {
      CreateChildAlreadyExists(true);
    }

    private void CreateChildAlreadyExists(bool subscribe)
    {
      TestUtils2.TestMailbox(".", "Sent", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("CREATE Sent.Child\r\n",
                                       "$tag NO [ALREADYEXISTS] mailbox already exists\r\n");
        });

        var ex = Assert.Throws<ImapErrorResponseException>(() => {
          if (subscribe)
            mailbox.CreateChild("Child", true);
          else
            mailbox.CreateChild("Child");
        });

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(ImapResponseCode.AlreadyExists,
                        ex.Result.TaggedStatusResponse.ResponseText.Code);
      });
    }

    [Test]
    public void TestGetChild()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" INBOX.Child\r\n",
                                       "* LIST () \".\" INBOX.Child\r\n" +
                                       "$tag OK done\r\n");
        });

        var child = mailbox.GetChild("Child");

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(child);
        Assert.AreEqual("Child", child.Name);
        Assert.AreEqual("INBOX.Child", child.FullName);
      });
    }

    [Test]
    public void TestGetChild_NameEmpty()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentException>(() => mailbox.GetChild(string.Empty));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetChild_NameNull()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.GetChild((string)null));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetChild_NotFound()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("$tag OK done\r\n");
        });

        var ex = Assert.Throws<ImapMailboxNotFoundException>(() => {
          mailbox.GetChild("Child");
          Assert.Fail("ImapMailboxNotFoundException not thrown");
        });

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(ex.Mailbox);
        Assert.AreEqual("INBOX.Child", ex.Mailbox);

        Smdn.Net.TestUtils.SerializeBinary(ex, delegate(ImapMailboxNotFoundException deserialized) {
          Assert.IsNotNull(deserialized.Mailbox);
          Assert.AreEqual(ex.Mailbox, deserialized.Mailbox);
        });
      });
    }

    [Test]
    public void TestGetOrCreate_ExistentChild()
    {
      GetOrCreateChildExistent(false);
    }

    [Test]
    public void TestGetOrCreate_ExistentChild_Subscribe()
    {
      GetOrCreateChildExistent(true);
    }

    private void GetOrCreateChildExistent(bool subscribe)
    {
      TestUtils2.TestMailbox(".", "Sent", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" Sent.Child\r\n",
                                       "* LIST () \".\" Sent.Child\r\n" +
                                       "$tag OK done\r\n");
        });

        var child = subscribe
          ? mailbox.GetOrCreateChild("Child", true)
          : mailbox.GetOrCreateChild("Child");

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(child);
        Assert.AreEqual("Child", child.Name);
        Assert.AreEqual("Sent.Child", child.FullName);
      });
    }

    [Test]
    public void TestGetOrCreate_NonExistentChild()
    {
      GetOrCreateChildNonExistent(false);
    }

    [Test]
    public void TestGetOrCreate_NonExistentChild_Subscribe()
    {
      GetOrCreateChildNonExistent(true);
    }

    private void GetOrCreateChildNonExistent(bool subscribe)
    {
      TestUtils2.TestMailbox(".", "Sent", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" Sent.Child\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("CREATE Sent.Child\r\n",
                                       "$tag OK done\r\n");
          if (subscribe)
            server.EnqueueTaggedResponse("SUBSCRIBE Sent.Child\r\n",
                                         "$tag OK done\r\n");
          server.EnqueueTaggedResponse("LIST \"\" Sent.Child\r\n",
                                       "* LIST () \".\" Sent.Child\r\n" +
                                       "$tag OK done\r\n");
        });

        var child = subscribe
          ? mailbox.GetOrCreateChild("Child", true)
          : mailbox.GetOrCreateChild("Child");

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(child);
        Assert.AreEqual("Child", child.Name);
        Assert.AreEqual("Sent.Child", child.FullName);
      });
    }

    [Test]
    public void TestGetOrCreateChild_NameEmpty()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentException>(() => mailbox.GetOrCreateChild(string.Empty));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetOrCreateChild_NameNull()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.GetOrCreateChild((string)null));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetChild_NonHierarchical()
    {
      TestNonHierarchicalMailbox(delegate(ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetChild("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetOrCreateChild_NonHierarchical()
    {
      TestNonHierarchicalMailbox(delegate(ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetOrCreateChild("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestCreateChild_NonHierarchical()
    {
      TestNonHierarchicalMailbox(delegate(ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.CreateChild("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetFullNameOf_NonHierarchical()
    {
      TestNonHierarchicalMailbox(delegate(ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetFullNameOf("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    private void TestNonHierarchicalMailbox(Action<ImapMailboxInfo> action)
    {
      TestUtils2.TestMailbox(string.Empty, "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.CanHaveChild);

        action(mailbox);
      });
    }

    [Test]
    public void TestGetChild_NoInferiors()
    {
      TestNoInferiorsMailbox(delegate(ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetChild("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetOrCreateChild_NoInferiors()
    {
      TestNoInferiorsMailbox(delegate(ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetOrCreateChild("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestCreateChild_NoInferiors()
    {
      TestNoInferiorsMailbox(delegate(ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.CreateChild("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetFullNameOf_NoInferiors()
    {
      TestNoInferiorsMailbox(delegate(ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetFullNameOf("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    private void TestNoInferiorsMailbox(Action<ImapMailboxInfo> action)
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             string.Empty,
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.CanHaveChild);

        action(mailbox);
      });
    }

    [Test]
    public void TestGetQuota()
    {
      TestUtils2.TestMailbox(new[] {ImapCapability.Quota},
                             null,
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("GETQUOTAROOT INBOX\r\n",
                                       "* QUOTAROOT INBOX \"\"\r\n" +
                                       "* QUOTA \"\" (STORAGE 10 512)\r\n" +
                                       "$tag OK done\r\n");
        });

        var quotas = new List<ImapQuota>(mailbox.GetQuota());

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(1, quotas.Count);

        var quota = quotas[0];

        Assert.AreEqual(string.Empty, quota.Root);
        Assert.AreEqual(1, quota.Resources.Count);
        Assert.AreEqual("STORAGE", quota.Resources[0].Name);
        Assert.AreEqual(10, quota.Resources[0].Usage);
        Assert.AreEqual(512, quota.Resources[0].Limit);
      });
    }

    [Test]
    public void TestGetQuota_QuotaIncapable()
    {
      TestUtils2.TestMailbox("INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var quotas = new List<ImapQuota>(mailbox.GetQuota());

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(0, quotas.Count());
      });
    }

    [Test]
    public void TestGetQuota_QuotaDisabled()
    {
      TestUtils2.TestMailbox(new[] {ImapCapability.Quota},
                             null,
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        mailbox.Client.DisabledCapabilities.Add(ImapCapability.Quota);

        var quotas = new List<ImapQuota>(mailbox.GetQuota());

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(0, quotas.Count());
      });
    }
  }
}
