using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapMailboxInfoOperationsAppendTests {
    [Test]
    public void TestAppendMessage_IAppendMessage()
    {
      TestUtils2.TestMailbox("INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        const string uploadingMessage = "MIME-Version: 1.0\r\n\r\ntest message";

        var message = new ImapAppendMessage(Encoding.ASCII.GetBytes(uploadingMessage),
                                            new DateTimeOffset(2011, 1, 24, 14, 52, 30, TimeSpan.FromHours(+9.0)),
                                            new ImapMessageFlagSet(new[] {"$label1"}, ImapMessageFlag.Seen));

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("APPEND INBOX ",
                                       "$tag OK [APPENDUID 38505 3955] APPEND completed\r\n");
        });

        mailbox.AppendMessage(message);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.That(server.DequeueRequest(),
                    Is.StringEnding(string.Format("APPEND INBOX (\\Seen $label1) \"24-Jan-2011 14:52:30 +0900\" {{{0}}}\r\n{1}\r\n",
                                                  uploadingMessage.Length,
                                                  uploadingMessage)));
      });
    }

    [Test]
    public void TestAppendMessage_IAppendMessage_MailboxNonExistent()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NonExistent},
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Exists);

        using (var stream = new MemoryStream(new byte[] {0x40})) {
          Assert.Throws<ImapProtocolViolationException>(() => mailbox.AppendMessage(stream));
        }

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestAppendMessage_IAppendMessage_MailboxUnselectable()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoSelect},
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.IsUnselectable);

        using (var stream = new MemoryStream(new byte[] {0x40})) {
          Assert.Throws<ImapProtocolViolationException>(() => mailbox.AppendMessage(stream));
        }

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestAppendMessage_IAppendMessage_NoTryCreate()
    {
      TestUtils2.TestMailbox("NonExistent",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueLiteralContinuationResponse("APPEND NonExistent {1}\r\n",
                                                    "$tag NO [TRYCREATE] mailbox not exist\r\n",
                                                    false);
        });

        var ex = Assert.Throws<ImapMailboxNotFoundException>(() => mailbox.AppendMessage(new MemoryStream(new byte[] {0x40})));

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(ex.Mailbox);
        Assert.AreEqual("NonExistent", ex.Mailbox);

        Smdn.Net.TestUtils.SerializeBinary(ex, delegate(ImapMailboxNotFoundException deserialized) {
          Assert.IsNotNull(deserialized.Mailbox);
          Assert.AreEqual(ex.Mailbox, deserialized.Mailbox);
        });
      });
    }

    [Test]
    public void TestAppendMessages()
    {
      TestUtils2.TestMailbox("INBOX", AppendMessages);
    }

    [Test]
    public void TestAppendMessages_MultiappendDisabled()
    {
      TestUtils2.TestMailbox(new[] {ImapCapability.MultiAppend},
                             null,
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        mailbox.Client.DisabledCapabilities.Add(ImapCapability.MultiAppend);

        AppendMessages(server, mailbox);
      });
    }

    private void AppendMessages(ImapPseudoServer2 server, ImapMailboxInfo mailbox)
    {
      const string uploadingMessage1 = "MIME-Version: 1.0\r\n\r\ntest message1";
      const string uploadingMessage2 = "MIME-Version: 1.0\r\n\r\ntest message2";
      const string uploadingMessage3 = "MIME-Version: 1.0\r\n\r\ntest message3";

      var messages = new[] {
        new ImapAppendMessage(Encoding.ASCII.GetBytes(uploadingMessage1),
                              new DateTimeOffset(2011, 1, 24, 14, 52, 30, TimeSpan.FromHours(+9.0)),
                              new ImapMessageFlagSet(new[] {"$label1"}, ImapMessageFlag.Seen)),
        new ImapAppendMessage(Encoding.ASCII.GetBytes(uploadingMessage2),
                              new DateTimeOffset(2011, 1, 24, 14, 52, 31, TimeSpan.FromHours(+9.0)),
                              new ImapMessageFlagSet(new[] {"$label2"}, ImapMessageFlag.Seen)),
        new ImapAppendMessage(Encoding.ASCII.GetBytes(uploadingMessage3),
                              new DateTimeOffset(2011, 1, 24, 14, 52, 32, TimeSpan.FromHours(+9.0)),
                              new ImapMessageFlagSet(new[] {"$label3"}, ImapMessageFlag.Seen)),
      };

      var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
        server.EnqueueTaggedResponse(string.Format("APPEND INBOX (\\Seen $label1) \"24-Jan-2011 14:52:30 +0900\" {{{0}}}\r\n{1}\r\n",
                                                   uploadingMessage1.Length,
                                                   uploadingMessage1),
                                     "$tag OK [APPENDUID 38505 3955] APPEND completed\r\n");
        server.EnqueueTaggedResponse(string.Format("APPEND INBOX (\\Seen $label2) \"24-Jan-2011 14:52:31 +0900\" {{{0}}}\r\n{1}\r\n",
                                                   uploadingMessage2.Length,
                                                   uploadingMessage2),
                                     "$tag OK [APPENDUID 38505 3956] APPEND completed\r\n");
        server.EnqueueTaggedResponse(string.Format("APPEND INBOX (\\Seen $label3) \"24-Jan-2011 14:52:32 +0900\" {{{0}}}\r\n{1}\r\n",
                                                   uploadingMessage3.Length,
                                                   uploadingMessage3),
                                     "$tag OK [APPENDUID 38505 3957] APPEND completed\r\n");
      });

      mailbox.AppendMessages(messages);

      responseTask.Wait();

      Assert.IsFalse(mailbox.Client.IsBusy);
    }

    [Test]
    public void TestAppendMessages_MultiappendCapable()
    {
      TestUtils2.TestMailbox(new[] {ImapCapability.MultiAppend},
                             null,
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        const string uploadingMessage1 = "MIME-Version: 1.0\r\n\r\ntest message1";
        const string uploadingMessage2 = "MIME-Version: 1.0\r\n\r\ntest message2";
        const string uploadingMessage3 = "MIME-Version: 1.0\r\n\r\ntest message3";

        var messages = new[] {
          new ImapAppendMessage(Encoding.ASCII.GetBytes(uploadingMessage1),
                                new DateTimeOffset(2011, 1, 24, 14, 52, 30, TimeSpan.FromHours(+9.0)),
                                new ImapMessageFlagSet(new[] {"$label1"}, ImapMessageFlag.Seen)),
          new ImapAppendMessage(Encoding.ASCII.GetBytes(uploadingMessage2),
                                new DateTimeOffset(2011, 1, 24, 14, 52, 31, TimeSpan.FromHours(+9.0)),
                                new ImapMessageFlagSet(new[] {"$label2"}, ImapMessageFlag.Seen)),
          new ImapAppendMessage(Encoding.ASCII.GetBytes(uploadingMessage3),
                                new DateTimeOffset(2011, 1, 24, 14, 52, 32, TimeSpan.FromHours(+9.0)),
                                new ImapMessageFlagSet(new[] {"$label3"}, ImapMessageFlag.Seen)),
        };

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          var expectedCommand =
            "APPEND INBOX " +
            string.Format("(\\Seen $label1) \"24-Jan-2011 14:52:30 +0900\" {{{0}}}\r\n{1} ",
                          uploadingMessage1.Length,
                          uploadingMessage1) +
            string.Format("(\\Seen $label2) \"24-Jan-2011 14:52:31 +0900\" {{{0}}}\r\n{1} ",
                          uploadingMessage2.Length,
                          uploadingMessage2) +
            string.Format("(\\Seen $label3) \"24-Jan-2011 14:52:32 +0900\" {{{0}}}\r\n{1}\r\n",
                          uploadingMessage3.Length,
                          uploadingMessage3);

          server.EnqueueTaggedResponse(expectedCommand,
                                       "$tag OK [APPENDUID 38505 3955,3956,3957] APPEND completed\r\n");
        });

        mailbox.AppendMessages(messages);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestAppendMessages_NonExistentMailbox()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NonExistent},
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Exists);

        var messages = new[] {
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
        };

        Assert.Throws<ImapProtocolViolationException>(() => mailbox.AppendMessages(messages));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestAppendMessages_UnselectableMailbox()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoSelect},
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.IsUnselectable);

        var messages = new[] {
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
        };

        Assert.Throws<ImapProtocolViolationException>(() => mailbox.AppendMessages(messages));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestAppendMessages_NoTryCreate()
    {
      TestUtils2.TestMailbox("NonExistent",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueLiteralContinuationResponse("APPEND NonExistent {1}\r\n",
                                                    "$tag NO [TRYCREATE] mailbox not exist\r\n",
                                                    false);
        });

        var messages = new[] {
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
        };

        var ex = Assert.Throws<ImapMailboxNotFoundException>(() => mailbox.AppendMessages(messages));

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(ex.Mailbox);
        Assert.AreEqual("NonExistent", ex.Mailbox);

        Smdn.Net.TestUtils.SerializeBinary(ex, delegate(ImapMailboxNotFoundException deserialized) {
          Assert.IsNotNull(deserialized.Mailbox);
          Assert.AreEqual(ex.Mailbox, deserialized.Mailbox);
        });
      });
    }

    [Test]
    public void TestAppendMessage_MessageInfoOfSameSession()
    {
      TestUtils2.TestMailbox("INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo sourceMailbox) {
        server.EnqueueTaggedResponse("LIST \"\" dest\r\n",
                                     "* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = sourceMailbox.Client.GetMailbox("dest");

        server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                     "* EXISTS 3\r\n" +
                                     "* OK [UIDVALIDITY 23]\r\n" +
                                     "$tag OK [READ-WRITE] done\r\n");

        using (var openedSourceMailbox = sourceMailbox.Open()) {
          server.EnqueueTaggedResponse("FETCH 1 (UID)\r\n",
                                       "* FETCH 1 (UID 3)\r\n" +
                                       "$tag OK done\r\n");

          var message = openedSourceMailbox.GetMessageBySequence(1L);

          var responseTask = TestUtils2.QueueResponseTask(server, sourceMailbox.Client, delegate {
            server.EnqueueTaggedResponse("UID COPY 3 dest\r\n",
                                         "$tag OK done\r\n");
          });

          destMailbox.AppendMessage(message);

          responseTask.Wait();

          Assert.IsFalse(destMailbox.Client.IsBusy);

          server.EnqueueTaggedResponse("CLOSE\r\n",
                                       "$tag OK done\r\n");
        }
      });
    }

    [Test]
    public void TestAppendMessage_MessageInfoOfDifferentSession()
    {
      TestUtils2.TestMailbox("dest", delegate(ImapPseudoServer2 destServer, ImapMailboxInfo destMailbox) {
        var selectResp = "* EXISTS 3\r\n" +
          "* OK [UIDVALIDITY 23]\r\n" +
            "$tag OK [READ-WRITE] done\r\n";

        TestUtils2.TestOpenedMailbox("INBOX",
                                     selectResp,
                                     delegate(ImapPseudoServer2 sourceServer, ImapOpenedMailboxInfo sourceMailbox) {
          sourceServer.EnqueueTaggedResponse("FETCH 1 (UID)\r\n",
                                             "* FETCH 1 (UID 3)\r\n" +
                                             "$tag OK done\r\n");

          var message = sourceMailbox.GetMessageBySequence(1L);

          sourceServer.EnqueueTaggedResponse("UID FETCH 3 (RFC822.SIZE INTERNALDATE FLAGS BODY.PEEK[]<0.10240>)\r\n",
                                             "* FETCH 1 (" +
                                             "FLAGS (\\Answered $label1) " +
                                             "INTERNALDATE \"25-Jan-2011 15:29:06 +0900\" " +
                                             "RFC822.SIZE 12 " +
                                             "BODY[] {12}\r\ntest message)\r\n" +
                                             "$tag OK done\r\n");

          var responseTask = TestUtils2.QueueResponseTask(destServer, destMailbox.Client, delegate {
            destServer.EnqueueTaggedResponse("APPEND dest (\\Answered $label1) \"25-Jan-2011 15:29:06 +0900\" {12}\r\ntest message\r\n",
                                             "$tag OK [APPENDUID 38505 3955] APPEND completed\r\n");
          });

          destMailbox.AppendMessage(message);

          responseTask.Wait();

          Assert.IsFalse(destMailbox.Client.IsBusy);
        });
      });
    }

    [Test]
    public void TestAppendMessage_Action()
    {
      AppendMessageAction(false);
    }

    [Test]
    public void TestAppendMessage_Action_LengthSpecified()
    {
      AppendMessageAction(true);
    }

    private void AppendMessageAction(bool specifyLength)
    {
      TestUtils2.TestMailbox("INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        const string uploadingMessage = "MIME-Version: 1.0\r\n\r\ntest message";

        var length = uploadingMessage.Length;
        var flags = new ImapMessageFlagSet(new[] {"$label1"}, ImapMessageFlag.Seen);
        var internalDate = new DateTimeOffset(2011, 1, 24, 14, 52, 30, TimeSpan.FromHours(+9.0));
        var writeUploadMessage = new Action<Stream>(delegate(Stream stream) {
          var writer = new StreamWriter(stream, Encoding.ASCII);

          writer.Write(uploadingMessage);
          writer.Flush();
        });

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse(string.Format("APPEND INBOX (\\Seen $label1) \"24-Jan-2011 14:52:30 +0900\" {{{0}}}\r\n{1}\r\n",
                                                     length,
                                                     uploadingMessage),
                                       "$tag OK [APPENDUID 38505 3955] APPEND completed\r\n");
        });

        if (specifyLength)
          mailbox.AppendMessage(length,
                                internalDate,
                                flags,
                                writeUploadMessage);
        else
          mailbox.AppendMessage(internalDate,
                                flags,
                                writeUploadMessage);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestAppendMessage_Action_MailboxNonExistent()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NonExistent},
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Exists);

        Assert.Throws<ImapProtocolViolationException>(delegate {
          mailbox.AppendMessage(delegate(Stream stream) {
          });
        });

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestAppendMessage_Action_MailboxUnselectable()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoSelect},
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.IsUnselectable);

        Assert.Throws<ImapProtocolViolationException>(delegate {
          mailbox.AppendMessage(delegate(Stream stream) {
          });
        });

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestAppendMessage_Action_LengthSpecified_OutOfRange()
    {
      TestUtils2.TestMailbox("INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(delegate {
          mailbox.AppendMessage(-1L, delegate(Stream stream) {
          });
        });

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestAppendMessage_Action_NoTryCreate()
    {
      TestUtils2.TestMailbox("NonExistent",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueLiteralContinuationResponse("APPEND NonExistent {1}\r\n",
                                                    "$tag NO [TRYCREATE] mailbox not exist\r\n",
                                                    false);
        });

        var ex = Assert.Throws<ImapMailboxNotFoundException>(() => {
          mailbox.AppendMessage(1, delegate(Stream stream) {
            stream.WriteByte(0x40);
          });
        });

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(ex.Mailbox);
        Assert.AreEqual("NonExistent", ex.Mailbox);

        Smdn.Net.TestUtils.SerializeBinary(ex, delegate(ImapMailboxNotFoundException deserialized) {
          Assert.IsNotNull(deserialized.Mailbox);
          Assert.AreEqual(ex.Mailbox, deserialized.Mailbox);
        });
      });
    }
  }
}