using System;
using System.Linq;
using System.Text;
using NUnit.Framework;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapMessageInfoBaseTests {
    private class MessageSet : ImapMessageInfoBase {
      public MessageSet(ImapOpenedMailboxInfo mailbox, ImapSequenceSet sequenceOrUidSet)
        : base(mailbox)
      {
        this.sequenceOrUidSet = sequenceOrUidSet;
      }

      protected override ImapSequenceSet GetSequenceOrUidSet()
      {
        return sequenceOrUidSet;
      }

      private ImapSequenceSet sequenceOrUidSet;
    }

    private void TestMessageSet(int existMesasgeCount,
                                ImapSequenceSet sequenceOrUidSet,
                                Action<ImapPseudoServer2, ImapMessageInfoBase, ImapMessageInfo[]> action)
    {
      TestMessageSet(existMesasgeCount, null, sequenceOrUidSet, action);
    }

    private void TestMessageSet(int existMessageCount,
                                ImapCapability[] capabilities,
                                ImapSequenceSet sequenceOrUidSet,
                                Action<ImapPseudoServer2, ImapMessageInfoBase, ImapMessageInfo[]> action)
    {
      TestUtils2.TestAuthenticated(capabilities,
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        server.EnqueueTaggedResponse("LIST ",
                                     "* LIST () \".\" INBOX\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                     string.Format("* EXISTS {0}\r\n", existMessageCount) +
                                     "* OK [UIDVALIDITY 1]\r\n" +
                                     "$tag OK done\r\n");

        using (var inbox = client.OpenInbox()) {
          server.DequeueRequest(); // LIST
          server.DequeueRequest(); // SELECT

          var messages = new ImapMessageInfo[0];

          if (0 < existMessageCount) {
            server.EnqueueTaggedResponse("NOOP\r\n",
                                         "$tag OK done\r\n");

            var fetchResp = new StringBuilder();

            for (var seq = 1; seq <= existMessageCount; seq++) {
              fetchResp.AppendFormat("* FETCH {0} (UID {0})\r\n", seq);
            }

            fetchResp.Append("$tag OK done\r\n");

            server.EnqueueTaggedResponse("FETCH ",
                                         fetchResp.ToString());

            var msgs = inbox.GetMessages();

            messages = msgs.ToArray();

            server.DequeueRequest(); // NOOP
            server.DequeueRequest(); // FETCH

            Assert.AreSame(client, msgs.Client);
            Assert.AreSame(inbox, msgs.Mailbox);
            Assert.IsTrue(msgs.IsAvailable);
          }

          try {
            var messageSet = new MessageSet(inbox, sequenceOrUidSet);

            Assert.AreSame(client, messageSet.Client);
            Assert.AreSame(inbox, messageSet.Mailbox);
            Assert.IsTrue(messageSet.IsAvailable);
            Assert.AreEqual(1L, messageSet.UidValidity);

            action(server, messageSet, messages);
          }
          finally {
            if (inbox.IsOpen)
              server.EnqueueTaggedResponse("CLOSE\r\n",
                                           "$tag OK done\r\n");
          }
        }
      });
    }

    private ImapSequenceSet CreateEmptySet()
    {
      return ImapSequenceSet.CreateSet(new long[0]);
    }

    [Test]
    public void TestAddFlags()
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        var responseTask = TestUtils2.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("UID STORE 1 +FLAGS (\\Draft)\r\n",
                                     "* FETCH 1 (FLAGS (\\Seen \\Draft))\r\n" +
                                     "$tag OK done\r\n");
        });

        messageSet.AddFlags(ImapMessageFlag.Draft);

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(2, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue(messages[0].IsSeen);
        Assert.IsTrue(messages[0].IsDraft);
      });
    }

    [Test]
    public void TestRemoveFlags()
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        var responseTask = TestUtils2.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("UID STORE 1 -FLAGS (\\Draft)\r\n",
                                       "* FETCH 1 (FLAGS (\\Seen))\r\n" +
                                       "$tag OK done\r\n");
        });

        messageSet.RemoveFlags(ImapMessageFlag.Draft);

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[0].IsSeen);
      });
    }

    [Test]
    public void TestSetFlags()
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        var responseTask = TestUtils2.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("UID STORE 1 FLAGS (\\Draft \\Flagged)\r\n",
                                       "* FETCH 1 (FLAGS (\\Draft \\Flagged))\r\n" +
                                       "$tag OK done\r\n");
        });

        messageSet.SetFlags(ImapMessageFlag.Draft, ImapMessageFlag.Flagged);

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(2, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Flagged));
        Assert.IsTrue(messages[0].IsDraft);
        Assert.IsTrue(messages[0].IsFlagged);
      });
    }

    [Test]
    public void TestAddFlags_NonApplicableFlagsRemoved()
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        var responseTask = TestUtils2.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("UID STORE 1 +FLAGS (\\Draft)\r\n",
                                       "* FETCH 1 (FLAGS (\\Draft))\r\n" +
                                       "$tag OK done\r\n");
        });

        messageSet.AddFlags(ImapMessageFlag.Draft, ImapMessageFlag.Recent);

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsFalse(messages[0].Flags.Contains(ImapMessageFlag.Recent));
      });
    }

    [Test]
    public void TestRemoveFlags_NonApplicableFlagsRemoved()
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        var responseTask = TestUtils2.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("UID STORE 1 -FLAGS (\\Draft)\r\n",
                                       "* FETCH 1 (FLAGS (\\Draft \\Recent))\r\n" +
                                       "$tag OK done\r\n");
        });

        messageSet.RemoveFlags(ImapMessageFlag.Draft, ImapMessageFlag.Recent);

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(2, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Recent));
      });
    }

    [Test]
    public void TestSetFlags_NonApplicableFlagsRemoved()
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        var responseTask = TestUtils2.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("UID STORE 1 FLAGS (\\Draft)\r\n",
                                       "* FETCH 1 (FLAGS (\\Draft))\r\n" +
                                       "$tag OK done\r\n");
        });

        messageSet.SetFlags(ImapMessageFlag.Draft, ImapMessageFlag.Recent);

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsFalse(messages[0].Flags.Contains(ImapMessageFlag.Recent));
      });
    }

    [Test]
    public void TestAddFlags_ContainsNoApplicableFlags()
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        Assert.Throws<ArgumentException>(() => messageSet.AddFlags(ImapMessageFlag.Recent));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestRemoveFlags_ContainsNoApplicableFlags()
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        Assert.Throws<ArgumentException>(() => messageSet.RemoveFlags(ImapMessageFlag.Recent));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestSetFlags_ContainsNoApplicableFlags()
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        Assert.Throws<ArgumentException>(() => messageSet.SetFlags(ImapMessageFlag.Recent));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestStore_EmptySet()
    {
      TestMessageSet(0, CreateEmptySet(),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // do nothing
        messageSet.Store(ImapStoreDataItem.AddFlags(ImapMessageFlag.Seen));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestStore_SingleUidSet()
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        var responseTask = TestUtils2.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("UID STORE 1 +FLAGS (\\Seen)\r\n",
                                       "* FETCH 1 (FLAGS (\\Seen))\r\n" +
                                       "$tag OK done\r\n");
        });

        messageSet.Store(ImapStoreDataItem.AddFlags(ImapMessageFlag.Seen));

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[0].IsSeen);
      });
    }

    [Test]
    public void TestStore_SequenceRangeSet()
    {
      TestMessageSet(3, ImapSequenceSet.CreateRangeSet(1L, 3L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        var responseTask = TestUtils2.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("STORE 1:3 -FLAGS (\\Seen)\r\n",
                                       "* FETCH 1 (FLAGS (\\Draft))\r\n" +
                                       "* FETCH 2 (FLAGS ($label1))\r\n" +
                                       "* FETCH 3 (FLAGS ())\r\n" +
                                       "$tag OK done\r\n");
        });

        messageSet.Store(ImapStoreDataItem.RemoveFlags(ImapMessageFlag.Seen));

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue(messages[0].IsDraft);

        Assert.IsNotNull(messages[1].Flags);
        Assert.AreEqual(1, messages[1].Flags.Count);
        Assert.IsTrue(messages[1].Flags.Contains("$label1"));

        Assert.IsNotNull(messages[2].Flags);
        Assert.AreEqual(0, messages[2].Flags.Count);
      });
    }

    [Test]
    public void TestStore_MailboxClosed()
    {
      TestMessageSet(10, ImapSequenceSet.CreateRangeSet(1L, 10L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK done\r\n");

        var mailbox = messageSet.Mailbox;

        messageSet.Mailbox.Close();

        Assert.IsFalse(mailbox.IsOpen);

        var ex = Assert.Throws<ImapUnavailableException>(() => messageSet.Store(ImapStoreDataItem.RemoveFlags(ImapMessageFlag.Seen)));

        Assert.IsFalse(messageSet.Client.IsBusy);

        Smdn.Net.TestUtils.SerializeBinary(ex);
      });
    }

    [Test, Ignore("not implemented; uid validity")]
    public void TestStore_UidValidityChanged()
    {
    }

    [Test]
    public void TestMarkAsDeleted()
    {
      TestMessageSet(3, ImapSequenceSet.CreateRangeSet(1L, 3L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        var responseTask = TestUtils2.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("STORE 1:3 +FLAGS (\\Deleted)\r\n",
                                       "* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                       "* FETCH 2 (FLAGS (\\Deleted \\Seen))\r\n" +
                                       "* FETCH 3 (FLAGS (\\Deleted $label1))\r\n" +
                                       "$tag OK done\r\n");
        });

        messageSet.MarkAsDeleted();

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsTrue(messages[0].IsAvailable);
        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[0].IsMarkedAsDeleted);

        Assert.IsTrue(messages[1].IsAvailable);
        Assert.IsNotNull(messages[1].Flags);
        Assert.AreEqual(2, messages[1].Flags.Count);
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[1].IsSeen);
        Assert.IsTrue(messages[1].IsMarkedAsDeleted);

        Assert.IsTrue(messages[2].IsAvailable);
        Assert.IsNotNull(messages[2].Flags);
        Assert.AreEqual(2, messages[2].Flags.Count);
        Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[2].Flags.Contains("$label1"));
        Assert.IsTrue(messages[2].IsMarkedAsDeleted);
      });
    }

    [Test]
    public void TestMarkAsSeen()
    {
      TestMessageSet(3, ImapSequenceSet.CreateRangeSet(1L, 3L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        var responseTask = TestUtils2.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("STORE 1:3 +FLAGS (\\Seen)\r\n",
                                       "* FETCH 1 (FLAGS (\\Seen))\r\n" +
                                       "* FETCH 2 (FLAGS (\\Seen \\Draft))\r\n" +
                                       "* FETCH 3 (FLAGS (\\Seen $label1))\r\n" +
                                       "$tag OK done\r\n");
        });

        messageSet.MarkAsSeen();

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[0].IsSeen);

        Assert.IsNotNull(messages[1].Flags);
        Assert.AreEqual(2, messages[1].Flags.Count);
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue(messages[1].IsSeen);
        Assert.IsTrue(messages[1].IsDraft);

        Assert.IsNotNull(messages[2].Flags);
        Assert.AreEqual(2, messages[2].Flags.Count);
        Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[2].Flags.Contains("$label1"));
        Assert.IsTrue(messages[2].IsSeen);
      });
    }

    [Test]
    public void TestDelete_UidPlusCapable()
    {
      TestMessageSet(3,
                     new[] {ImapCapability.UidPlus},
                     ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        Delete(server, messageSet, messages, true);
      });
    }

    [Test]
    public void TestDelete_UidPlusIncapable()
    {
      TestMessageSet(3,
                     null,
                     ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        Delete(server, messageSet, messages, false);
      });
    }

    [Test]
    public void TestDeleteUidPlusDisabled()
    {
      TestMessageSet(3,
                     new[] {ImapCapability.UidPlus},
                     ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        messageSet.Client.DisabledCapabilities.Add(ImapCapability.UidPlus);

        Delete(server, messageSet, messages, false);
      });
    }

    private void Delete(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages, bool uidPlusAvailable)
    {
      var responseTask = TestUtils2.QueueResponseTask(server, messageSet.Client, delegate {
        server.EnqueueTaggedResponse("STORE 1:3 +FLAGS (\\Deleted)\r\n",
                                     "* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                     "* FETCH 2 (FLAGS (\\Deleted \\Seen))\r\n" +
                                     "* FETCH 3 (FLAGS (\\Deleted $label1))\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse(uidPlusAvailable ? "UID EXPUNGE 1:3\r\n" : "EXPUNGE\r\n",
                                     "* EXPUNGE 3\r\n" +
                                     "* EXPUNGE 2\r\n" +
                                     "* EXPUNGE 1\r\n" +
                                     "$tag OK done\r\n");
      });

      var mailbox = messageSet.Mailbox;

      messageSet.Delete();

      responseTask.Wait();

      Assert.IsFalse(messageSet.Client.IsBusy);

      Assert.IsFalse(messageSet.IsAvailable);
      Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messageSet.Mailbox));
      Assert.AreEqual(0L, mailbox.ExistMessageCount);

      Assert.IsFalse(messages[0].IsAvailable);
      Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messages[0].Mailbox));
      Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[0].Sequence);
      Assert.IsTrue(messages[0].IsDeleted);

      Assert.IsNotNull(messages[0].Flags);
      Assert.AreEqual(1, messages[0].Flags.Count);
      Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Deleted));
      Assert.IsTrue(messages[0].IsMarkedAsDeleted);

      Assert.IsFalse(messages[1].IsAvailable);
      Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messages[1].Mailbox));
      Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[1].Sequence);
      Assert.IsTrue(messages[1].IsDeleted);

      Assert.IsNotNull(messages[1].Flags);
      Assert.AreEqual(2, messages[1].Flags.Count);
      Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Deleted));
      Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Seen));
      Assert.IsTrue(messages[1].IsMarkedAsDeleted);
      Assert.IsTrue(messages[1].IsSeen);

      Assert.IsFalse(messages[0].IsAvailable);
      Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messages[2].Mailbox));
      Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[2].Sequence);
      Assert.IsTrue(messages[2].IsDeleted);

      Assert.IsNotNull(messages[2].Flags);
      Assert.AreEqual(2, messages[2].Flags.Count);
      Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Deleted));
      Assert.IsTrue(messages[2].Flags.Contains("$label1"));
      Assert.IsTrue(messages[2].IsMarkedAsDeleted);
    }

    [Test]
    public void TestDelete_EmptySet()
    {
      TestMessageSet(0, CreateEmptySet(),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // do nothing
        messageSet.Delete();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsTrue(messageSet.IsAvailable);
        Assert.DoesNotThrow(() => Assert.IsNotNull(messageSet.Mailbox));
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationMailbox()
    {
      TestMessageSet(3, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        CopyOrMoveToDestinationMailbox(server, messageSet, messages, false);
      });
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox()
    {
      TestMessageSet(3, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        CopyOrMoveToDestinationMailbox(server, messageSet, messages, true);
      });
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_MoveDisabled()
    {
      TestMessageSet(3, new[] {ImapCapability.Move}, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        messageSet.Client.DisabledCapabilities.Add(ImapCapability.Move);

        CopyOrMoveToDestinationMailbox(server, messageSet, messages, true);
      });
    }

    private void CopyOrMoveToDestinationMailbox(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages, bool move)
    {
      server.EnqueueTaggedResponse("LIST ",
                                   "* LIST () \"\" dest\r\n" +
                                   "$tag OK done\r\n");

      var destMailbox = messageSet.Client.GetMailbox("dest");

      var responseTask = TestUtils2.QueueResponseTask(server, messageSet.Client, delegate {
        server.EnqueueTaggedResponse("UID COPY 1,3,5 dest\r\n",
                                     "$tag OK done\r\n");
        if (move)
          server.EnqueueTaggedResponse("UID STORE 1,3,5 +FLAGS (\\Deleted)\r\n",
                                       "* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                       "* FETCH 2 (FLAGS (\\Deleted \\Seen))\r\n" +
                                       "* FETCH 3 (FLAGS (\\Deleted $label1))\r\n" +
                                       "$tag OK done\r\n");
      });

      if (move)
        messageSet.MoveTo(destMailbox);
      else
        messageSet.CopyTo(destMailbox);

      responseTask.Wait();

      Assert.IsFalse(messageSet.Client.IsBusy);

      if (move) {
        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[0].IsMarkedAsDeleted);

        Assert.IsNotNull(messages[1].Flags);
        Assert.AreEqual(2, messages[1].Flags.Count);
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[1].IsMarkedAsDeleted);
        Assert.IsTrue(messages[1].IsSeen);

        Assert.IsNotNull(messages[2].Flags);
        Assert.AreEqual(2, messages[2].Flags.Count);
        Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[2].Flags.Contains("$label1"));
        Assert.IsTrue(messages[2].IsMarkedAsDeleted);
      }
    }
    
    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationMailbox()
    {
      TestMessageSet(3,
                     new[] {ImapCapability.Move},
                     ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        server.EnqueueTaggedResponse("LIST ",
                                     "* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var sourceMailbox = messageSet.Mailbox;
        var destMailbox = messageSet.Client.GetMailbox("dest");

        var responseTask = TestUtils2.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("UID MOVE 1:3 dest\r\n",
                                       "* EXPUNGE 3\r\n" +
                                       "* EXPUNGE 2\r\n" +
                                       "* EXPUNGE 1\r\n" +
                                       "$tag OK done\r\n");
        });

        messageSet.MoveTo(destMailbox);

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsFalse(messageSet.IsAvailable);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messageSet.Mailbox));
        Assert.AreEqual(0L, sourceMailbox.ExistMessageCount);

        Assert.IsFalse(messages[0].IsAvailable);
        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[0].Sequence);
        Assert.IsTrue(messages[0].IsDeleted);

        Assert.IsFalse(messages[1].IsAvailable);
        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[1].Sequence);
        Assert.IsTrue(messages[1].IsDeleted);

        Assert.IsFalse(messages[2].IsAvailable);
        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[2].Sequence);
        Assert.IsTrue(messages[2].IsDeleted);
      });
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_CopyFailed()
    {
      MoveToDestinationMailboxCopyFailed(false);
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationMailbox_CopyFailed()
    {
      MoveToDestinationMailboxCopyFailed(true);
    }

    private void MoveToDestinationMailboxCopyFailed(bool moveCapable)
    {
      var capabilities = moveCapable ? new[] { ImapCapability.Move } : EmptyArray<ImapCapability>.Instance;

      TestMessageSet(3, capabilities, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        server.EnqueueTaggedResponse("LIST ",
                                     "* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = messageSet.Client.GetMailbox("dest");

        var responseTask = TestUtils2.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse(moveCapable ? "UID MOVE 1,3,5 dest\r\n" : "UID COPY 1,3,5 dest\r\n",
                                       "$tag NO failed\r\n");
        });

        // STORE won't be issued
        Assert.Throws<ImapErrorResponseException>(() => messageSet.MoveTo(destMailbox));

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsTrue(messageSet.IsAvailable);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationMailbox_MailboxItself()
    {
      CopyOrMoveToDestinationMailboxItself(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_MailboxItself()
    {
      CopyOrMoveToDestinationMailboxItself(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationMailbox_MailboxItself()
    {
      CopyOrMoveToDestinationMailboxItself(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationMailboxItself(bool move, ImapCapability[] capabilities)
    {
      TestMessageSet(3, capabilities, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        Assert.AreEqual(3, messageSet.Mailbox.ExistMessageCount);
        Assert.AreEqual(0, messageSet.Mailbox.RecentMessageCount);

        if (move) {
          messageSet.MoveTo(messageSet.Mailbox);
        }
        else {
          var responseTask = TestUtils2.QueueResponseTask(server, messageSet.Client, delegate {
            server.EnqueueTaggedResponse("UID COPY 1,3,5 " + messageSet.Mailbox.FullName + "\r\n",
                                         "* 3 RECENT\r\n" +
                                         "* 6 EXISTS\r\n" +
                                         "$tag OK done\r\n");
          });

          messageSet.CopyTo(messageSet.Mailbox);

          responseTask.Wait();
        }

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsTrue(messageSet.IsAvailable);

        if (move) {
          Assert.AreEqual(3, messageSet.Mailbox.ExistMessageCount);
          Assert.AreEqual(0, messageSet.Mailbox.RecentMessageCount);
        }
        else {
          Assert.AreEqual(6, messageSet.Mailbox.ExistMessageCount);
          Assert.AreEqual(3, messageSet.Mailbox.RecentMessageCount);
        }
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationMailbox_MailboxOfDifferentSession()
    {
      CopyOrMoveToDestinationMailboxOfDifferentSession(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_MailboxOfDifferentSession()
    {
      CopyOrMoveToDestinationMailboxOfDifferentSession(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationMailbox_MailboxOfDifferentSession()
    {
      CopyOrMoveToDestinationMailboxOfDifferentSession(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationMailboxOfDifferentSession(bool move, ImapCapability[] capabilities)
    {
      TestMessageSet(3, capabilities, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     delegate(ImapPseudoServer2 serverSrc, ImapMessageInfoBase srcMessageSet, ImapMessageInfo[] srcMessages) {
        TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 serverDest, ImapClient clientDest) {
          serverDest.EnqueueTaggedResponse("LIST ",
                                           "* LIST () \"\" dest\r\n" +
                                           "$tag OK done\r\n");

          var destMailbox = clientDest.GetMailbox("dest");

          Assert.AreNotSame(srcMessageSet.Client, destMailbox.Client);

          var responseTask = TestUtils2.QueueResponseTask(serverSrc, srcMessageSet.Client, delegate {
            serverSrc.EnqueueTaggedResponse("UID FETCH 1,3,5 (UID)\r\n",
                                            "$tag NO FETCH failed\r\n");
          });

          Assert.Throws<ImapErrorResponseException>(delegate {
            if (move)
              srcMessageSet.MoveTo(destMailbox);
            else
              srcMessageSet.CopyTo(destMailbox);
          });

          responseTask.Wait();

          Assert.IsFalse(srcMessageSet.Client.IsBusy);
        });
      });
    }

    [Test, Ignore("not implemented; uid validity")]
    public void TestCopyTo_CopyToDestinationMailbox_UidValidityChanged()
    {
    }

    [Test, Ignore("not implemented; uid validity")]
    public void TestMoveTo_MoveToDestinationMailbox_UidValidityChanged()
    {
    }

    [Test, Ignore("not implemented; uid validity")]
    public void TestMoveTo_MoveCapable_MoveToDestinationMailbox_UidValidityChanged()
    {
    }

    [Test]
    public void TestCopyTo_CopyToDestinationMailbox_ArgumentNull()
    {
      CopyOrMoveToDestinationMailboxArgumentNull(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_ArgumentNull()
    {
      CopyOrMoveToDestinationMailboxArgumentNull(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationMailbox_ArgumentNull()
    {
      CopyOrMoveToDestinationMailboxArgumentNull(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationMailboxArgumentNull(bool move, ImapCapability[] capabilities)
    {
      TestMessageSet(3, capabilities, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        if (move)
          Assert.Throws<ArgumentNullException>(() => messageSet.MoveTo((ImapMailboxInfo)null));
        else
          Assert.Throws<ArgumentNullException>(() => messageSet.CopyTo((ImapMailboxInfo)null));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationMailbox_NonExistent()
    {
      CopyOrMoveToDestinationMailboxNonExistent(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_NonExistent()
    {
      CopyOrMoveToDestinationMailboxNonExistent(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationMailbox_NonExistent()
    {
      CopyOrMoveToDestinationMailboxNonExistent(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationMailboxNonExistent(bool move, ImapCapability[] capabilities)
    {
      TestMessageSet(10,
                     capabilities.Prepend(ImapCapability.ListExtended),
                     ImapSequenceSet.CreateUidRangeSet(1L, 10L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        server.EnqueueTaggedResponse("LIST ",
                                     "* LIST (\\Subscribed \\NonExistent) \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = messageSet.Client.GetMailbox("dest", ImapMailboxListOptions.SubscribedOnly);

        Assert.IsFalse(destMailbox.Exists);

        if (move)
          Assert.Throws<ImapProtocolViolationException>(() => messageSet.MoveTo(destMailbox));
        else
          Assert.Throws<ImapProtocolViolationException>(() => messageSet.CopyTo(destMailbox));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationMailbox_Unselectable()
    {
      CopyOrMoveToDestinationMailboxUnselectable(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_Unselectable()
    {
      CopyOrMoveToDestinationMailboxUnselectable(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationMailbox_Unselectable()
    {
      CopyOrMoveToDestinationMailboxUnselectable(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationMailboxUnselectable(bool move, ImapCapability[] capabilities)
    {
      TestMessageSet(10,
                     capabilities.Prepend(ImapCapability.ListExtended),
                     ImapSequenceSet.CreateUidRangeSet(1L, 10L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        server.EnqueueTaggedResponse("LIST ",
                                     "* LIST (\\Noselect) \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = messageSet.Client.GetMailbox("dest", ImapMailboxListOptions.SubscribedOnly);

        server.DequeueRequest();

        Assert.IsTrue(destMailbox.IsUnselectable);

        if (move)
          Assert.Throws<ImapProtocolViolationException>(() => messageSet.MoveTo(destMailbox));
        else
          Assert.Throws<ImapProtocolViolationException>(() => messageSet.CopyTo(destMailbox));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationMailbox_EmptySet()
    {
      CopyOrMoveToDestinationMailboxEmptySet(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_EmptySet()
    {
      CopyOrMoveToDestinationMailboxEmptySet(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationMailbox_EmptySet()
    {
      CopyOrMoveToDestinationMailboxEmptySet(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationMailboxEmptySet(bool move, ImapCapability[] capabilities)
    {
      TestMessageSet(0, capabilities, CreateEmptySet(),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        server.EnqueueTaggedResponse("LIST ",
                                     "* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = messageSet.Client.GetMailbox("dest");

        // do nothing
        if (move)
          messageSet.MoveTo(destMailbox);
        else
          messageSet.CopyTo(destMailbox);

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsTrue(messageSet.IsAvailable);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationMailbox_MailboxClosed()
    {
      CopyOrMoveToDestinationMailboxMailboxClosed(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_MailboxClosed()
    {
      CopyOrMoveToDestinationMailboxMailboxClosed(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationMailbox_MailboxClosed()
    {
      CopyOrMoveToDestinationMailboxMailboxClosed(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationMailboxMailboxClosed(bool move, ImapCapability[] capabilities)
    {
      TestMessageSet(3, capabilities, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        server.EnqueueTaggedResponse("LIST ",
                                     "* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = messageSet.Client.GetMailbox("dest");

        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK done\r\n");

        var mailbox = messageSet.Mailbox;

        mailbox.Close();

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsFalse(messageSet.IsAvailable);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messageSet.Mailbox));

        var ex = Assert.Throws<ImapUnavailableException>(delegate {
          if (move)
            messageSet.MoveTo(destMailbox);
          else
            messageSet.CopyTo(destMailbox);
        });

        Assert.IsFalse(messageSet.Client.IsBusy);

        Smdn.Net.TestUtils.SerializeBinary(ex);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationName()
    {
      TestMessageSet(3, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        CopyOrMoveToDestinationName(server, messageSet, messages, false, false);
      });
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName()
    {
      TestMessageSet(3, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        CopyOrMoveToDestinationName(server, messageSet, messages, true, false);
      });
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName_MoveDisabled()
    {
      TestMessageSet(3, new[] {ImapCapability.Move}, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        messageSet.Client.DisabledCapabilities.Add(ImapCapability.Move);

        CopyOrMoveToDestinationName(server, messageSet, messages, true, false);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationName_TryCreate()
    {
      TestMessageSet(3, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        CopyOrMoveToDestinationName(server, messageSet, messages, false, true);
      });
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName_TryCreate()
    {
      TestMessageSet(3, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        CopyOrMoveToDestinationName(server, messageSet, messages, true, true);
      });
    }

    private void CopyOrMoveToDestinationName(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages, bool move, bool tryCreate)
    {
      var responseTask = TestUtils2.QueueResponseTask(server, messageSet.Client, delegate {
        server.EnqueueTaggedResponse("UID COPY 1:3 dest\r\n",
                                     "$tag OK done\r\n");
        if (move)
          server.EnqueueTaggedResponse("UID STORE 1:3 +FLAGS (\\Deleted)\r\n",
                                       "* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                       "* FETCH 2 (FLAGS (\\Deleted \\Seen))\r\n" +
                                       "* FETCH 3 (FLAGS (\\Deleted $label1))\r\n" +
                                       "$tag OK done\r\n");
      });

      ImapMailboxInfo createdMailbox = null;

      if (tryCreate) {
        createdMailbox = move
          ? messageSet.MoveTo("dest", true)
          : messageSet.CopyTo("dest", true);
      }
      else {
        if (move)
          messageSet.MoveTo("dest");
        else
          messageSet.CopyTo("dest");
      }

      responseTask.Wait();

      Assert.IsFalse(messageSet.Client.IsBusy);

      if (tryCreate)
        Assert.IsNull(createdMailbox);

      if (move) {
        Assert.IsFalse(messageSet.IsAvailable);

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[0].IsMarkedAsDeleted);

        Assert.IsNotNull(messages[1].Flags);
        Assert.AreEqual(2, messages[1].Flags.Count);
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[1].IsMarkedAsDeleted);
        Assert.IsTrue(messages[1].IsSeen);

        Assert.IsNotNull(messages[2].Flags);
        Assert.AreEqual(2, messages[2].Flags.Count);
        Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[2].Flags.Contains("$label1"));
        Assert.IsTrue(messages[2].IsMarkedAsDeleted);
      }
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationName()
    {
      MoveToMoveCapableDestinationName(false);
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationName_NoTryCreateResponse()
    {
      MoveToMoveCapableDestinationName(true);
    }

    private void MoveToMoveCapableDestinationName(bool tryCreate)
    {
      TestMessageSet(3,
                     new[] {ImapCapability.Move},
                     ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        var responseTask = TestUtils2.QueueResponseTask(server, messageSet.Client, delegate {
          if (tryCreate) {
            server.EnqueueTaggedResponse("UID MOVE 1:3 dest\r\n",
                                         "$tag NO [TRYCREATE] mailbox not exist\r\n");
            server.EnqueueTaggedResponse("CREATE dest\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("UID MOVE 1:3 dest\r\n",
                                         "* EXPUNGE 3\r\n" +
                                         "* EXPUNGE 2\r\n" +
                                         "* EXPUNGE 1\r\n" +
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("LIST \"\" dest\r\n",
                                         "* LIST () \"\" dest\r\n" +
                                         "$tag OK done\r\n");
          }
          else {
            server.EnqueueTaggedResponse("UID MOVE 1:3 dest\r\n",
                                         "* EXPUNGE 3\r\n" +
                                         "* EXPUNGE 2\r\n" +
                                         "* EXPUNGE 1\r\n" +
                                         "$tag OK done\r\n");
          }
        });

        var sourceMailbox = messageSet.Mailbox;
        var createdMailbox = messageSet.MoveTo("dest", tryCreate);

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        if (tryCreate) {
          Assert.IsNotNull(createdMailbox);
          Assert.AreEqual("dest", createdMailbox.Name);
          Assert.AreEqual("dest", createdMailbox.FullName);
        }
        else {
          Assert.IsNull(createdMailbox);
        }

        Assert.IsFalse(messageSet.IsAvailable);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messageSet.Mailbox));
        Assert.AreEqual(0L, sourceMailbox.ExistMessageCount);

        Assert.IsFalse(messages[0].IsAvailable);
        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[0].Sequence);
        Assert.IsTrue(messages[0].IsDeleted);

        Assert.IsFalse(messages[1].IsAvailable);
        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[1].Sequence);
        Assert.IsTrue(messages[1].IsDeleted);

        Assert.IsFalse(messages[2].IsAvailable);
        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[2].Sequence);
        Assert.IsTrue(messages[2].IsDeleted);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationName_SameMailbox()
    {
      CopyOrMoveToDestinationNameSameMailbox(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName_SameMailbox()
    {
      CopyOrMoveToDestinationNameSameMailbox(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationName_SameMailbox()
    {
      CopyOrMoveToDestinationNameSameMailbox(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationNameSameMailbox(bool move, ImapCapability[] capabilities)
    {
      TestMessageSet(3, capabilities, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        Assert.AreEqual(3, messageSet.Mailbox.ExistMessageCount);
        Assert.AreEqual(0, messageSet.Mailbox.RecentMessageCount);

        if (move) {
          messageSet.MoveTo(messageSet.Mailbox.FullName);
        }
        else {
          var responseTask = TestUtils2.QueueResponseTask(server, messageSet.Client, delegate {
            server.EnqueueTaggedResponse("UID COPY 1:3 " + messageSet.Mailbox.FullName + "\r\n",
                                         "* 3 RECENT\r\n" +
                                         "* 6 EXISTS\r\n" +
                                         "$tag OK done\r\n");
          });

          messageSet.CopyTo(messageSet.Mailbox.FullName);

          responseTask.Wait();
        }

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsTrue(messageSet.IsAvailable);

        if (move) {
          Assert.AreEqual(3, messageSet.Mailbox.ExistMessageCount);
          Assert.AreEqual(0, messageSet.Mailbox.RecentMessageCount);
        }
        else {
          Assert.AreEqual(6, messageSet.Mailbox.ExistMessageCount);
          Assert.AreEqual(3, messageSet.Mailbox.RecentMessageCount);
        }
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationName_ArgumentEmpty()
    {
      CopyOrMoveToDestinationNameArgumentEmpty(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName_ArgumentEmpty()
    {
      CopyOrMoveToDestinationNameArgumentEmpty(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationName_ArgumentEmpty()
    {
      CopyOrMoveToDestinationNameArgumentEmpty(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationNameArgumentEmpty(bool move, ImapCapability[] capabilities)
    {
      TestMessageSet(3, capabilities, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        if (move)
          Assert.Throws<ArgumentException>(() => messageSet.MoveTo(string.Empty));
        else
          Assert.Throws<ArgumentException>(() => messageSet.CopyTo(string.Empty));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationName_ArgumentNull()
    {
      CopyOrMoveToDestinationNameArgumentNull(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName_ArgumentNull()
    {
      CopyOrMoveToDestinationNameArgumentNull(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationName_ArgumentNull()
    {
      CopyOrMoveToDestinationNameArgumentNull(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationNameArgumentNull(bool move, ImapCapability[] capabilities)
    {
      TestMessageSet(3, capabilities, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        if (move)
          Assert.Throws<ArgumentNullException>(() => messageSet.MoveTo((string)null));
        else
          Assert.Throws<ArgumentNullException>(() => messageSet.CopyTo((string)null));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationName_EmptySet()
    {
      CopyOrMoveToDestinationNameEmptySet(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName_EmptySet()
    {
      CopyOrMoveToDestinationNameEmptySet(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationName_EmptySet()
    {
      CopyOrMoveToDestinationNameEmptySet(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationNameEmptySet(bool move, ImapCapability[] capabilities)
    {
      TestMessageSet(0, capabilities, CreateEmptySet(),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        // do nothing
        if (move)
          messageSet.MoveTo("dest");
        else
          messageSet.CopyTo("dest");

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsTrue(messageSet.IsAvailable);
      });
    }

    [Test, Ignore("not implemented; uid validity")]
    public void TestCopyTo_CopyToDestinationName_UidValidityChanged()
    {
    }

    [Test, Ignore("not implemented; uid validity")]
    public void TestMoveTo_MoveToDestinationName_UidValidityChanged()
    {
    }

    [Test, Ignore("not implemented; uid validity")]
    public void TestMoveTo_MoveCapable_MoveToDestinationName_UidValidityChanged()
    {
    }

    [Test]
    public void TestCopyTo_CopyToDestinationName_MailboxClosed()
    {
      CopyOrMoveToDestinationNameMailboxClosed(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName_MailboxClosed()
    {
      CopyOrMoveToDestinationNameMailboxClosed(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationName_MailboxClosed()
    {
      CopyOrMoveToDestinationNameMailboxClosed(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationNameMailboxClosed(bool move, ImapCapability[] capabilities)
    {
      TestMessageSet(10, capabilities, ImapSequenceSet.CreateUidRangeSet(1L, 10L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK done\r\n");

        var mailbox = messageSet.Mailbox;

        mailbox.Close();

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsFalse(messageSet.IsAvailable);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messageSet.Mailbox));

        var ex = Assert.Throws<ImapUnavailableException>(delegate {
          if (move)
            messageSet.MoveTo("dest");
          else
            messageSet.CopyTo("dest");
        });

        Assert.IsFalse(messageSet.Client.IsBusy);

        Smdn.Net.TestUtils.SerializeBinary(ex);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationName_NoResponse()
    {
      CopyOrMoveToDestinationNameNoResponse(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName_NoResponse()
    {
      CopyOrMoveToDestinationNameNoResponse(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationName_NoResponse()
    {
      CopyOrMoveToDestinationNameNoResponse(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationNameNoResponse(bool move, ImapCapability[] capabilities)
    {
      TestMessageSet(10, capabilities, ImapSequenceSet.CreateUidRangeSet(1L, 10L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        var responseTask = TestUtils2.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse((move && 0 < capabilities.Length) ? "UID MOVE 1:10 dest\r\n" : "UID COPY 1:10 dest\r\n",
                                       "$tag NO error occurred\r\n");
        });

        if (move)
          Assert.Throws<ImapErrorResponseException>(() => messageSet.MoveTo("dest", false));
        else
          Assert.Throws<ImapErrorResponseException>(() => messageSet.CopyTo("dest", false));

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationName_NoResponse_DontTryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseDontTryCreate(false, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName_NoResponse_DontTryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseDontTryCreate(true, EmptyArray<ImapCapability>.Instance);
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationName_NoResponse_DontTryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseDontTryCreate(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationNameNoResponseDontTryCreate(bool move, ImapCapability[] capabilities)
    {
      TestMessageSet(10, capabilities, ImapSequenceSet.CreateUidRangeSet(1L, 10L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        var responseTask = TestUtils2.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse((move && 0 < capabilities.Length) ? "UID MOVE 1:10 dest\r\n" : "UID COPY 1:10 dest\r\n",
                                       "$tag NO [TRYCREATE] mailbox not exist\r\n");
        });

        var ex = Assert.Throws<ImapMailboxNotFoundException>(() => {
          if (move)
            messageSet.MoveTo("dest", false);
          else
            messageSet.CopyTo("dest", false);
        });

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(ex.Mailbox);
        Assert.AreEqual("dest", ex.Mailbox);

        Smdn.Net.TestUtils.SerializeBinary(ex, delegate(ImapMailboxNotFoundException deserialized) {
          Assert.IsNotNull(deserialized.Mailbox);
          Assert.AreEqual(ex.Mailbox, deserialized.Mailbox);
        });


        Assert.IsTrue(messageSet.IsAvailable);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationName_NoResponse_TryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseTryCreate(false);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName_NoResponse_TryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseTryCreate(true);
    }

    private void CopyOrMoveToDestinationNameNoResponseTryCreate(bool move)
    {
      TestMessageSet(3, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        var responseTask = TestUtils2.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("UID COPY 1:3 dest\r\n",
                                       "$tag NO [TRYCREATE] mailbox not exist\r\n");
          server.EnqueueTaggedResponse("CREATE dest\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("UID COPY 1:3 dest\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("LIST \"\" dest\r\n",
                                       "* LIST () \"\" dest\r\n" +
                                       "$tag OK done\r\n");
          if (move)
            server.EnqueueTaggedResponse("UID STORE 1:3 +FLAGS (\\Deleted)\r\n",
                                         "* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                         "* FETCH 2 (FLAGS (\\Deleted \\Seen))\r\n" +
                                         "* FETCH 3 (FLAGS (\\Deleted $label1))\r\n" +
                                         "$tag OK done\r\n");
        });

        var createdMailbox = move
          ? messageSet.MoveTo("dest", true)
          : messageSet.CopyTo("dest", true);

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(createdMailbox);
        Assert.AreEqual("dest", createdMailbox.Name);
        Assert.AreEqual("dest", createdMailbox.FullName);

        if (move) {
          Assert.IsFalse(messageSet.IsAvailable);

          Assert.IsNotNull(messages[0].Flags);
          Assert.AreEqual(1, messages[0].Flags.Count);
          Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Deleted));
          Assert.IsTrue(messages[0].IsMarkedAsDeleted);

          Assert.IsNotNull(messages[1].Flags);
          Assert.AreEqual(2, messages[1].Flags.Count);
          Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Deleted));
          Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Seen));
          Assert.IsTrue(messages[1].IsMarkedAsDeleted);
          Assert.IsTrue(messages[1].IsSeen);

          Assert.IsNotNull(messages[2].Flags);
          Assert.AreEqual(2, messages[2].Flags.Count);
          Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Deleted));
          Assert.IsTrue(messages[2].Flags.Contains("$label1"));
          Assert.IsTrue(messages[2].IsMarkedAsDeleted);
        }
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationName_NoResponse_TryCreate_NameContainsWildcard()
    {
      CopyOrMoveToDestinationNameNoResponseTryCreateNameContainsWildcard(true);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName_NoResponse_TryCreate_NameContainsWildcard()
    {
      CopyOrMoveToDestinationNameNoResponseTryCreateNameContainsWildcard(false);
    }

    private void CopyOrMoveToDestinationNameNoResponseTryCreateNameContainsWildcard(bool move)
    {
      TestMessageSet(3, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     delegate(ImapPseudoServer2 server, ImapMessageInfoBase messageSet, ImapMessageInfo[] messages) {
        var responseTask = TestUtils2.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("UID COPY 1:3 \"Box*\"\r\n",
                                       "$tag NO [TRYCREATE] mailbox not exist\r\n");
          server.EnqueueTaggedResponse("CREATE \"Box*\"\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("UID COPY 1:3 \"Box*\"\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("LIST \"\" Box*\r\n",
                                       "* LIST (\\Marked) \".\" Box*\r\n" +
                                       "* LIST () \".\" Box\r\n" +
                                       "* LIST () \".\" Box.Child\r\n" +
                                       "* LIST () \".\" Boxes\r\n" +
                                       "$tag OK done\r\n");
          if (move)
            server.EnqueueTaggedResponse("UID STORE 1:3 +FLAGS (\\Deleted)\r\n",
                                         "* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                         "* FETCH 2 (FLAGS (\\Deleted \\Seen))\r\n" +
                                         "* FETCH 3 (FLAGS (\\Deleted $label1))\r\n" +
                                         "$tag OK done\r\n");
        });

        var createdMailbox = move
          ? messageSet.MoveTo("Box*", true)
          : messageSet.CopyTo("Box*", true);

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(createdMailbox);
        Assert.AreEqual("Box*", createdMailbox.Name);
        Assert.AreEqual("Box*", createdMailbox.FullName);
        Assert.AreEqual(1, createdMailbox.Flags.Count);
        CollectionAssert.Contains(createdMailbox.Flags, ImapMailboxFlag.Marked);

        if (move) {
          Assert.IsFalse(messageSet.IsAvailable);

          Assert.IsNotNull(messages[0].Flags);
          Assert.AreEqual(1, messages[0].Flags.Count);
          Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Deleted));
          Assert.IsTrue(messages[0].IsMarkedAsDeleted);

          Assert.IsNotNull(messages[1].Flags);
          Assert.AreEqual(2, messages[1].Flags.Count);
          Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Deleted));
          Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Seen));
          Assert.IsTrue(messages[1].IsMarkedAsDeleted);
          Assert.IsTrue(messages[1].IsSeen);

          Assert.IsNotNull(messages[2].Flags);
          Assert.AreEqual(2, messages[2].Flags.Count);
          Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Deleted));
          Assert.IsTrue(messages[2].Flags.Contains("$label1"));
          Assert.IsTrue(messages[2].IsMarkedAsDeleted);
        }
      });
    }
  }
}
