#if NET_4_0
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using NUnit.Framework;

using Smdn.IO;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapMessageInfoOperationsFetchAsyncTests {
    private void TestFetchMessage(Action<ImapPseudoServer2, ImapMessageInfo, string, int> action)
    {
      TestFetchMessage(false, string.Empty, action);
    }

    private void TestFetchMessage(bool setSeen,
                                  string initialFlags,
                                  Action<ImapPseudoServer2, ImapMessageInfo, string, int> action)
    {
      const string messageBody = @"Return-Path: <test@example.net>
X-Original-To: test@example.net
Delivered-To: test@example.net
Message-ID: <4BAA561E.5000004@test@example.net>
Date: Thu, 25 Mar 2010 03:12:46 +0900
From: test <test@example.net>
Reply-To: test@example.net
User-Agent: Thunderbird 2.0.0.24 (X11/20100317)
MIME-Version: 1.0
To: hoge@example.net
Subject: =?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=
Content-Type: text/plain; charset=ISO-2022-JP
Content-Transfer-Encoding: 7bit

test message

3 uid fetch 6 (body[])
* 6 FETCH (UID 6 BODY[] {1476}
hogehogehoge
)
3 OK Fetch completed.
";

      TestUtils2.TestOpenedMailbox("INBOX",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        server.EnqueueTaggedResponse("UID FETCH 1 (UID FLAGS)\r\n",
                                     string.Format("* FETCH 1 (UID 1 FLAGS ({0}))\r\n", initialFlags ?? string.Empty) +
                                     "$tag OK done\r\n");

        var message = mailbox.GetMessageByUid(1L, ImapMessageFetchAttributeOptions.DynamicAttributes);

        server.DequeueRequest(); // FETCH

        var octets = NetworkTransferEncoding.Transfer8Bit.GetByteCount(messageBody);

        server.EnqueueTaggedResponse(string.Format("* {0} FETCH (RFC822.SIZE {1} {3}BODY[] {{{1}}}\r\n{2})\r\n",
                                                   message.Uid,
                                                   octets,
                                                   messageBody,
                                                   setSeen ? "FLAGS (\\Seen) " : string.Empty) +
                                     "$tag OK done\r\n");

        action(server, message, messageBody, octets);
      });
    }

    [Test]
    public void TestOpenReadAsync()
    {
      TestFetchMessage(false,
                       null,
                       delegate(ImapPseudoServer2 server,
                                ImapMessageInfo message,
                                string expectedMessageBody,
                                int expectedOctets) {
        var t = message.OpenReadAsync();

        Assert.IsTrue(message.Client.IsBusy, "IsBusy; task started");

        Assert.IsFalse(t.IsCanceled, "task.IsCanceled #1");
        Assert.IsFalse(t.IsFaulted, "task.IsFaulted #1");

        using (var stream = t.Result) {
          Assert.That(server.DequeueRequest(), Is.StringEnding("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n"));

          Assert.IsFalse(t.IsCanceled, "task.IsCanceled #2");
          Assert.IsFalse(t.IsFaulted, "task.IsFaulted #2");
          Assert.IsTrue(t.IsCompleted, "task.IsCompleted #2");

          Assert.IsNotNull(stream);
          Assert.AreEqual(expectedOctets, stream.Length);
          Assert.AreEqual(expectedMessageBody,
                          (new StreamReader(stream, NetworkTransferEncoding.Transfer8Bit)).ReadToEnd());

          Assert.IsFalse(message.Flags.Contains(ImapMessageFlag.Seen));
          Assert.IsFalse(message.IsSeen);

          Assert.IsTrue(message.Client.IsBusy, "IsBusy; still open");
        }

        Assert.IsFalse(message.Client.IsBusy, "IsBusy; after close");
      });
    }

    [Test]
    public void TestOpenReadAsync_Overloads()
    {
#if NET_4_5
      var progress = (IProgress<ImapMessageFetchProgressInfo>)null;
      var openReadAsyncMethods = new[] {
        new Func<ImapMessageInfo, Task<Stream>>(m => m.OpenReadAsync()),
        new Func<ImapMessageInfo, Task<Stream>>(m => m.OpenReadAsync(progress)),
        new Func<ImapMessageInfo, Task<Stream>>(m => m.OpenReadAsync(CancellationToken.None)),
        new Func<ImapMessageInfo, Task<Stream>>(m => m.OpenReadAsync(CancellationToken.None, progress)),
        new Func<ImapMessageInfo, Task<Stream>>(m => m.OpenReadAsync(ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, Task<Stream>>(m => m.OpenReadAsync(ImapMessageFetchBodyOptions.Default, progress)),
        new Func<ImapMessageInfo, Task<Stream>>(m => m.OpenReadAsync(ImapMessageFetchBodyOptions.Default, CancellationToken.None)),
        new Func<ImapMessageInfo, Task<Stream>>(m => m.OpenReadAsync(ImapMessageFetchBodyOptions.Default, CancellationToken.None, progress)),
      };
#else
      var openReadAsyncMethods = new[] {
        new Func<ImapMessageInfo, Task<Stream>>(m => m.OpenReadAsync()),
        new Func<ImapMessageInfo, Task<Stream>>(m => m.OpenReadAsync(CancellationToken.None)),
        new Func<ImapMessageInfo, Task<Stream>>(m => m.OpenReadAsync(ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, Task<Stream>>(m => m.OpenReadAsync(ImapMessageFetchBodyOptions.Default, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < openReadAsyncMethods.Length; i++) {
        TestFetchMessage(false,
                       null,
                       delegate(ImapPseudoServer2 server,
                                ImapMessageInfo message,
                                string expectedMessageBody,
                                int expectedOctets) {
          try {
            var ret = openReadAsyncMethods[i](message)
              .ContinueWith(t => { using (var r = new StreamReader(t.Result, NetworkTransferEncoding.Transfer8Bit)) return r.ReadToEnd(); })
              .Result;

            Assert.AreEqual(expectedMessageBody, ret);
          }
          catch (Exception ex) {
            Assert.Fail("unexpected exception (overload #{0}) : {1}", i, ex);
          }

          Assert.IsFalse(message.Client.IsBusy);
        });
      }
    }

    [Test]
    public void TestOpenReadAsync_CancellationRequestedBeforeStart()
    {
      TestUtils2.TestMessage(delegate(ImapPseudoServer2 server, ImapMessageInfo message) {
        var cts = new CancellationTokenSource();

        cts.Cancel();

        var t = message.OpenReadAsync(cts.Token);

        Assert.IsFalse(message.Client.IsBusy);

        var ex = Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsInstanceOf<OperationCanceledException>(ex.InnerException);

        Assert.IsNull(server.DequeueRequest()); // nothing requested

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenReadAsync_CancellationRequestedBeforeRead()
    {
      TestFetchMessage(false,
                       null,
                       delegate(ImapPseudoServer2 server,
                                ImapMessageInfo message,
                                string expectedMessageBody,
                                int expectedOctets) {
        var cts = new CancellationTokenSource();
        var t = message.OpenReadAsync(cts.Token);

        using (var stream = t.Result) {
          cts.Cancel();

          Assert.Throws<OperationCanceledException>(() => stream.ReadByte());

          Assert.IsTrue(message.Client.IsBusy, "IsBusy; still open");
        }

        Assert.IsFalse(message.Client.IsBusy, "IsBusy; after close");
      });
    }

    private class MessageGenerator {
      private ImapPseudoServer2 server;
      private ImapMessageInfo message;
      private string messageBody;
      private readonly Action<int> actionBeforeRespond;
      private readonly Action<int> actionAfterRespond;

      public int MessageBodyOctet {
        get { return NetworkTransferEncoding.Transfer8Bit.GetByteCount(messageBody); }
      }

      public MessageGenerator(ImapPseudoServer2 server,
                              ImapMessageInfo message,
                              int maxBlockCount)
        : this(server,
               message,
               maxBlockCount,
               null,
               null)
      {
      }

      public MessageGenerator(ImapPseudoServer2 server,
                              ImapMessageInfo message,
                              int maxBlockCount,
                              Action<int> actionBeforeRespond,
                              Action<int> actionAfterRespond)
      {
        this.server = server;
        this.message = message;
        this.actionBeforeRespond = actionBeforeRespond;
        this.actionAfterRespond = actionAfterRespond;

        const string messageHeader = @"Return-Path: <test@example.net>
X-Original-To: test@example.net
Delivered-To: test@example.net
Message-ID: <4BAA561E.5000004@test@example.net>
Date: Thu, 25 Mar 2010 03:12:46 +0900
From: test <test@example.net>
Reply-To: test@example.net
User-Agent: Thunderbird 2.0.0.24 (X11/20100317)
MIME-Version: 1.0
To: hoge@example.net
Subject: =?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=
Content-Type: text/plain; charset=ISO-2022-JP
Content-Transfer-Encoding: 7bit

";
        var messageContent = new string('x', (maxBlockCount - 1) * Smdn.Net.Imap4.Client.Session.ImapFetchMessageBodyStream.DefaultFetchBlockSize);

        messageBody = messageHeader + messageContent;
      }

      public IEnumerable<string> EnqueueNextBlock()
      {
        var messageBodyBytes = NetworkTransferEncoding.Transfer8Bit.GetBytes(messageBody);
        var position = 0L;

        for (var blockIndex = 0; ; blockIndex++) {
          var lengthToEnqueue = Math.Min(messageBodyBytes.Length - position,
                                         Smdn.Net.Imap4.Client.Session.ImapFetchMessageBodyStream.DefaultFetchBlockSize);
          var enqueueingMessageBody = NetworkTransferEncoding.Transfer8Bit.GetString(messageBodyBytes, (int)position, (int)lengthToEnqueue);

          string responseText;

          if (blockIndex == 0) {
            responseText = string.Format("* {0} FETCH (RFC822.SIZE {1} BODY[] {{{2}}}\r\n{3})\r\n",
                                         message.Uid,
                                         messageBodyBytes.Length,
                                         lengthToEnqueue,
                                         enqueueingMessageBody) +
                           "$tag OK done\r\n";
          }
          else {
            responseText = string.Format("* {0} FETCH (BODY[] {{{1}}}\r\n{2})\r\n",
                                         message.Uid,
                                         lengthToEnqueue,
                                         enqueueingMessageBody) +
                           "$tag OK done\r\n";
          }

          server.EnqueueTaggedResponse(null,
                                       responseText,
                                       () => { if (actionBeforeRespond != null) actionBeforeRespond(blockIndex); },
                                       () => { if (actionAfterRespond != null) actionAfterRespond(blockIndex); });

          position += Smdn.Net.Imap4.Client.Session.ImapFetchMessageBodyStream.DefaultFetchBlockSize;

          if (messageBodyBytes.Length <= position)
            yield break;
          else
            yield return enqueueingMessageBody;
        }
      }
    }

    [Test]
    public void TestOpenReadAsync_CancellationRequestedWhileFetching()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        // FETCH
        server.EnqueueTaggedResponse(string.Format("* FETCH 1 (UID 1 FLAGS ({0}))\r\n", string.Empty) +
                                     "$tag OK done\r\n");

        var message = mailbox.GetMessageByUid(1L, ImapMessageFetchAttributeOptions.DynamicAttributes);

        Assert.That(server.DequeueRequest(), Is.StringEnding("UID FETCH 1 (UID FLAGS)\r\n"));

        var generator = new MessageGenerator(server, message, 3);
        var cts = new CancellationTokenSource();
        var buffer = new byte[Smdn.Net.Imap4.Client.Session.ImapFetchMessageBodyStream.DefaultFetchBlockSize];

        var task = message.OpenReadAsync(cts.Token).ContinueWith(t => {
          using (var stream = t.Result) {
            // read #1 block (-1 bytes), fetching #1 block
            Assert.AreEqual(buffer.Length - 1,
                            stream.Read(buffer, 0, buffer.Length - 1));

            // read last 1 byte of #1 block and #2 block (-1 bytes), fetching #2 block
            Assert.AreEqual(buffer.Length,
                            stream.Read(buffer, 0, buffer.Length));

            // try to read last 1 byte of #2 block and #3 block (-1 bytes), and throws OperationCanceledException
            stream.Read(buffer, 0, buffer.Length);
          }
        });

        var block = 0;

        foreach (var m in generator.EnqueueNextBlock()) {
          if (block == 1) {
            cts.Cancel();
            break;
          }

          block++;
        }

        var ex = Assert.Throws<AggregateException>(() => task.Wait());

        Assert.IsInstanceOf<OperationCanceledException>(ex.Flatten().InnerException);

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenReadAsync_MailboxClosed()
    {
      TestUtils2.TestMessageClosedMailbox(delegate(ImapPseudoServer2 server, ImapMessageInfo message) {
        var ex = Assert.Throws<ImapUnavailableException>(() => message.OpenReadAsync());

        Smdn.Net.TestUtils.SerializeBinary(ex);

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenReadAsync_VanishedMessage()
    {
      TestUtils2.TestMessageVanished(delegate(ImapPseudoServer2 server, ImapMessageInfo message) {
        var ex = Assert.Throws<ImapUnavailableException>(() => message.OpenReadAsync());

        Smdn.Net.TestUtils.SerializeBinary(ex);

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenReadAsync_MessageVanishingByFetch()
    {
      TestUtils2.TestMessage(delegate(ImapPseudoServer2 server, ImapMessageInfo message) {
        Assert.AreNotEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, message.Sequence);
        Assert.IsFalse(message.IsDeleted);

        // FETCH
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        var t = message.OpenReadAsync();

        var ex = Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsInstanceOf<ImapMessageVanishedException>(ex.InnerException);

        Assert.IsFalse(message.Client.IsBusy);

        var exMessageVanished = ex.InnerException as ImapMessageVanishedException;

        Assert.IsNotNull(exMessageVanished.VanishedMessage);
        Assert.AreSame(message, exMessageVanished.VanishedMessage);

        Assert.That(server.DequeueRequest(), Is.StringEnding("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n"));

        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, message.Sequence);
        Assert.IsTrue(message.IsDeleted);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(message.Mailbox));
        Assert.IsFalse(message.IsAvailable);
      });
    }

#if NET_4_5
    [Test]
    [TestCase(1)]
    [TestCase(2)]
    [TestCase(3)]
    [TestCase(50)]
    public void TestOpenReadAsync_ReportProgress(int fetchBlockCount)
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        // FETCH
        server.EnqueueTaggedResponse(string.Format("* FETCH 1 (UID 1 FLAGS ({0}))\r\n", string.Empty) +
                                     "$tag OK done\r\n");

        var message = mailbox.GetMessageByUid(1L, ImapMessageFetchAttributeOptions.DynamicAttributes);

        var generator = new MessageGenerator(server, message, fetchBlockCount);
        var cts = new CancellationTokenSource();
        var lastProgressPercentage = int.MinValue;
        var reportedCount = 0;
        var reportedFinish = false;
        var progress = new Progress<ImapMessageFetchProgressInfo>(p => {
          Interlocked.Increment(ref reportedCount);

          if (p.FetchedOctetCount == p.TotalOctetCount)
            reportedFinish = true;

          Assert.AreEqual(cts.Token, p.CancellationToken, "CancellationToken #{0}", reportedCount);
          Assert.AreEqual(generator.MessageBodyOctet, p.TotalOctetCount, "Total #{0}", reportedCount);
          Assert.GreaterOrEqual(p.ProgressPercentage, 0, "0 <= ProgressPercentage #{0}", reportedCount);
          Assert.LessOrEqual(p.ProgressPercentage, 100, "ProgressPercentage <= 100 #{0}", reportedCount);

          if (SynchronizationContext.Current != null && lastProgressPercentage < 100)
            Assert.Greater(p.ProgressPercentage, lastProgressPercentage, "lastProgressPercentage < ProgressPercentage #{0}", reportedCount);

          Interlocked.Exchange(ref lastProgressPercentage, p.ProgressPercentage);
        });

        var task = message.OpenReadAsync(cts.Token, progress).ContinueWith(t => {t.Result.ReadToEnd(); t.Result.Close();});

        server.WaitForClientRequest(TimeSpan.FromSeconds(3.0));

        foreach (var m in generator.EnqueueNextBlock()) {
        }

        Assert.DoesNotThrow(() => task.Wait());

        Assert.IsFalse(message.Client.IsBusy);

        Thread.Sleep(100); // wait for progress task

        Assert.AreEqual(fetchBlockCount, reportedCount, "reported count");
        Assert.IsTrue(reportedFinish, "reported finish");
      });
    }
#endif

    [Test]
    public void TestReadAsAsync_StreamConverter()
    {
      var converter = new Converter<Stream, byte[]>(s => s.ReadToEnd());
#if NET_4_5
      var progress = (IProgress<ImapMessageFetchProgressInfo>)null;
      var readAsAsyncMethods = new[] {
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(converter)),
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(converter, progress)),
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(converter, CancellationToken.None)),
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(converter, CancellationToken.None, progress)),
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(ImapMessageFetchBodyOptions.Default, converter)),
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(ImapMessageFetchBodyOptions.Default, converter, progress)),
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(ImapMessageFetchBodyOptions.Default, converter, CancellationToken.None)),
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(ImapMessageFetchBodyOptions.Default, converter, CancellationToken.None, progress)),
      };
#else
      var readAsAsyncMethods = new[] {
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(converter)),
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(converter, CancellationToken.None)),
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(ImapMessageFetchBodyOptions.Default, converter)),
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(ImapMessageFetchBodyOptions.Default, converter, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < readAsAsyncMethods.Length; i++) {
        TestFetchMessage(false,
                         null,
                         delegate(ImapPseudoServer2 server,
                                ImapMessageInfo message,
                                string expectedMessageBody,
                                int expectedOctets) {
          try {
            var ret = readAsAsyncMethods[i](message).Result;

            Assert.IsNotNull(ret);
            Assert.AreEqual(expectedOctets, ret.Length);
            Assert.AreEqual(expectedMessageBody,
                            NetworkTransferEncoding.Transfer8Bit.GetString(ret));
          }
          catch (Exception ex) {
            Assert.Fail("unexpected exception (overload #{0}) : {1}", i, ex);
          }

          Assert.IsFalse(message.Client.IsBusy);
        });
      }
    }

    [Test]
    public void TestReadAsAsync_StreamConverter_ConverterNull()
    {
      var converter = (Converter<Stream, byte[]>)null;

      TestUtils2.TestMessage(delegate(ImapPseudoServer2 server, ImapMessageInfo message) {
        Assert.Throws<ArgumentNullException>(() => message.ReadAsAsync(converter));

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAllBytesAsync()
    {
#if NET_4_5
      var progress = (IProgress<ImapMessageFetchProgressInfo>)null;
      var readAllBytesAsyncMethods = new[] {
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync()),
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(progress)),
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(CancellationToken.None)),
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(CancellationToken.None, progress)),
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(ImapMessageFetchBodyOptions.Default, progress)),
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(ImapMessageFetchBodyOptions.Default, CancellationToken.None)),
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(ImapMessageFetchBodyOptions.Default, CancellationToken.None, progress)),
      };
#else
      var readAllBytesAsyncMethods = new[] {
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync()),
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(CancellationToken.None)),
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(ImapMessageFetchBodyOptions.Default, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < readAllBytesAsyncMethods.Length; i++) {
        TestFetchMessage(false,
                         null,
                         delegate(ImapPseudoServer2 server,
                                  ImapMessageInfo message,
                                  string expectedMessageBody,
                                  int expectedOctets) {
          try {
            var ret = readAllBytesAsyncMethods[i](message).Result;

            Assert.IsNotNull(ret);
            Assert.AreEqual(expectedOctets, ret.Length);
            Assert.AreEqual(expectedMessageBody,
                            NetworkTransferEncoding.Transfer8Bit.GetString(ret));
          }
          catch (Exception ex) {
            Assert.Fail("unexpected exception (overload #{0}) : {1}", i, ex);
          }

          Assert.IsFalse(message.Client.IsBusy);
        });
      }
    }

    [Test]
    public void TestWriteToAsync_Stream()
    {
#if NET_4_5
      var progress = (IProgress<ImapMessageFetchProgressInfo>)null;
      var writeToAsyncMethods = new[] {
        new Func<ImapMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s)),
        new Func<ImapMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s, progress)),
        new Func<ImapMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s, CancellationToken.None)),
        new Func<ImapMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s, CancellationToken.None, progress)),
        new Func<ImapMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s, ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s, ImapMessageFetchBodyOptions.Default, progress)),
        new Func<ImapMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s, ImapMessageFetchBodyOptions.Default, CancellationToken.None)),
        new Func<ImapMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s, ImapMessageFetchBodyOptions.Default, CancellationToken.None, progress)),
      };
#else
      var writeToAsyncMethods = new[] {
        new Func<ImapMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s)),
        new Func<ImapMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s, CancellationToken.None)),
        new Func<ImapMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s, ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s, ImapMessageFetchBodyOptions.Default, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < writeToAsyncMethods.Length; i++) {
        TestFetchMessage(false,
                         null,
                         delegate(ImapPseudoServer2 server,
                                  ImapMessageInfo message,
                                  string expectedMessageBody,
                                  int expectedOctets) {
          try {
            using (var outputStream = new MemoryStream()) {
              writeToAsyncMethods[i](message, outputStream).Wait();

              Assert.AreEqual(expectedOctets, outputStream.Length);

              outputStream.Close();

              Assert.AreEqual(NetworkTransferEncoding.Transfer8Bit.GetBytes(expectedMessageBody),
                              outputStream.ToArray());
            }
          }
          catch (Exception ex) {
            Assert.Fail("unexpected exception (overload #{0}) : {1}", i, ex);
          }

          Assert.IsFalse(message.Client.IsBusy);
        });
      }
    }

    [Test]
    public void TestWriteToAsync_Stream_DestinationNull()
    {
      TestUtils2.TestMessage(delegate(ImapPseudoServer2 server, ImapMessageInfo message) {
        Assert.Throws<ArgumentNullException>(() => message.WriteToAsync((Stream)null));

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestWriteToAsync_Stream_DestinationNotWritable()
    {
      TestUtils2.TestMessage(delegate(ImapPseudoServer2 server, ImapMessageInfo message) {
        using (var stream = new MemoryStream(new byte[0], false)) {
          Assert.Throws<ArgumentException>(() => message.WriteToAsync(stream));

          Assert.IsFalse(message.Client.IsBusy);
        }
      });
    }

    [Test]
    public void TestWriteToAsync_BinaryWriter()
    {
#if NET_4_5
      var progress = (IProgress<ImapMessageFetchProgressInfo>)null;
      var writeToAsyncMethods = new[] {
        new Func<ImapMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w)),
        new Func<ImapMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w, progress)),
        new Func<ImapMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w, CancellationToken.None)),
        new Func<ImapMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w, CancellationToken.None, progress)),
        new Func<ImapMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w, ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w, ImapMessageFetchBodyOptions.Default, progress)),
        new Func<ImapMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w, ImapMessageFetchBodyOptions.Default, CancellationToken.None)),
        new Func<ImapMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w, ImapMessageFetchBodyOptions.Default, CancellationToken.None, progress)),
      };
#else
      var writeToAsyncMethods = new[] {
        new Func<ImapMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w)),
        new Func<ImapMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w, CancellationToken.None)),
        new Func<ImapMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w, ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w, ImapMessageFetchBodyOptions.Default, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < writeToAsyncMethods.Length; i++) {
        TestFetchMessage(false,
                         null,
                         delegate(ImapPseudoServer2 server,
                                  ImapMessageInfo message,
                                  string expectedMessageBody,
                                  int expectedOctets) {
          try {
            using (var outputStream = new MemoryStream()) {
              var writer = new BinaryWriter(outputStream);

              writeToAsyncMethods[i](message, writer).Wait();

              Assert.AreEqual(expectedOctets, outputStream.Length);

              outputStream.Close();

              Assert.AreEqual(NetworkTransferEncoding.Transfer8Bit.GetBytes(expectedMessageBody),
                              outputStream.ToArray());
            }
          }
          catch (Exception ex) {
            Assert.Fail("unexpected exception (overload #{0}) : {1}", i, ex);
          }

          Assert.IsFalse(message.Client.IsBusy);
        });
      }
    }

    [Test]
    public void TestWriteToAsync_BinaryWriter_DestinationNull()
    {
      TestUtils2.TestMessage(delegate(ImapPseudoServer2 server, ImapMessageInfo message) {
        Assert.Throws<ArgumentNullException>(() => message.WriteToAsync((BinaryWriter)null));

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestSaveAsync()
    {
#if NET_4_5
      var progress = (IProgress<ImapMessageFetchProgressInfo>)null;
      var saveAsyncMethods = new[] {
        new Func<ImapMessageInfo, string, Task>((m, f) => m.SaveAsync(f)),
        new Func<ImapMessageInfo, string, Task>((m, f) => m.SaveAsync(f, progress)),
        new Func<ImapMessageInfo, string, Task>((m, f) => m.SaveAsync(f, CancellationToken.None)),
        new Func<ImapMessageInfo, string, Task>((m, f) => m.SaveAsync(f, CancellationToken.None, progress)),
        new Func<ImapMessageInfo, string, Task>((m, f) => m.SaveAsync(f, ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, string, Task>((m, f) => m.SaveAsync(f, ImapMessageFetchBodyOptions.Default, progress)),
        new Func<ImapMessageInfo, string, Task>((m, f) => m.SaveAsync(f, ImapMessageFetchBodyOptions.Default, CancellationToken.None)),
        new Func<ImapMessageInfo, string, Task>((m, f) => m.SaveAsync(f, ImapMessageFetchBodyOptions.Default, CancellationToken.None, progress)),
      };
#else
      var saveAsyncMethods = new[] {
        new Func<ImapMessageInfo, string, Task>((m, f) => m.SaveAsync(f)),
        new Func<ImapMessageInfo, string, Task>((m, f) => m.SaveAsync(f, CancellationToken.None)),
        new Func<ImapMessageInfo, string, Task>((m, f) => m.SaveAsync(f, ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, string, Task>((m, f) => m.SaveAsync(f, ImapMessageFetchBodyOptions.Default, CancellationToken.None)),
      };
#endif
      const string outfile = "fetched.eml";

      for (var i = 0; i < saveAsyncMethods.Length; i++) {
        try {
          File.Delete(outfile);

          TestFetchMessage(false,
                           null,
                           delegate(ImapPseudoServer2 server,
                                  ImapMessageInfo message,
                                  string expectedMessageBody,
                                  int expectedOctets) {
            try {
              saveAsyncMethods[i](message, outfile).Wait();

              Assert.IsTrue(File.Exists(outfile));
              Assert.AreEqual(expectedOctets, (new FileInfo(outfile)).Length);
              Assert.AreEqual(NetworkTransferEncoding.Transfer8Bit.GetBytes(expectedMessageBody),
                              File.ReadAllBytes(outfile));
            }
            catch (Exception ex) {
              Assert.Fail("unexpected exception (overload #{0}) : {1}", i, ex);
            }

            Assert.IsFalse(message.Client.IsBusy);
          });
        }
        finally {
          File.Delete(outfile);
        }
      }
    }

    [Test]
    public void TestSaveAsync_CancellationRequestedBeforeStart()
    {
      const string outfile = "fetched.eml";

      TestUtils2.TestMessage(delegate(ImapPseudoServer2 server, ImapMessageInfo message) {
        var cts = new CancellationTokenSource();

        cts.Cancel();

        var t = message.SaveAsync(outfile, cts.Token);

        var ex = Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsInstanceOf<OperationCanceledException>(ex.Flatten().InnerException);

        Assert.IsFalse(File.Exists(outfile), "file existence");

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestSaveAsync_CancellationRequestedBeforeCopyTaskRunning()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        const string outfile = "fetched.eml";

        try {
          File.Delete(outfile);

          server.EnqueueTaggedResponse("UID FETCH 1 (UID FLAGS)\r\n",
                                       string.Format("* FETCH 1 (UID 1 FLAGS ({0}))\r\n", string.Empty) +
                                       "$tag OK done\r\n");

          var message = mailbox.GetMessageByUid(1L, ImapMessageFetchAttributeOptions.DynamicAttributes);

          server.DequeueRequest(); // FETCH

          var generator = new MessageGenerator(server, message, 3);
          var cts = new CancellationTokenSource();
          var task = message.SaveAsync(outfile, cts.Token);

          server.WaitForClientRequest();

          cts.Cancel();

          foreach (var m in generator.EnqueueNextBlock()) {
            break;
          }

          var ex = Assert.Throws<AggregateException>(() => task.Wait());

          Assert.IsInstanceOf<OperationCanceledException>(ex.Flatten().InnerException);
          Assert.IsFalse(File.Exists(outfile), "file existence");

          Assert.That(server.DequeueRequest(), Is.StringEnding("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n"));
          Assert.IsNull(server.DequeueRequest());

          Assert.IsFalse(message.Client.IsBusy);
        }
        finally {
          File.Delete(outfile);
        }
      });
    }

    [Test]
    public void TestSaveAsync_CancellationRequestedWhileFetching()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        const string outfile = "fetched.eml";

        try {
          File.Delete(outfile);

          server.EnqueueTaggedResponse("UID FETCH 1 (UID FLAGS)\r\n",
                                       string.Format("* FETCH 1 (UID 1 FLAGS ({0}))\r\n", string.Empty) +
                                     "$tag OK done\r\n");

          var message = mailbox.GetMessageByUid(1L, ImapMessageFetchAttributeOptions.DynamicAttributes);

          server.DequeueRequest(); // FETCH

          var cts = new CancellationTokenSource();
          var generator = new MessageGenerator(server, message, 3,
                                               (block) => { if (block == 1) cts.Cancel(); },
                                               null);
          var task = message.SaveAsync(outfile, cts.Token);

          server.WaitForClientRequest();

          foreach (var m in generator.EnqueueNextBlock()) {
            if (cts.IsCancellationRequested)
              break;
            else
              server.WaitForClientRequest(TimeSpan.FromSeconds(1.0));
          }

          var ex = Assert.Throws<AggregateException>(() => task.Wait());

          Assert.IsInstanceOf<OperationCanceledException>(ex.Flatten().InnerException);
          Assert.IsTrue(File.Exists(outfile), "file existence");

          var fileLength = (new FileInfo(outfile)).Length;

          Assert.Greater(fileLength, 0, "file length");
          Assert.Less(fileLength, generator.MessageBodyOctet, "file length");

          Assert.That(server.DequeueRequest(), Is.StringEnding("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n"));
          Assert.That(server.DequeueRequest(), Is.StringEnding("UID FETCH 1 (BODY.PEEK[]<10240.10240>)\r\n"));
          Assert.IsNull(server.DequeueRequest());

          Assert.IsFalse(message.Client.IsBusy);
        }
        finally {
          File.Delete(outfile);
        }
      });
    }

    [Test]
    public void TestSaveAsync_OverwriteExistingFile()
    {
      const string outfile = "fetched.eml";
      const int existingFileLength = 10 * 1024;

      try {
        File.Delete(outfile);

        File.WriteAllText(outfile, new string('x', existingFileLength));

        TestFetchMessage(delegate(ImapPseudoServer2 server, ImapMessageInfo message, string expectedMessageBody, int expectedOctets) {
          Assert.Less(expectedOctets, (new FileInfo(outfile)).Length, "less than existing file length");

          message.SaveAsync(outfile).Wait();

          Assert.That(server.DequeueRequest(), Is.StringEnding("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n"));

          Assert.IsTrue(File.Exists(outfile), "file existence");
          Assert.AreEqual(expectedOctets, (new FileInfo(outfile)).Length, "file length");
          Assert.AreEqual(NetworkTransferEncoding.Transfer8Bit.GetBytes(expectedMessageBody),
                          File.ReadAllBytes(outfile));

          Assert.IsFalse(message.Client.IsBusy);
        });
      }
      finally {
        File.Delete(outfile);
      }
    }

    [Test]
    public void TestSaveAsync_PathNull()
    {
      TestFetchMessage(false,
                       null,
                       delegate(ImapPseudoServer2 server,
                                  ImapMessageInfo message,
                                  string expectedMessageBody,
                                  int expectedOctets) {
        Assert.Throws<ArgumentNullException>(() => message.SaveAsync(null));

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestSaveAsync_PathEmpty()
    {
      TestFetchMessage(false,
                       null,
                       delegate(ImapPseudoServer2 server,
                                  ImapMessageInfo message,
                                  string expectedMessageBody,
                                  int expectedOctets) {
        Assert.Throws<ArgumentException>(() => message.SaveAsync(string.Empty));

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestSaveAsync_CanNotOpenFetchStream()
    {
      const string outfile = "fetched.eml";

      try {
        File.Delete(outfile);

        TestUtils2.TestMessage(delegate(ImapPseudoServer2 server, ImapMessageInfo message) {
          // FETCH (message vanished)
          server.EnqueueTaggedResponse("$tag OK done\r\n");

          var t = message.SaveAsync(outfile);

          var ex = Assert.Throws<AggregateException>(() => t.Wait());

          Assert.IsInstanceOf<ImapMessageVanishedException>(ex.Flatten().InnerException);

          Assert.IsFalse(File.Exists(outfile), "file existence");

          Assert.IsFalse(message.Client.IsBusy);
        });
      }
      finally {
        File.Delete(outfile);
      }
    }

    [Test]
    public void TestDownloadAsync()
    {
#if NET_4_5
      var progress = (IProgress<ImapMessageFetchProgressInfo>)null;
      var downloadAsyncMethods = new[] {
        new Func<ImapMessageInfo, Task<Stream>>(m => m.DownloadAsync()),
        new Func<ImapMessageInfo, Task<Stream>>(m => m.DownloadAsync(progress)),
        new Func<ImapMessageInfo, Task<Stream>>(m => m.DownloadAsync(CancellationToken.None)),
        new Func<ImapMessageInfo, Task<Stream>>(m => m.DownloadAsync(CancellationToken.None, progress)),
        new Func<ImapMessageInfo, Task<Stream>>(m => m.DownloadAsync(ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, Task<Stream>>(m => m.DownloadAsync(ImapMessageFetchBodyOptions.Default, progress)),
        new Func<ImapMessageInfo, Task<Stream>>(m => m.DownloadAsync(ImapMessageFetchBodyOptions.Default, CancellationToken.None)),
        new Func<ImapMessageInfo, Task<Stream>>(m => m.DownloadAsync(ImapMessageFetchBodyOptions.Default, CancellationToken.None, progress)),
      };
#else
      var downloadAsyncMethods = new[] {
        new Func<ImapMessageInfo, Task<Stream>>(m => m.DownloadAsync()),
        new Func<ImapMessageInfo, Task<Stream>>(m => m.DownloadAsync(CancellationToken.None)),
        new Func<ImapMessageInfo, Task<Stream>>(m => m.DownloadAsync(ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, Task<Stream>>(m => m.DownloadAsync(ImapMessageFetchBodyOptions.Default, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < downloadAsyncMethods.Length; i++) {
        TestFetchMessage(false,
                         null,
                         delegate(ImapPseudoServer2 server,
                                  ImapMessageInfo message,
                                  string expectedMessageBody,
                                  int expectedOctets) {
          try {
            var ret = downloadAsyncMethods[i](message).Result;

            Assert.IsNotNull(ret);
            Assert.IsTrue(ret.CanSeek, "CanSeek");
            Assert.IsTrue(ret.CanRead, "CanRead");
            Assert.IsTrue(ret.CanWrite, "CanWrite");
            Assert.AreEqual(0L, ret.Position, "Position");
            Assert.AreEqual(expectedOctets, ret.Length, "Length");

            Assert.DoesNotThrow(() => ret.Seek(-1L, SeekOrigin.End));
            Assert.AreNotEqual(-1, ret.ReadByte());
            Assert.DoesNotThrow(() => ret.Seek(0L, SeekOrigin.Begin));

            Assert.AreEqual(expectedMessageBody,
                            NetworkTransferEncoding.Transfer8Bit.GetString(ret.ReadToEnd()));
          }
          catch (Exception ex) {
            Assert.Fail("unexpected exception (overload #{0}) : {1}", i, ex);
          }

          Assert.IsFalse(message.Client.IsBusy);
        });
      }
    }

    [Test]
    public void TestOpenTextAsync()
    {
      var encoding = Encoding.GetEncoding("ISO-8859-1");
#if NET_4_5
      var progress = (IProgress<ImapMessageFetchProgressInfo>)null;
      var openTextAsyncMethods = new[] {
        new Func<ImapMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync()),
        new Func<ImapMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(encoding)),
        new Func<ImapMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(encoding, ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(encoding, ImapMessageFetchBodyOptions.Default, CancellationToken.None)),
        new Func<ImapMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(encoding, ImapMessageFetchBodyOptions.Default, progress)),
        new Func<ImapMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(encoding, ImapMessageFetchBodyOptions.Default, CancellationToken.None, progress)),
      };
#else
      var openTextAsyncMethods = new[] {
        new Func<ImapMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync()),
        new Func<ImapMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(encoding)),
        new Func<ImapMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(encoding, ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(encoding, ImapMessageFetchBodyOptions.Default, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < openTextAsyncMethods.Length; i++) {
        TestFetchMessage(false,
                         null,
                         delegate(ImapPseudoServer2 server,
                                ImapMessageInfo message,
                                string expectedMessageBody,
                                int expectedOctets) {
          try {
            using (var reader = openTextAsyncMethods[i](message).Result) {
              Assert.IsNotNull(reader);
              Assert.AreEqual(encoding, reader.CurrentEncoding);
              Assert.AreEqual(expectedMessageBody, reader.ReadToEnd());
            }
          }
          catch (Exception ex) {
            Assert.Fail("unexpected exception (overload #{0}) : {1}", i, ex);
          }

          Assert.IsFalse(message.Client.IsBusy);
        });
      }
    }

    [Test]
    public void TestOpenTextAsync_CharsetNull()
    {
      TestUtils2.TestMessage(delegate(ImapPseudoServer2 server, ImapMessageInfo message) {
        Assert.Throws<ArgumentNullException>(() => message.OpenTextAsync((Encoding)null));

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAsAsync_StreamReaderConverter()
    {
      var encoding = Encoding.GetEncoding("ISO-8859-1");
      var converter = new Converter<StreamReader, string>(s => s.ReadToEnd());
#if NET_4_5
      var progress = (IProgress<ImapMessageFetchProgressInfo>)null;
      var readAsAsyncMethods = new[] {
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAsAsync(converter)),
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAsAsync(ImapMessageFetchBodyOptions.Default, converter)),
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAsAsync(encoding, converter)),
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAsAsync(encoding, ImapMessageFetchBodyOptions.Default, converter)),
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAsAsync(encoding, ImapMessageFetchBodyOptions.Default, converter, CancellationToken.None)),
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAsAsync(encoding, ImapMessageFetchBodyOptions.Default, converter, progress)),
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAsAsync(encoding, ImapMessageFetchBodyOptions.Default, converter, CancellationToken.None, progress)),
      };
#else
      var readAsAsyncMethods = new[] {
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAsAsync(converter)),
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAsAsync(ImapMessageFetchBodyOptions.Default, converter)),
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAsAsync(encoding, converter)),
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAsAsync(encoding, ImapMessageFetchBodyOptions.Default, converter)),
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAsAsync(encoding, ImapMessageFetchBodyOptions.Default, converter, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < readAsAsyncMethods.Length; i++) {
        TestFetchMessage(false,
                         null,
                         delegate(ImapPseudoServer2 server,
                                  ImapMessageInfo message,
                                  string expectedMessageBody,
                                  int expectedOctets) {
          try {
            Assert.AreEqual(expectedMessageBody,
                            readAsAsyncMethods[i](message).Result);
          }
          catch (Exception ex) {
            Assert.Fail("unexpected exception (overload #{0}) : {1}", i, ex);
          }

          Assert.IsFalse(message.Client.IsBusy);
        });
      }
    }

    [Test]
    public void TestReadAsAsync_StreamReaderConverter_ConverterNull()
    {
      var converter = (Converter<StreamReader, string>)null;

      TestUtils2.TestMessage(delegate(ImapPseudoServer2 server, ImapMessageInfo message) {
        Assert.Throws<ArgumentNullException>(() => message.ReadAsAsync(converter));

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAsAsync_StreamReaderConverter_CharsetNull()
    {
      var converter = (Converter<StreamReader, string>)(r => r.ReadToEnd());

      TestUtils2.TestMessage(delegate(ImapPseudoServer2 server, ImapMessageInfo message) {
        Assert.Throws<ArgumentNullException>(() => message.ReadAsAsync((Encoding)null, converter));

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAllLinesAsync()
    {
      var encoding = Encoding.GetEncoding("ISO-8859-1");
#if NET_4_5
      var progress = (IProgress<ImapMessageFetchProgressInfo>)null;
      var readAsAsyncMethods = new[] {
        new Func<ImapMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync()),
        new Func<ImapMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(encoding)),
        new Func<ImapMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(encoding, ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(encoding, ImapMessageFetchBodyOptions.Default, CancellationToken.None)),
        new Func<ImapMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(encoding, ImapMessageFetchBodyOptions.Default, progress)),
        new Func<ImapMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(encoding, ImapMessageFetchBodyOptions.Default, CancellationToken.None, progress)),
      };
#else
      var readAsAsyncMethods = new[] {
        new Func<ImapMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync()),
        new Func<ImapMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(encoding)),
        new Func<ImapMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(encoding, ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(encoding, ImapMessageFetchBodyOptions.Default, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < readAsAsyncMethods.Length; i++) {
        TestFetchMessage(false,
                         null,
                         delegate(ImapPseudoServer2 server,
                                  ImapMessageInfo message,
                                  string expectedMessageBody,
                                  int expectedOctets) {
          try {
            Assert.AreEqual(expectedMessageBody.Replace("\r\n", "\n").Replace("\n", "\r\n").TrimEnd(),
                            string.Join("\r\n", message.ReadAllLinesAsync().Result));
          }
          catch (Exception ex) {
            Assert.Fail("unexpected exception (overload #{0}) : {1}", i, ex);
          }

          Assert.IsFalse(message.Client.IsBusy);
        });
      }
    }

    [Test]
    public void TestReadAllTextAsync()
    {
      var encoding = Encoding.GetEncoding("ISO-8859-1");
#if NET_4_5
      var progress = (IProgress<ImapMessageFetchProgressInfo>)null;
      var readAllTextAsyncMethods = new[] {
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAllTextAsync()),
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAllTextAsync(ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAllTextAsync(encoding)),
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAllTextAsync(encoding, ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAllTextAsync(encoding, ImapMessageFetchBodyOptions.Default, CancellationToken.None)),
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAllTextAsync(encoding, ImapMessageFetchBodyOptions.Default, progress)),
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAllTextAsync(encoding, ImapMessageFetchBodyOptions.Default, CancellationToken.None, progress)),
      };
#else
      var readAllTextAsyncMethods = new[] {
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAllTextAsync()),
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAllTextAsync(ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAllTextAsync(encoding)),
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAllTextAsync(encoding, ImapMessageFetchBodyOptions.Default)),
        new Func<ImapMessageInfo, Task<string>>(m => m.ReadAllTextAsync(encoding, ImapMessageFetchBodyOptions.Default, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < readAllTextAsyncMethods.Length; i++) {
        TestFetchMessage(false,
                         null,
                         delegate(ImapPseudoServer2 server,
                                  ImapMessageInfo message,
                                  string expectedMessageBody,
                                  int expectedOctets) {
          try {
            Assert.AreEqual(expectedMessageBody,
                            readAllTextAsyncMethods[i](message).Result);
          }
          catch (Exception ex) {
            Assert.Fail("unexpected exception (overload #{0}) : {1}", i, ex);
          }

          Assert.IsFalse(message.Client.IsBusy);
        });
      }
    }
  }
}
#endif // NET_4_0
