using System;
using System.Collections.Generic;
using System.Linq;
using NUnit.Framework;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapOpenedMailboxInfoTests {
    [Test]
    public void TestRefresh()
    {
      var selectResp =
        "* EXISTS 3\r\n" +
        "* RECENT 1\r\n" +
        "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.AreEqual(3L, mailbox.ExistMessageCount);
        Assert.AreEqual(1L, mailbox.RecentMessageCount);
        Assert.IsNotNull(mailbox.ApplicableFlags);
        Assert.AreEqual(5, mailbox.ApplicableFlags.Count);
        Assert.IsFalse(mailbox.ApplicableFlags.Contains("$label1"));

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("NOOP\r\n",
                                       "* EXPUNGE 1\r\n" +
                                       "* EXPUNGE 1\r\n" +
                                       "* EXISTS 1\r\n" +
                                       "* RECENT 2\r\n" +
                                       "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft $label1)\r\n" +
                                       "$tag OK done\r\n");
        });

        mailbox.Refresh();

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(1L, mailbox.ExistMessageCount);
        Assert.AreEqual(2L, mailbox.RecentMessageCount);
        Assert.IsNotNull(mailbox.ApplicableFlags);
        Assert.AreEqual(6, mailbox.ApplicableFlags.Count);
        Assert.IsTrue(mailbox.ApplicableFlags.Contains("$label1"));
      });
    }

    [Test]
    public void TestRefresh_ReRetrieveMailboxFlags()
    {
      var selectResp =
        "* EXISTS 3\r\n" +
        "* RECENT 1\r\n" +
        "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        CollectionAssert.DoesNotContain(mailbox.Flags, ImapMailboxFlag.Marked);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" INBOX",
                                       "* LIST (\\Marked) \"/\" \"INBOX\"\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("NOOP\r\n",
                                       "$tag OK done\r\n");
        });

        mailbox.Refresh(true);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.Marked);
      });
    }

    [Test]
    public void TestRefresh_Closed()
    {
      TestUtils2.TestClosedMailbox("INBOX",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var ex = Assert.Throws<ImapUnavailableException>(() => mailbox.Refresh());

        Assert.IsFalse(mailbox.Client.IsBusy);

        Smdn.Net.TestUtils.SerializeBinary(ex);
      });
    }

    [Test]
    public void TestOpen()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.IsOpen);

        Assert.AreSame(mailbox, mailbox.Open());

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreSame(mailbox, mailbox.Client.OpenedMailbox);

        Assert.IsTrue(mailbox.IsOpen);
      });
    }

    [Test]
    public void TestOpen_AsReadOnly()
    {
      var selectResp =
        "$tag OK [READ-ONLY] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.IsOpen);

        Assert.AreSame(mailbox, mailbox.Open());

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreSame(mailbox, mailbox.Client.OpenedMailbox);
        Assert.IsTrue(mailbox.IsOpen);
        Assert.IsTrue(mailbox.IsReadOnly);
      });
    }

    [Test]
    public void TestOpen_UidNotSticky()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "* NO [UIDNOTSTICKY]\r\n" +
        "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
        "* OK [PERMANENTFLAGS (\\Deleted \\Seen)] Limited\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.IsOpen);

        Assert.AreSame(mailbox, mailbox.Open());

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreSame(mailbox, mailbox.Client.OpenedMailbox);
        Assert.IsTrue(mailbox.IsOpen);
        Assert.IsFalse(mailbox.IsReadOnly);
        Assert.AreEqual(23L, mailbox.UidValidity);
        Assert.IsFalse(mailbox.IsUidPersistent);

        Assert.IsNotNull(mailbox.ApplicableFlags);
        Assert.AreEqual(5, mailbox.ApplicableFlags.Count);

        Assert.IsNotNull(mailbox.PermanentFlags);
        Assert.AreEqual(2, mailbox.PermanentFlags.Count);
        Assert.IsFalse(mailbox.PermanentFlags.Contains(ImapMessageFlag.AllowedCreateKeywords));
        Assert.IsFalse(mailbox.IsAllowedToCreateKeywords);
      });
    }

    [Test, Ignore("not implemented; reopen")]
    public void TestOpen_OpenNewly()
    {
    }

    [Test, Ignore("not implemented; reopen")]
    public void TestOpen_ReopenClosed()
    {
    }

    [Test, Ignore("not implemented; reopen")]
    public void TestOpen_ReopenReadWriteChange()
    {
    }

    [Test]
    public void TestOpen_ReopenMailbox()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.IsOpen);
        Assert.AreSame(mailbox, mailbox.Client.OpenedMailbox);

        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK done\r\n");

        mailbox.Close();

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                       string.Format("* LIST () \"\" {0}\r\n", mailbox.FullName) +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                       "$tag OK [READ-WRITE] done\r\n"); 
        });

        var reopenedMailbox = mailbox.Client.OpenMailbox(mailbox.FullName);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreSame(mailbox.Client.OpenedMailbox, reopenedMailbox);

        Assert.IsFalse(mailbox.IsOpen);
        Assert.AreNotSame(mailbox, mailbox.Client.OpenedMailbox);
      });
    }

    [Test]
    public void TestClose()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.IsOpen);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("CLOSE\r\n",
                                       "$tag OK done\r\n");
        });

        mailbox.Close();

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsNull(mailbox.Client.OpenedMailbox);
      });
    }

    [Test]
    public void TestClose_AlreadyClosed()
    {
      TestUtils2.TestClosedMailbox("INBOX",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.DoesNotThrow(() => mailbox.Close());

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsNull(mailbox.Client.OpenedMailbox);
      });
    }

    [Test]
    public void TestClose_SameMailboxOpened()
    {
      CloseDisposeSameMailboxOpened(true);
    }

    [Test]
    public void TestDispose()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                     "* LIST () \"\" INBOX\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                     "$tag OK [READ-WRITE] done\r\n");

        var mailbox = client.OpenInbox();

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("CLOSE\r\n",
                                       "$tag OK done\r\n");
        });

        Assert.DoesNotThrow(() => (mailbox as IDisposable).Dispose());

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsNull(client.OpenedMailbox);
      });
    }

    [Test]
    public void TestDispose_AlreadyClosed()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                     "* LIST () \"\" INBOX\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                     "$tag OK [READ-WRITE] done\r\n");

        var mailbox = client.OpenInbox();

        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK done\r\n");

        client.CloseMailbox();

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsNull(client.OpenedMailbox);

        Assert.DoesNotThrow(() => (mailbox as IDisposable).Dispose());

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsNull(client.OpenedMailbox);
      });
    }

    [Test]
    public void TestDispose_AlreadyLoggedOut()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                     "* LIST () \"\" INBOX\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                     "$tag OK [READ-WRITE] done\r\n");

        var mailbox = client.OpenInbox();

        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK closed\r\n");
        server.EnqueueUntaggedByeResponse("* BYE logged out\r\n"); // LOGOUT

        try {
          client.Logout();
        }
        catch (Smdn.Net.Imap4.Protocol.ImapConnectionException) {
          // ignore disconnected
        }

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsFalse(client.IsConnected);

        Assert.DoesNotThrow(() => (mailbox as IDisposable).Dispose());
      });
    }

    [Test]
    public void TestDispose_SameMailboxOpened()
    {
      CloseDisposeSameMailboxOpened(false);
    }

    private void CloseDisposeSameMailboxOpened(bool close)
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                     "* LIST () \"\" INBOX\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                     "$tag OK [READ-WRITE] done\r\n");

        var firstOpenedMailbox = client.OpenInbox();

        Assert.AreSame(client.OpenedMailbox, firstOpenedMailbox);
        Assert.IsTrue(firstOpenedMailbox.IsOpen);

        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK done\r\n");

        firstOpenedMailbox.Close();

        Assert.IsNull(client.OpenedMailbox);
        Assert.IsFalse(firstOpenedMailbox.IsOpen);

        server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                     "* LIST () \"\" INBOX\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                     "$tag OK [READ-WRITE] done\r\n");

        var secondOpenedMailbox = client.OpenInbox();

        Assert.AreNotSame(client.OpenedMailbox, firstOpenedMailbox);
        Assert.AreSame(client.OpenedMailbox, secondOpenedMailbox);
        Assert.IsFalse(firstOpenedMailbox.IsOpen);
        Assert.IsTrue(secondOpenedMailbox.IsOpen);

        if (close)
          Assert.DoesNotThrow(() => firstOpenedMailbox.Close());
        else
          Assert.DoesNotThrow(() => (firstOpenedMailbox as IDisposable).Dispose());

        Assert.IsFalse(firstOpenedMailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestExpunge()
    {
      var selectResp =
        "* EXISTS 3\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.AreEqual(3L, mailbox.ExistMessageCount);

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("FETCH 1:3 (UID FLAGS)\r\n",
                                     "* FETCH 1 (UID 1 FLAGS (\\Deleted))\r\n" +
                                     "* FETCH 2 (UID 2 FLAGS (\\Deleted))\r\n" +
                                     "* FETCH 3 (UID 3 FLAGS (\\Deleted))\r\n" +
                                     "$tag OK done\r\n");

        var messages = mailbox.GetMessages(ImapMessageFetchAttributeOptions.DynamicAttributes).ToArray();

        Assert.AreEqual(3, messages.Length);

        Assert.AreEqual(1L, messages[0].Sequence);
        Assert.AreEqual(1L, messages[0].Uid);
        Assert.IsTrue(messages[0].IsMarkedAsDeleted);
        Assert.IsFalse(messages[0].IsDeleted);

        Assert.AreEqual(2L, messages[1].Sequence);
        Assert.AreEqual(2L, messages[1].Uid);
        Assert.IsTrue(messages[1].IsMarkedAsDeleted);
        Assert.IsFalse(messages[1].IsDeleted);

        Assert.AreEqual(3L, messages[2].Sequence);
        Assert.AreEqual(3L, messages[2].Uid);
        Assert.IsTrue(messages[2].IsMarkedAsDeleted);
        Assert.IsFalse(messages[2].IsDeleted);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("EXPUNGE\r\n",
                                       "* EXPUNGE 3\r\n" +
                                       "* EXPUNGE 2\r\n" +
                                       "* EXPUNGE 1\r\n" +
                                       "* EXISTS 0\r\n" +
                                       "$tag OK done\r\n");
        });

        mailbox.Expunge();

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(0L, mailbox.ExistMessageCount);

        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[0].Sequence);
        Assert.AreEqual(1L, messages[0].Uid);
        Assert.IsTrue(messages[0].IsMarkedAsDeleted);
        Assert.IsTrue(messages[0].IsDeleted);

        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[1].Sequence);
        Assert.AreEqual(2L, messages[1].Uid);
        Assert.IsTrue(messages[1].IsMarkedAsDeleted);
        Assert.IsTrue(messages[1].IsDeleted);

        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[2].Sequence);
        Assert.AreEqual(3L, messages[2].Uid);
        Assert.IsTrue(messages[2].IsMarkedAsDeleted);
        Assert.IsTrue(messages[2].IsDeleted);
      });
    }

    [Test]
    public void TestExpunge_Closed()
    {
      TestUtils2.TestClosedMailbox("INBOX",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var ex = Assert.Throws<ImapUnavailableException>(() => mailbox.Expunge());

        Assert.IsFalse(mailbox.Client.IsBusy);

        Smdn.Net.TestUtils.SerializeBinary(ex);
      });
    }

    [Test]
    public void TestMoveMessagesTo()
    {
      CopyOrMoveMessagesTo(true);
    }

    [Test]
    public void TestCopyMessagesTo()
    {
      CopyOrMoveMessagesTo(false);
    }

    private void CopyOrMoveMessagesTo(bool move)
    {
      var selectResp =
        "* 5 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        server.EnqueueTaggedResponse("LIST \"\" dest\r\n",
                                     "* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = mailbox.Client.GetMailbox("dest");

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("NOOP\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("COPY 1:5 dest\r\n",
                                       "$tag OK done\r\n");

          if (move)
            server.EnqueueTaggedResponse("STORE 1:5 +FLAGS (\\Deleted)\r\n",
                                         "* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                         "* FETCH 2 (FLAGS (\\Deleted))\r\n" +
                                         "* FETCH 3 (FLAGS (\\Deleted))\r\n" +
                                         "* FETCH 4 (FLAGS (\\Deleted))\r\n" +
                                         "* FETCH 5 (FLAGS (\\Deleted))\r\n" +
                                         "$tag OK done\r\n");
        });

        if (move)
          mailbox.MoveMessagesTo(destMailbox);
        else
          mailbox.CopyMessagesTo(destMailbox);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestMoveMessagesTo_DestinationNull()
    {
      var selectResp =
        "* 5 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.MoveMessagesTo(null));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestCopyMessagesTo_DestinationNull()
    {
      var selectResp =
        "* 5 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.CopyMessagesTo(null));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestMoveMessagesTo_MoveToMailboxOfDifferentSession()
    {
      CopyOrMoveMessagesToMailboxOfDifferentSession(true);
    }

    [Test]
    public void TestCopyMessagesTo_CopyToMailboxOfDifferentSession()
    {
      CopyOrMoveMessagesToMailboxOfDifferentSession(false);
    }

    private void CopyOrMoveMessagesToMailboxOfDifferentSession(bool move)
    {
      var selectResp =
        "* 3 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer2 serverSource, ImapOpenedMailboxInfo mailboxSource) {
        TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 serverDest, ImapClient clientDest) {
          serverDest.EnqueueTaggedResponse("LIST ", "* LIST () \"\" dest\r\n" +
                                           "$tag OK done\r\n");

          var mailboxDest = clientDest.GetMailbox("dest");

          serverDest.DequeueRequest();

          Assert.AreNotSame(mailboxSource.Client, mailboxDest.Client);

          var responseTask = TestUtils2.QueueResponseTask(serverSource, mailboxSource.Client, delegate {
            // NOOP (source)
            serverSource.EnqueueTaggedResponse("NOOP\r\n",
                                               "$tag OK done\r\n");
            // FETCH (source)
            serverSource.EnqueueTaggedResponse("FETCH 1:3 (UID)\r\n",
                                               "* FETCH 1 (UID 1)\r\n" +
                                               "* FETCH 2 (UID 2)\r\n" +
                                               "* FETCH 3 (UID 3)\r\n" +
                                               "$tag OK done\r\n");
            serverSource.EnqueueTaggedResponse("UID FETCH 1 (RFC822.SIZE INTERNALDATE FLAGS BODY.PEEK[]<0.10240>)\r\n",
                                               "* FETCH 1 (" +
                                               "FLAGS (\\Answered $label1) " +
                                               "INTERNALDATE \"25-Jan-2011 15:29:06 +0900\" " +
                                               "RFC822.SIZE 13 " +
                                               "BODY[] {13}\r\ntest message1)\r\n" +
                                               "$tag OK done\r\n");
            serverSource.EnqueueTaggedResponse("UID FETCH 2 (RFC822.SIZE INTERNALDATE FLAGS BODY.PEEK[]<0.10240>)\r\n",
                                               "* FETCH 2 (" +
                                               "FLAGS (\\Answered $label2) " +
                                               "INTERNALDATE \"25-Jan-2011 15:29:06 +0900\" " +
                                               "RFC822.SIZE 13 " +
                                               "BODY[] {13}\r\ntest message2)\r\n" +
                                               "$tag OK done\r\n");
            serverSource.EnqueueTaggedResponse("UID FETCH 3 (RFC822.SIZE INTERNALDATE FLAGS BODY.PEEK[]<0.10240>)\r\n",
                                               "* FETCH 3 (" +
                                               "FLAGS (\\Answered $label3) " +
                                               "INTERNALDATE \"25-Jan-2011 15:29:06 +0900\" " +
                                               "RFC822.SIZE 13 " +
                                               "BODY[] {13}\r\ntest message3)\r\n" +
                                               "$tag OK done\r\n");

            if (move)
              // UID STORE (source)
              serverSource.EnqueueTaggedResponse("STORE 1:3 +FLAGS (\\Deleted)\r\n",
                                                 "* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                                 "* FETCH 2 (FLAGS (\\Deleted))\r\n" +
                                                 "* FETCH 3 (FLAGS (\\Deleted))\r\n" +
                                                 "$tag OK done\r\n");
          });

          // APPEND (dest)
          serverDest.EnqueueTaggedResponse("APPEND dest (\\Answered $label1) \"25-Jan-2011 15:29:06 +0900\" {13}\r\n" +
                                           "test message1\r\n",
                                           "$tag OK [APPENDUID 38505 3955] APPEND completed\r\n");
          serverDest.EnqueueTaggedResponse("APPEND dest (\\Answered $label2) \"25-Jan-2011 15:29:06 +0900\" {13}\r\n" +
                                           "test message2\r\n",
                                           "$tag OK [APPENDUID 38505 3956] APPEND completed\r\n");
          serverDest.EnqueueTaggedResponse("APPEND dest (\\Answered $label3) \"25-Jan-2011 15:29:06 +0900\" {13}\r\n" +
                                           "test message3\r\n",
                                           "$tag OK [APPENDUID 38505 3957] APPEND completed\r\n");

          if (move)
            mailboxSource.MoveMessagesTo(mailboxDest);
          else
            mailboxSource.CopyMessagesTo(mailboxDest);

          responseTask.Wait();

          Assert.IsFalse(mailboxSource.Client.IsBusy);
        });
      });
    }
  }
}
