using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapOpenedMailboxInfoOperationsWaitingTests {
    private void AssertBusyAndIdlingStateProperties(ImapOpenedMailboxInfo mailbox, bool expected, string message)
    {
      Assert.AreEqual(expected, mailbox.Client.IsBusy, "Client.IsBusy " + message);
      Assert.AreEqual(expected, mailbox.Client.IsIdling, "Client.IsIdling " + message);
      Assert.AreEqual(expected, mailbox.IsIdling, "IsIdling " + message);
    }

#if DEBUG
    private const int testIdleRepetition = 3;
#else
    private const int testIdleRepetition = 25;
#endif

    [Test]
    public void TestIdle_IdleCapable()
    {
      TestUtils2.TestOpenedMailbox(new[] { ImapCapability.Idle },
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        TestUtils.Repeat(testIdleRepetition, delegate {
          var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
            server.EnqueueResponse("IDLE\r\n",
                                   "+ idling\r\n");

            server.WaitForClientRequest(); // DONE

            AssertBusyAndIdlingStateProperties(mailbox, true, "while idling");

            // DONE
            server.EnqueueTaggedResponse("$tag OK done\r\n");
          });

          AssertBusyAndIdlingStateProperties(mailbox, false, "before idle");

          var timeout = TimeSpan.FromSeconds(0.5);

          Smdn.Net.TestUtils.Elapses(timeout, delegate {
            mailbox.Idle(timeout);
          });

          responseTask.Wait();

          AssertBusyAndIdlingStateProperties(mailbox, false, "after idle");
        });
      });
    }

    [Test]
    public void TestIdle_IdleCapable_NoResponse()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          AssertBusyAndIdlingStateProperties(mailbox, true, "after idle started");

          server.EnqueueTaggedResponse("IDLE\r\n",
                                       "$tag NO not available\r\n");
        });

        AssertBusyAndIdlingStateProperties(mailbox, false, "before idle");

        var timeout = TimeSpan.FromSeconds(1.0);

        Smdn.Net.TestUtils.NotElapse(timeout, delegate {
          Assert.Throws<ImapErrorResponseException>(() => mailbox.Idle(timeout));
        });

        responseTask.Wait();

        AssertBusyAndIdlingStateProperties(mailbox, false, "after idle");
      });
    }

    [Test]
    [TestCase(false)]
    [TestCase(true)]
    public void TestIdle_IdleCapable_ProcessSizeAndStatusResponseOnlyOnce_Expunge(bool asPushResponse)
    {
      var selectResp =
        "* EXISTS 3\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.AreEqual(3, mailbox.ExistMessageCount);

        var eventExistMessageCountChangedRaisedCount = 0;

        mailbox.Client.ExistMessageCountChanged += delegate(object sender, ImapMailboxSizeChangedEventArgs e) {
          eventExistMessageCountChangedRaisedCount++;
        };

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueResponse("IDLE\r\n",
                               "+ idling\r\n");

          if (asPushResponse)
            server.EnqueuePushResponse("* EXPUNGE 2\r\n");

          server.WaitForClientRequest(); // DONE

          AssertBusyAndIdlingStateProperties(mailbox, true, "while idling");

          // DONE
          if (asPushResponse)
            server.EnqueueTaggedResponse("$tag OK done\r\n");
          else
            server.EnqueueTaggedResponse("* EXPUNGE 2\r\n" +
                                         "$tag OK done\r\n");
        });

        AssertBusyAndIdlingStateProperties(mailbox, false, "before idle");

        var timeout = TimeSpan.FromSeconds(0.5);

        Smdn.Net.TestUtils.Elapses(timeout, delegate {
          mailbox.Idle(timeout);
        });

        responseTask.Wait();

        AssertBusyAndIdlingStateProperties(mailbox, false, "after idle");

        Assert.AreEqual(2, mailbox.ExistMessageCount);
        Assert.AreEqual(1, eventExistMessageCountChangedRaisedCount);
      });
    }

    [Test]
    [TestCase(false)]
    [TestCase(true)]
    public void TestIdle_IdleCapable_ProcessSizeAndStatusResponseOnlyOnce_Recent(bool asPushResponse)
    {
      var selectResp =
        "* RECENT 0\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.AreEqual(0, mailbox.RecentMessageCount);

        var eventRecentMessageCountChangedRaisedCount = 0;

        mailbox.Client.RecentMessageCountChanged += delegate(object sender, ImapMailboxSizeChangedEventArgs e) {
          eventRecentMessageCountChangedRaisedCount++;
        };

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n");

          if (asPushResponse)
            server.EnqueuePushResponse("* RECENT 1\r\n");

          server.WaitForClientRequest(); // DONE

          AssertBusyAndIdlingStateProperties(mailbox, true, "while idling");

          // DONE
          if (asPushResponse)
            server.EnqueueTaggedResponse("$tag OK done\r\n");
          else
            server.EnqueueTaggedResponse("* RECENT 1\r\n" +
                                         "$tag OK done\r\n");
        });

        AssertBusyAndIdlingStateProperties(mailbox, false, "before idle");

        var timeout = TimeSpan.FromSeconds(0.5);

        Smdn.Net.TestUtils.Elapses(timeout, delegate {
          mailbox.Idle(timeout);
        });

        responseTask.Wait();

        AssertBusyAndIdlingStateProperties(mailbox, false, "after idle");

        Assert.AreEqual(1, mailbox.RecentMessageCount);
        Assert.AreEqual(1, eventRecentMessageCountChangedRaisedCount);
      });
    }

    [Test]
    public void TestIdle_IdleCapable_TimeoutZero()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        TestUtils.Repeat(testIdleRepetition, delegate {
          var delay = TimeSpan.FromSeconds(0.1);

          var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
            server.EnqueueResponse("IDLE\r\n",
                                   "+ idling\r\n");

            Thread.Sleep(delay);

            // DONE
            server.EnqueueTaggedResponse("$tag OK done\r\n");
          });

          Smdn.Net.TestUtils.Elapses(delay, delegate {
            mailbox.Idle(0);
          });

          responseTask.Wait();
        });
      });
    }

    [Test]
    public void TestIdle_IdleCapable_DisconnectFromServer()
    {
      TestUtils.Repeat(testIdleRepetition, delegate {
        TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                     "INBOX",
                                     "$tag OK [READ-WRITE] done\r\n",
                                     delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
          var delay = TimeSpan.FromSeconds(0.1);

          var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
            server.EnqueueResponse("IDLE\r\n",
                                   "+ idling\r\n");

            Thread.Sleep(delay);

            server.EnqueueUntaggedByeResponse("* BYE Server shutting down.\r\n", true);
          });

          Smdn.Net.TestUtils.Elapses(delay, delegate {
            Assert.Throws<ImapConnectionException>(() => mailbox.Idle(TimeSpan.FromSeconds(1.0)));
          });

          responseTask.Wait();

          Assert.IsFalse(mailbox.Client.IsBusy);
        });
      });
    }

    [Test]
    public void TestIdle_IdleIncapable()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   IdleUnavailable);
    }

    [Test]
    public void TestIdle_IdleDisabled()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        mailbox.Client.DisabledCapabilities.Add(ImapCapability.Idle);

        IdleUnavailable(server, mailbox);
      });
    }

    private void IdleUnavailable(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox)
    {
      TestUtils.Repeat(testIdleRepetition, delegate {
        // NOOP (initial)
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");
        // NOOP (first polling)
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");
        // NOOP (stopped)
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");

        var timeout = TimeSpan.FromSeconds(0.5);

        Smdn.Net.TestUtils.Elapses(timeout, delegate {
          AssertBusyAndIdlingStateProperties(mailbox, false, "before idle");

          mailbox.Idle(timeout, TimeSpan.FromSeconds(0.35));
        });

        AssertBusyAndIdlingStateProperties(mailbox, false, "after idle");
      });
    }

    [Test]
    public void TestIdle_IdleIncapable_TimeoutZero()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        TestUtils.Repeat(testIdleRepetition, delegate {
          // NOOP (initial)
          server.EnqueueTaggedResponse("NOOP\r\n",
                                       "$tag OK done\r\n");
          // NOOP (stopped)
          server.EnqueueTaggedResponse("NOOP\r\n",
                                       "$tag OK done\r\n");

          Smdn.Net.TestUtils.Elapses(TimeSpan.Zero, delegate {
            mailbox.Idle(0, 1000);
          });
        });
      });
    }

    [Test]
    public void TestIdle_IdleIncapable_DisconnectFromServer()
    {
      TestUtils.Repeat(testIdleRepetition, delegate {
        TestUtils2.TestOpenedMailbox(new ImapCapability[0],
                                     "INBOX",
                                     "$tag OK [READ-WRITE] done\r\n",
                                     delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
          // NOOP (initial)
          server.EnqueueTaggedResponse("NOOP\r\n",
                                       "$tag OK done\r\n");
          // NOOP (first polling)
          server.EnqueueUntaggedByeResponse("* BYE Server shutting down.\r\n", false);

          var timeout = TimeSpan.FromSeconds(2.5);
          var pollingInterval = TimeSpan.FromSeconds(0.25);

          Smdn.Net.TestUtils.Elapses(pollingInterval, delegate {
            Assert.Throws<ImapConnectionException>(() => mailbox.Idle(timeout, pollingInterval));
          });

          Assert.IsFalse(mailbox.Client.IsBusy);
        });
      });
    }

    [Test]
    public void TestIdle_StopIdleWhileAwaiting()
    {
      TestUtils2.TestOpenedMailbox(new[] { ImapCapability.Idle },
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        TestUtils.Repeat(testIdleRepetition, delegate {
          var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
            server.EnqueueResponse("IDLE\r\n",
                                   "+ idling\r\n");

            Thread.Sleep(TimeSpan.FromSeconds(0.1));

            // DONE
            server.EnqueueTaggedResponse("$tag OK done\r\n");

            Assert.DoesNotThrow(() => mailbox.StopIdle());
          });

          var timeout = TimeSpan.FromSeconds(1.0);

          Smdn.Net.TestUtils.NotElapse(timeout, delegate {
            Assert.Throws<InvalidOperationException>(() => mailbox.Idle(timeout));
          });

          responseTask.Wait();

          Assert.IsFalse(mailbox.Client.IsBusy);
        });
      });
    }

    [Test]
    public void TestIdle_TimeoutInfinite()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.Idle(Timeout.Infinite));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestIdle_TimeoutArgumentOutOfRange()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.Idle(-2));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestStartIdle_IdleCapable_PollingIntervalInifinite()
    {
      StartIdleIdleCapablePollingIntervalIgnored(Timeout.Infinite);
    }

    [Test]
    public void TestStartIdle_IdleCapable_PollingIntervalArgumentOutOfRange_NegativeValue()
    {
      StartIdleIdleCapablePollingIntervalIgnored(-2);
    }

    [Test]
    public void TestStartIdle_IdleCapable_PollingIntervalArgumentOutOfRange_GreaterThan30Minutes()
    {
      StartIdleIdleCapablePollingIntervalIgnored((int)TimeSpan.FromMinutes(30.0).TotalMilliseconds);
    }

    private void StartIdleIdleCapablePollingIntervalIgnored(int millisecondsPollingInterval)
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        TestUtils.Repeat(testIdleRepetition, delegate {
          using (var waitHandle = new ManualResetEvent(false)) {
            var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
              server.EnqueueResponse("IDLE\r\n",
                                     "+ idling\r\n",
                                     () => waitHandle.Set(),
                                     null);

              server.WaitForClientRequest(); // DONE

              // DONE
              server.EnqueueTaggedResponse("$tag OK done\r\n");
            });

            var delay = TimeSpan.FromSeconds(0.25);

            Smdn.Net.TestUtils.Elapses(delay, delegate {
              AssertBusyAndIdlingStateProperties(mailbox, false, "before idle");

              mailbox.StartIdle(millisecondsPollingInterval);

              waitHandle.WaitOne();

              AssertBusyAndIdlingStateProperties(mailbox, true, "while idling");

              Thread.Sleep(delay);

              mailbox.StopIdle();

              AssertBusyAndIdlingStateProperties(mailbox, false, "after idle");
            });

            responseTask.Wait();
          }
        });
      });
    }

    [Test]
    public void TestStartIdle_IdleIncapable_PollingIntervalInfinite()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.StartIdle(Timeout.Infinite));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestStartIdle_IdleIncapable_PollingIntervalArgumentOutOfRange_NegativeValue()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.StartIdle(-2));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestStartIdle_IdleIncapable_PollingIntervalArgumentOutOfRange_GreaterThan30Minutes()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.StartIdle(TimeSpan.FromMinutes(30.0)));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestStartIdle_AnotherWatchOperationProceeding_1()
    {
      TestUtils2.TestOpenedMailbox(new[] { ImapCapability.Idle },
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        TestUtils.Repeat(testIdleRepetition, delegate {
          AssertBusyAndIdlingStateProperties(mailbox, false, "before idle");

          using (var waitHandle = new ManualResetEvent(false)) {
            // first
            server.EnqueueResponse("IDLE\r\n",
                                   "+ idling\r\n",
                                   () => waitHandle.Set(),
                                   null);

            mailbox.StartIdle();

            waitHandle.WaitOne();

            AssertBusyAndIdlingStateProperties(mailbox, true, "while idling");
          }

          // second
          Assert.Throws<InvalidOperationException>(() => mailbox.StartIdle());

          server.EnqueueTaggedResponse("$tag OK done\r\n");

          mailbox.StopIdle();

          AssertBusyAndIdlingStateProperties(mailbox, false, "after idle");
        });
      });
    }

    [Test]
    public void TestStartIdle_AnotherWatchOperationProceeding_2()
    {
      TestUtils2.TestOpenedMailbox(new[] { ImapCapability.Idle },
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        TestUtils.Repeat(testIdleRepetition, delegate {
          AssertBusyAndIdlingStateProperties(mailbox, false, "before idle");

          using (var waitHandle = new ManualResetEvent(false)) {
            // first
            server.EnqueueResponse("IDLE\r\n",
                                   "+ idling\r\n",
                                   () => waitHandle.Set(),
                                   null);

            mailbox.StartWaitForMessageArrival();

            waitHandle.WaitOne();

            AssertBusyAndIdlingStateProperties(mailbox, true, "while idling");
          }

          // second
          Assert.Throws<InvalidOperationException>(() => mailbox.StartIdle());

          server.EnqueueTaggedResponse("$tag OK done\r\n");

          mailbox.StopWaitForMessageArrival();

          AssertBusyAndIdlingStateProperties(mailbox, false, "after idle");
        });
      });
    }

    [Test]
    public void TestStopIdle_WatchOperationNotStarted()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<InvalidOperationException>(() => mailbox.StopIdle());

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestStopIdle_AnotherWatchOperationProceeding()
    {
      TestUtils2.TestOpenedMailbox(new[] { ImapCapability.Idle },
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        TestUtils.Repeat(testIdleRepetition, delegate {
          using (var waitHandle = new ManualResetEvent(false)) {
            server.EnqueueResponse("IDLE\r\n",
                                   "+ idling\r\n",
                                   () => waitHandle.Set(),
                                   null);

            mailbox.StartWaitForMessageArrival();

            waitHandle.WaitOne();

            AssertBusyAndIdlingStateProperties(mailbox, true, "while idling");
          }

          Assert.Throws<InvalidOperationException>(() => mailbox.StopIdle());

          server.EnqueueTaggedResponse("$tag OK done\r\n");

          mailbox.StopWaitForMessageArrival();

          AssertBusyAndIdlingStateProperties(mailbox, false, "after idle");
        });
      });
    }

    [Test]
    public void TestWaitForMessageArrival_IdleCapable()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 1]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] { ImapCapability.Idle },
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        TestUtils.Repeat(testIdleRepetition, delegate {
          var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
            server.EnqueueResponse("IDLE\r\n",
                                   "+ idling\r\n");

            AssertBusyAndIdlingStateProperties(mailbox, true, "while idling");

            server.WaitForClientRequest(); // DONE

            // DONE
            server.EnqueueTaggedResponse("$tag OK done\r\n");
          });

          AssertBusyAndIdlingStateProperties(mailbox, false, "before idle");

          var timeout = TimeSpan.FromSeconds(0.5);
          ImapMessageInfoList arrivalMessages = null;

          Smdn.Net.TestUtils.Elapses(timeout, delegate {
            arrivalMessages = mailbox.WaitForMessageArrival(timeout);
          });

          responseTask.Wait();

          Assert.IsNotNull(arrivalMessages);
          CollectionAssert.IsEmpty(arrivalMessages);

          AssertBusyAndIdlingStateProperties(mailbox, false, "after idle");
        });
      });
    }

    [Test]
    public void TestWaitForMessageArrival_IdleCapable_NoResponse()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 1]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          AssertBusyAndIdlingStateProperties(mailbox, true, "after idle started");

          server.EnqueueTaggedResponse("IDLE\r\n",
                                       "$tag NO unavailable\r\n");
        });

        AssertBusyAndIdlingStateProperties(mailbox, false, "before idle");

        var timeout = TimeSpan.FromSeconds(1.0);

        Smdn.Net.TestUtils.NotElapse(timeout, delegate {
          Assert.Throws<ImapErrorResponseException>(() => mailbox.WaitForMessageArrival(timeout));
        });

        responseTask.Wait();

        AssertBusyAndIdlingStateProperties(mailbox, false, "after idle");
      });
    }

    [Test]
    public void TestWaitForMessageArrival_IdleCapable_TimeoutZero()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 1]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var delay = TimeSpan.FromSeconds(0.1);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n");

          Thread.Sleep(delay);

          // DONE
          server.EnqueueTaggedResponse("$tag OK done\r\n");
        });

        ImapMessageInfoList arrivalMessages = null;

        Smdn.Net.TestUtils.Elapses(delay, delegate {
          arrivalMessages = mailbox.WaitForMessageArrival(0);
        });

        responseTask.Wait();

        Assert.IsNotNull(arrivalMessages);
        CollectionAssert.IsEmpty(arrivalMessages);
      });
    }

    [Test]
    public void TestWaitForMessageArrival_IdleCapable_TimeoutInfinite()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 1]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var delay = TimeSpan.FromSeconds(0.1);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n");

          Thread.Sleep(delay);

          server.EnqueuePushResponse("* RECENT 1\r\n");
          server.EnqueuePushResponse("* EXISTS 1\r\n");

          server.WaitForClientRequest();

          // DONE
          server.EnqueueTaggedResponse("$tag OK done\r\n");
        });

        ImapMessageInfoList arrivalMessages = null;

        Smdn.Net.TestUtils.Elapses(delay, delegate {
          arrivalMessages = mailbox.WaitForMessageArrival(Timeout.Infinite);
        });

        responseTask.Wait();

        Assert.IsNotNull(arrivalMessages);

        server.EnqueueTaggedResponse("FETCH ",
                                     "* FETCH 1 (UID 1)\r\n" +
                                     "$tag OK done\r\n");

        var arrivalMessageArray = arrivalMessages.ToArray();

        Assert.AreEqual(1, arrivalMessageArray.Length);
        Assert.AreEqual(1L, arrivalMessageArray[0].Uid, "arrival message uid");
        Assert.AreEqual(1L, arrivalMessageArray[0].Sequence, "arrival message seq");
      });
    }

    [Test]
    public void TestWaitForMessageArrival_IdleIncapable()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 1]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        TestUtils.Repeat(testIdleRepetition, delegate {
          var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
            // NOOP (initial)
            server.EnqueueTaggedResponse("NOOP\r\n",
                                         "$tag OK done\r\n");

            server.WaitForClientRequest();

            // NOOP (first polling)
            server.EnqueueTaggedResponse("NOOP\r\n",
                                         "$tag OK done\r\n");

            server.WaitForClientRequest();

            // NOOP (stopped)
            server.EnqueueTaggedResponse("NOOP\r\n",
                                         "$tag OK done\r\n");
          });

          var timeout = TimeSpan.FromSeconds(0.5);
          ImapMessageInfoList arrivalMessages = null;

          Smdn.Net.TestUtils.Elapses(timeout, delegate {
            arrivalMessages = mailbox.WaitForMessageArrival(timeout,
                                                            TimeSpan.FromSeconds(0.35));
          });

          responseTask.Wait();

          Assert.IsNotNull(arrivalMessages);
          CollectionAssert.IsEmpty(arrivalMessages);
        });
      });
    }

    [Test]
    public void TestWaitForMessageArrival_IdleIncapable_TimeoutZero()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 1]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        // NOOP (initial)
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");
        // NOOP (first polling)
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");

        var pollingInterval = TimeSpan.FromSeconds(0.25);
        ImapMessageInfoList arrivalMessages = null;

        Smdn.Net.TestUtils.NotElapse(pollingInterval, delegate {
          arrivalMessages = mailbox.WaitForMessageArrival(0,
                                                          (int)pollingInterval.TotalMilliseconds);
        });

        Assert.IsNotNull(arrivalMessages);

        var arrivalMessageArray = arrivalMessages.ToArray();

        Assert.AreEqual(0, arrivalMessageArray.Length);
      });
    }

    [Test]
    public void TestWaitForMessageArrival_IdleIncapable_TimeoutInfinite()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 1]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        // NOOP (initial)
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");
        // NOOP (first polling)
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");
        // NOOP (second polling)
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "* RECENT 1\r\n" +
                                     "* EXISTS 1\r\n" +
                                     "$tag OK done\r\n");

        var pollingInterval = TimeSpan.FromSeconds(0.25);
        ImapMessageInfoList arrivalMessages = null;

        Smdn.Net.TestUtils.Elapses(TimeSpan.FromTicks(pollingInterval.Ticks * 2), delegate {
          arrivalMessages = mailbox.WaitForMessageArrival(Timeout.Infinite,
                                                          (int)pollingInterval.TotalMilliseconds);
        });

        Assert.IsNotNull(arrivalMessages);

        server.EnqueueTaggedResponse("FETCH ",
                                     "* FETCH 1 (UID 1)\r\n" +
                                     "$tag OK done\r\n");

        var arrivalMessageArray = arrivalMessages.ToArray();

        Assert.AreEqual(1, arrivalMessageArray.Length);
        Assert.AreEqual(1L, arrivalMessageArray[0].Uid, "arrival message uid");
        Assert.AreEqual(1L, arrivalMessageArray[0].Sequence, "arrival message seq");
      });
    }

    [Test]
    public void TestWaitForMessageArrival_TimeoutArgumentOutOfRange()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.WaitForMessageArrival(-2));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestWaitForMessageArrival_StopWaitForMessageArrivalWhileAwaiting()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 1]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] { ImapCapability.Idle },
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        TestUtils.Repeat(testIdleRepetition, delegate {
          var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
            server.EnqueueResponse("IDLE\r\n",
                                   "+ idling\r\n");

            Thread.Sleep(TimeSpan.FromSeconds(0.1));

            // DONE
            server.EnqueueTaggedResponse("$tag OK done\r\n");

            Assert.DoesNotThrow(() => mailbox.StopWaitForMessageArrival());
          });

          var timeout = TimeSpan.FromSeconds(1.0);

          Smdn.Net.TestUtils.NotElapse(timeout, delegate {
            Assert.Throws<InvalidOperationException>(() => mailbox.WaitForMessageArrival(timeout));
          });

          responseTask.Wait();

          Assert.IsFalse(mailbox.Client.IsBusy);
        });
      });
    }

    [Test]
    public void TestStartWaitForMessageArrival_IdleCapable_PollingIntervalInifinite()
    {
      StartWaitForMessageArrivalIdleCapablePollingIntervalIgnored(-1);
    }

    [Test]
    public void TestStartWaitForMessageArrival_IdleCapable_PollingIntervalArgumentOutOfRange_NegativeValue()
    {
      StartWaitForMessageArrivalIdleCapablePollingIntervalIgnored(-2);
    }

    [Test]
    public void TestStartWaitForMessageArrival_IdleCapable_PollingIntervalArgumentOutOfRange_GreaterThan30Minutes()
    {
      StartWaitForMessageArrivalIdleCapablePollingIntervalIgnored((int)TimeSpan.FromMinutes(30.0).TotalMilliseconds);
    }

    private void StartWaitForMessageArrivalIdleCapablePollingIntervalIgnored(int millisecondsPollingInterval)
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        using (var waitHandle = new ManualResetEvent(false)) {
          var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
            server.EnqueueResponse("IDLE\r\n",
                                   "+ idling\r\n",
                                   () => waitHandle.Set(),
                                   null);

            server.WaitForClientRequest(); // DONE

            // DONE
            server.EnqueueTaggedResponse("$tag OK done\r\n");
          });

          var delay = TimeSpan.FromSeconds(0.25);

          Smdn.Net.TestUtils.Elapses(delay, delegate {
            mailbox.StartWaitForMessageArrival(millisecondsPollingInterval);

            waitHandle.WaitOne();

            Thread.Sleep(delay);

            mailbox.StopWaitForMessageArrival();
          });

          responseTask.Wait();
        }
      });
    }

    [Test]
    public void TestStartWaitForMessageArrival_IdleIncapable_PollingIntervalInfinite()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.StartWaitForMessageArrival(Timeout.Infinite));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestStartWaitForMessageArrival_IdleIncapable_PollingIntervalArgumentOutOfRange_NegativeValue()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.StartWaitForMessageArrival(-2));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestStartWaitForMessageArrival_IdleIncapable_PollingIntervalArgumentOutOfRange_GreaterThan30Minutes()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.StartWaitForMessageArrival(TimeSpan.FromMinutes(30.0)));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestStartWaitForMessageArrival_AnotherWatchOperationProceeding1()
    {
      TestUtils2.TestOpenedMailbox(new[] { ImapCapability.Idle },
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        TestUtils.Repeat(testIdleRepetition, delegate {
          AssertBusyAndIdlingStateProperties(mailbox, false, "before idle");

          using (var waitHandle = new ManualResetEvent(false)) {
            // first
            server.EnqueueResponse("IDLE\r\n",
                                   "+ idling\r\n",
                                   () => waitHandle.Set(),
                                   null);

            mailbox.StartWaitForMessageArrival();

            waitHandle.WaitOne();

            AssertBusyAndIdlingStateProperties(mailbox, true, "while idling");
          }

          // second
          Assert.Throws<InvalidOperationException>(() => mailbox.StartWaitForMessageArrival());

          server.EnqueueTaggedResponse("$tag OK done\r\n");

          mailbox.StopWaitForMessageArrival();

          AssertBusyAndIdlingStateProperties(mailbox, false, "after idle");
        });
      });
    }

    [Test]
    public void TestStartWaitForMessageArrival_AnotherWatchOperationProceeding2()
    {
      TestUtils2.TestOpenedMailbox(new[] { ImapCapability.Idle },
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        TestUtils.Repeat(testIdleRepetition, delegate {
          AssertBusyAndIdlingStateProperties(mailbox, false, "before idle");

          using (var waitHandle = new ManualResetEvent(false)) {
            // first
            server.EnqueueResponse("IDLE\r\n",
                                   "+ idling\r\n",
                                   () => waitHandle.Set(),
                                   null);

            mailbox.StartIdle();

            waitHandle.WaitOne();

            AssertBusyAndIdlingStateProperties(mailbox, true, "while idling");
          }

          // second
          Assert.Throws<InvalidOperationException>(() => mailbox.StartWaitForMessageArrival());

          server.EnqueueTaggedResponse("$tag OK done\r\n");

          mailbox.StopIdle();

          AssertBusyAndIdlingStateProperties(mailbox, false, "after idle");
        });
      });
    }

    [Test]
    public void TestStopWaitForMessageArrival_WatchOperationNotStarted()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<InvalidOperationException>(() => mailbox.StopWaitForMessageArrival());

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestStopWaitForMessageArrival_AnotherWatchOperationProceeding()
    {
      TestUtils2.TestOpenedMailbox(new[] { ImapCapability.Idle },
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        TestUtils.Repeat(testIdleRepetition, delegate {
          AssertBusyAndIdlingStateProperties(mailbox, false, "before idle");

          using (var waitHandle = new ManualResetEvent(false)) {
            server.EnqueueResponse("IDLE\r\n",
                                   "+ idling\r\n",
                                   () => waitHandle.Set(),
                                   null);

            mailbox.StartIdle();

            waitHandle.WaitOne();

            AssertBusyAndIdlingStateProperties(mailbox, true, "while idling");
          }

          Assert.Throws<InvalidOperationException>(() => mailbox.StopWaitForMessageArrival());

          server.EnqueueTaggedResponse("$tag OK done\r\n");

          mailbox.StopIdle();

          AssertBusyAndIdlingStateProperties(mailbox, false, "after idling");
        });
      });
    }

    [Test]
    public void TestWatchOperationProceeding_Refresh()
    {
      TestUtils2.TestOpenedMailbox(new[] { ImapCapability.Idle },
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        TestUtils.Repeat(testIdleRepetition, delegate {
          AssertBusyAndIdlingStateProperties(mailbox, false, "before idle");

          using (var waitHandle = new ManualResetEvent(false)) {
            server.EnqueueResponse("IDLE\r\n",
                                   "+ idling\r\n",
                                   () => waitHandle.Set(),
                                   null);

            mailbox.StartIdle();

            waitHandle.WaitOne();

            AssertBusyAndIdlingStateProperties(mailbox, true, "while idling");
          }

          Assert.Throws<InvalidOperationException>(() => mailbox.Refresh());

          server.EnqueueTaggedResponse("$tag OK done\r\n");

          mailbox.StopIdle();

          AssertBusyAndIdlingStateProperties(mailbox, false, "after idle");
        });
      });
    }

    [Test]
    public void TestWatchOperationProceeding_Close()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        AssertBusyAndIdlingStateProperties(mailbox, false, "before idle");

        using (var waitHandle = new ManualResetEvent(false)) {
          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n",
                                 () => waitHandle.Set(),
                                 null);

          mailbox.StartIdle();

          waitHandle.WaitOne();

          AssertBusyAndIdlingStateProperties(mailbox, true, "while idling");
        }

        Assert.Throws<InvalidOperationException>(() => mailbox.Close());

        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.StopIdle();

        AssertBusyAndIdlingStateProperties(mailbox, false, "after idle");
      });
    }
  }
}

