// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;

using Smdn.Formats;

namespace Smdn.Net.Imap4 {
  // 7.4.2. FETCH Response
  //       ENVELOPE
  //         The fields of the envelope structure are in the following
  //         order: date, subject, from, sender, reply-to, to, cc, bcc,
  //         in-reply-to, and message-id.  The date, subject, in-reply-to,
  //         and message-id fields are strings.  The from, sender, reply-to,
  //         to, cc, and bcc fields are parenthesized lists of address
  //         structures.
  public sealed class ImapEnvelope {
    public string Date {
      get; private set;
    }

    public string Subject {
      get; private set;
    }

#if NET_4_5
    public IReadOnlyList<ImapAddress>
#else
    public IList<ImapAddress>
#endif
    From {
      get; private set;
    }

#if NET_4_5
    public IReadOnlyList<ImapAddress>
#else
    public IList<ImapAddress>
#endif
    Sender {
      get; private set;
    }

#if NET_4_5
    public IReadOnlyList<ImapAddress>
#else
    public IList<ImapAddress>
#endif
    ReplyTo {
      get; private set;
    }

#if NET_4_5
    public IReadOnlyList<ImapAddress>
#else
    public IList<ImapAddress>
#endif
    To {
      get; private set;
    }

#if NET_4_5
    public IReadOnlyList<ImapAddress>
#else
    public IList<ImapAddress>
#endif
    Cc {
      get; private set;
    }

#if NET_4_5
    public IReadOnlyList<ImapAddress>
#else
    public IList<ImapAddress>
#endif
    Bcc {
      get; private set;
    }

    public string InReplyTo {
      get; private set;
    }

    public string MessageId {
      get; private set;
    }

    public ImapEnvelope(string date,
                        string subject,
#if NET_4_5
                        IReadOnlyList<ImapAddress> @from,
                        IReadOnlyList<ImapAddress> sender,
                        IReadOnlyList<ImapAddress> replyTo,
                        IReadOnlyList<ImapAddress> @to,
                        IReadOnlyList<ImapAddress> cc,
                        IReadOnlyList<ImapAddress> bcc,
#else
                        IList<ImapAddress> @from,
                        IList<ImapAddress> sender,
                        IList<ImapAddress> replyTo,
                        IList<ImapAddress> @to,
                        IList<ImapAddress> cc,
                        IList<ImapAddress> bcc,
#endif
                        string inReplyTo,
                        string messageId)
    {
      this.Date       = date;
      this.Subject    = subject;
      this.From       = CollectionUtils.ToNonNullReadOnly(@from);
      this.Sender     = CollectionUtils.ToNonNullReadOnly(sender);
      this.ReplyTo    = CollectionUtils.ToNonNullReadOnly(replyTo);
      this.To         = CollectionUtils.ToNonNullReadOnly(@to);
      this.Cc         = CollectionUtils.ToNonNullReadOnly(cc);
      this.Bcc        = CollectionUtils.ToNonNullReadOnly(bcc);
      this.InReplyTo  = inReplyTo;
      this.MessageId  = messageId;
    }

    public string GetDecodedSubject()
    {
      return GetDecodedSubject(null, false);
    }

    public string GetDecodedSubject(EncodingSelectionCallback selectFallbackEncoding,
                                    bool throwIfMalformedOrUnsupported)
    {
      return MimeUtils.GetDecodedValue(Subject,
                                       selectFallbackEncoding,
                                       throwIfMalformedOrUnsupported);
    }

    public override string ToString()
    {
      return string.Format("{{Date={0}, Subject={1}, From={{{2}}}, Sender={{{3}}}, ReplyTo={{{4}}}, To={{{5}}}, Cc={{{6}}}, Bcc={{{7}}}, InReplyTo={8}, MessageId={9}}}",
                           Date,
                           Subject,
                           CollectionUtils.Join(", ", From),
                           CollectionUtils.Join(", ", Sender),
                           CollectionUtils.Join(", ", ReplyTo),
                           CollectionUtils.Join(", ", To),
                           CollectionUtils.Join(", ", Cc),
                           CollectionUtils.Join(", ", Bcc),
                           InReplyTo,
                           MessageId);
    }
  }
}
/*
envelope        = "(" env-date SP env-subject SP env-from SP
                  env-sender SP env-reply-to SP env-to SP env-cc SP
                  env-bcc SP env-in-reply-to SP env-message-id ")"
env-bcc         = "(" 1*address ")" / nil
env-cc          = "(" 1*address ")" / nil
env-date        = nstring
env-from        = "(" 1*address ")" / nil
env-in-reply-to = nstring
env-message-id  = nstring
env-reply-to    = "(" 1*address ")" / nil
env-sender      = "(" 1*address ")" / nil
env-subject     = nstring
env-to          = "(" 1*address ")" / nil
*/
